%### KARLIM.SL

% This file contains the routines that converts tensor index list to
% internal vector index and back. The algorithms for manipulating more
% that 2 (a)symmetric indices are by JEA.
% The functions in this file expect indices to be of limited size.
% Vector indices are XInums i.e. < 2^ 16. Basic tensor indices are
% Inums i.e. < 1000. Intermediate indices might be XInums but only if
% they are non symmetric, otherwise they are Inums.
% In the very unlikely event that larger indices are needed,
% substitute (X for (XI and (XI for (I . This would mean that vector
% indices are Xnums i.e. < 2^24 and tensor indices are XInums.

(SETLOCALGLOBALS
 SUBIX!$
 !$NOTASY0
 FINXL!$)               % Index list inside TENFIND and TENCFIND.

(SETFLUIDS
 XXND!$)                                % Dimension

(ENTRYPOINTS TENLIMB TENFIND RANGE ONEELM)

(LOCALFUNCTIONS GTSUINX XFIN DFIN DFINN VGTSUINX SYFINE SFIN HFIN AFIN
                XSADD XHADD XAADD SFIN2 AFIN2 HFIN2
%               XLIM DLIMN DLIM1 DLIM KLIM KLIM2 RANKERR
                TENCFIN GTSUCX DCXS DCXSN VGTSUCX ACXS XSSUB XASUB)

(DE TENFIND (XL LIML) (COND
% FINXL!$ (global): index list, LIML: symmetry specification.
% Returns vector index (or NIL in case of 0 by symmetry).
  ((PAIRP (SETQ FINXL!$ XL))
   (SETQ !$NOTASY0 T)
   (SETQ XL (GTSUINX (CAR LIML) (CDR LIML)))
   (COND (!$NOTASY0 XL)) )
  (T XL) ))  

(DE GTSUINX (DMX LIML) 
% DMX: maximum index for subrange, LIML: symmetry spec. for subrange
% Return index for subrange.
 (COND 
  ((NULL LIML) (XFIN DMX))              % one index
  (T (XCASE (CAR LIML) (UNSYMER (CAR LIML)) 
      (DFIN (CDR LIML))                 % no symmetry
      (SFIN DMX (CDR LIML))             % symmetric
      (AFIN DMX (CDR LIML))             % anti-symmetric
      (HFIN DMX (CDR LIML))             % hermitian symmetry
      (DFINN (CADR LIML) (CDDR LIML))   % no symmetry, same dimension
% The following are optimised versions for the common case of 2 indices:
      (SFIN2 DMX (CADR LIML) (CADDR LIML) (CADDDR LIML))
      (AFIN2 DMX (CADR LIML) (CADDR LIML) (CADDDR LIML))
      (HFIN2 DMX (CADR LIML) (CADDR LIML) (CADDDR LIML)) )) ) )

(DE XFIN (DMX) % Get one index and check that it's within range.
% DMX: maximum index.
 (COND
  ((NULL FINXL!$) (RANKERR))            % No more indices! Wrong rank.
  (T (SLET ((X (CAR FINXL!$)))          % Get index.
      (COND ((OR (IMINUSP X) (IGREATERP X DMX))
             (ERRORFT 544 "Index %r is not between 0 and %w" X DMX) ))
      (UNPUSH FINXL!$)                  % Remove index from index list.
      X)) ) )                           % Return index

(DE DFIN (LIML) % Get index for range of non-symmetric indices.
% LIML: sym. spec. has form ((DMX1 . LIML1) (DMX2 . LIML2)...)
% where DMXn is index maximum and LIMLn is sym. spec. for subrange n.
 (SLET ((X (CAR LIML)) (VX (GTSUINX (CAR X) (CDR X))))
  (WHILE (AND (UNPUSH LIML) !$NOTASY0)
   (SETQ X (CAR LIML))
   (SETQ VX         % VX := VX*(DMXn + 1) + GTSUINX(DMXn, LIMLn)
    (XIPLUS (XITIMES VX (XIADD1 (CAR X))) (GTSUINX (CAR X) (CDR X))) ) )
  (COND
   (!$NOTASY0 VX)
   (T 0) ) ) )

(DE DFINN (DMX CNT)
% Get index for CNT non-symmetric indices with dimension DMX and no
% recursion.
 (SLET ((VX (XFIN DMX)) (DMY (IADD1 DMX)))
  (REPEATS (ISUB1 CNT) (SETQ VX (XIPLUS (XITIMES VX DMY) (XFIN DMX))))
  VX) )

(DE VGTSUINX (DMX LIML) (IDIFFERENCE DMX (GTSUINX DMX LIML)))

(DE SYFINE (DMX XL SW) (SLET ((N 0)) (COND
   (!$NOTASY0 (FOREACH X IN XL DO
     (SETQ DMX (XIDIFFERENCE DMX (SXIOVERW X (IINCR N) SW))) )
    DMX)
   (T 0) )))

(DE SFIN (DMX LIML) % Get index for range of symmetric indices.
% DMX: index max. for range,
% LIML: sym. spec. has form (DN LIML1 LIML2 ...)
% All the subranges 1 2 ...  have the same index maximum DN
 (SLET ((DN (CAR LIML)) XL)
  (FOREACH X IN (CDR LIML) DO (SETQ XL (XSADD (VGTSUINX DN X) XL)))
  (SYFINE DMX XL T) ) )

(DE HFIN (DMX LIML) 
 (SLET ((DN (CAR LIML)) XL)
  (FOREACH X IN (CDR LIML) DO (SETQ XL (XHADD (VGTSUINX DN X) XL)))
  (SYFINE DMX XL T) ) )

(DE AFIN (DMX LIML) 
 (SLET ((DN (CAR LIML)) XL)
  (FOREACH X IN (CDR LIML) DO (SETQ XL (XAADD (VGTSUINX DN X) XL)))
  (SYFINE DMX XL NIL) ) )

(DE XSADD (X XL) 
 (COND 
  ((NULL XL) (LIST X))
  ((NOT (IGREATERP X (CAR XL))) (CONS X XL))
  (T (RPLACD XL (XSADD X (CDR XL)))) ) )

(DE XHADD (X XL) 
 (COND 
   ((NULL XL) (LIST X))
   ((ILESSP X (CAR XL)) (CONS X XL))
   ((IEQN X (CAR XL)) (SETQ HDIAG!* T) (CONS X XL))
   (T (SETQ CONJSW (NOT CONJSW)) (RPLACD XL (XHADD X (CDR XL)))) ) )

(DE XAADD (X XL) 
 (COND 
  ((NULL XL) (LIST X))
  ((ILESSP X (CAR XL)) (CONS X XL))
  ((IEQN X (CAR XL)) (SETQ !$NOTASY0 NIL) (CONS X XL))
  (T (SETQ MINSW (NOT MINSW)) (RPLACD XL (XAADD X (CDR XL)))) ) )

(COMMENT
(DE XFIN2 (DMX Y CX)
 (SLET ((DN (CAR Y)))
  (SLET ((X (VGTSUINX DN (CADR Y))))
   (SETQ Y (VGTSUINX DN (CADDR Y)))
   (COND
    ((IEQN X Y)
     (XCASE CX NIL NIL (SETQ !$NOTASY0 NIL) (SETQ HDIAG!* T)) )
    ((ILESSP X Y)
     (EXCHANGE X Y)
     (XCASE CX NIL NIL
      (SETQ MINSW (NOT MINSW))
      (SETQ CONJSW (NOT CONJSW)) ) ) )
   (COND
    (!$NOTASY0 (XIDIFFERENCE DMX (XIPLUS Y (XIQUOTIENT
                  (XITIMES X (IPLUS X (XCASE CX NIL 1 -1 1)))
                  2))))
    (T 0) ) ) ) )
)

(DE SFIN2 (DMX DN X Y)
 (SETQ X (VGTSUINX DN X))
 (SETQ Y (VGTSUINX DN Y))
 (COND ((ILESSP X Y) (EXCHANGE X Y)))
 (XIDIFFERENCE DMX (XIPLUS Y (XIQUOTIENT (XITIMES X (IADD1 X)) 2))) )

(DE AFIN2 (DMX DN X Y) (PROG ()
  (SETQ X (VGTSUINX DN X))
  (SETQ Y (VGTSUINX DN Y))
  (COND
   ((IEQN X Y) (SETQ !$NOTASY0 NIL) (RETURN 0))
   ((ILESSP X Y) (EXCHANGE X Y) (SETQ MINSW (NOT MINSW))) )
  (RETURN
   (XIDIFFERENCE DMX
    (XIPLUS Y (XIQUOTIENT (XITIMES X (ISUB1 X)) 2)) ) ) ))

(DE HFIN2 (DMX DN X Y)
 (SETQ X (VGTSUINX DN X))
 (SETQ Y (VGTSUINX DN Y))
 (COND
  ((IEQN X Y) (SETQ HDIAG!* T))
  ((ILESSP X Y) (EXCHANGE X Y) (SETQ CONJSW (NOT CONJSW))) )
 (XIDIFFERENCE DMX (XIPLUS Y (XIQUOTIENT (XITIMES X (IADD1 X)) 2))) )



(DE TENLIMB (XM LIML) (XLIM XM (CDR LIML)))

(DE XLIM (XM LIML) 
 (COND 
  ((NULL LIML) (LIST XM))
  (T (XCASE (CAR LIML) (UNSYMER (CAR LIML)) 
      (DLIM XM (CDR LIML))
      (KLIM XM (CDR LIML) -4 NIL)
      (KLIM XM (CDR LIML) NIL NIL)
      (KLIM XM (CDR LIML) -4 T)
      (DLIMN XM (CADR LIML) (CDDR LIML))
      (KLIM2 XM (CDR LIML) T NIL)
      (KLIM2 XM (CDR LIML) NIL NIL)
      (KLIM2 XM (CDR LIML) T T) )) ) )

(DE DLIMN (XM DMX CNT)
 (SLET (XL)
  (XIINCR DMX)
  (REPEATS CNT
   (XPUSH XL (XIREMAINDER XM DMX))
   (SETQ XM (XIQUOTIENT XM DMX)) )
  XL) )

(DE DLIM1 (LIML) 
 (SLET ((XL (COND ((CDR LIML) (DLIM1 (CDR LIML)))))
        (XD (XIADD1 (CAAR LIML))) )
  (SLET ((XR (XIREMAINDER XXND!$ XD)))
   (SETQ XXND!$ (XIQUOTIENT XXND!$ XD))
  (NCONC (XLIM XR (CDAR LIML)) XL) ) ) )

(DE DLIM (XXND!$ LIML) (DLIM1 LIML))

(DE KLIM (VX LIML SSW HSW)
% If SSW then N=dim and Y=dim-current index
%  else N=dim-2 and Y=dim-1-current index value
% Q= rank-1
 (SLET ((Y (CAR LIML)) (N (COND (SSW (SETQ Y (IADD1 Y))) (T (ISUB1 Y))))
        (Q (LENGTH (UNPUSH LIML))) NVX)
  (FOREACH TL IN LIML CONC
   (IDECR Q)
   (WHILE
    (NOT (XIMINUSP (SETQ NVX (XIDIFFERENCE VX (SXIOVERW Y Q SSW)))))
    (SETQ VX NVX) (IDECR Y) )
   (COND 
    ((NOT SSW) (IDECR Y))   % Anti symmetric case
    ((NOT (IEQN Y SSW)) (SETQ SSW Y))
    (HSW (SETQ HDIAG!* T))
    (T (SETQ SDIAG!* T)) )
   (XLIM (IDIFFERENCE N Y) TL) ) ) )

(DE KLIM2 (XM LIML SS HS) 
 (SLET ((XNX (CAR LIML)) X)
  (COND (SS (IINCR XNX)))
  (WHILE (NOT (XIMINUSP (SETQ X (XIDIFFERENCE XM XNX))))
         (SETQ XM X) (IDECR XNX) )
  (SETQ XNX (IDIFFERENCE (CAR LIML) XNX))
  (COND
   ((NOT SS) (IINCR XM))
   (T (XIINCR XNX) (COND ((IZEROP XM) (COND
         (HS (SETQ HDIAG!* T))
         (T (SETQ SDIAG!* T)) )))) )
  (NCONC (XLIM XNX (CADR LIML)) (XLIM (IPLUS XM XNX) (CADDR LIML))) ) )

(DE RANKERR NIL 
 (ERRORFT 540 "Number of indices inconsistent with rank") )

(DE UNSYMER (X) (ERRORFT 545 "%r is not a known symmetry type" X))



(DE RANGE (TN LWLIM HILIM)
 (SLET ((LIML (CAR (GETBLO TN))))
  (SETQ SUBIX!$ NIL)
  (SETQ LWLIM (COND
    (LWLIM
     (SETQ LWLIM (TENCFIN LWLIM LIML))
     (COND (SUBIX!$ (XIADD1 LWLIM)) (T LWLIM)) )
    (T 0) ))
  (SETQ SUBIX!$ NIL)                    % Not-memory-element switch off.
  (SETQ HILIM (COND
    (HILIM (TENCFIN HILIM LIML))
    (T (CAR LIML)) ))
  (COND
   ((XGREATERP LWLIM HILIM)
    (ERRORFT 546
     "Elements in the specified range are only defined by symmetry") ) )
  (CONS TN (CONS LWLIM HILIM)) ) )

(DE ONEELM (TN XL) (SLET (MINSW CONJSW HDIAG!* A)
 (SETQ SUBIX!$ NIL)
 (SETQ A (TENCFIN XL (CAR (GETBLO TN))))
 (CONS TN (COND
   (SUBIX!$
    (SETQ A (TENFIND XL (CAR (GETBLO TN))))
    (CONSA A MINSW (AND CONJSW (NOT HDIAG!*)) XL) )
   (T A) )) ))
  
(DE TENCFIN (XL LIML) (PROGN
 (SETQ FINXL!$ XL)
 (SETQ !$NOTASY0 T)
 (COND (LIML (SETQ LIML (GTSUCX (CAR LIML) (CDR LIML)))))
 (COND
  (FINXL!$ (RANKERR))
  (T LIML) ) ))

(DE GTSUCX (DMX LIML) 
 (COND 
  ((NULL LIML) (SETQ LIML (XFIN DMX)) (COND (SUBIX!$ DMX) (T LIML)))
  (T (XCASE (CAR LIML) (UNSYMER (CAR LIML)) 
      (DCXS (CDR LIML))
      (SCXS DMX (CDR LIML))
      (ACXS DMX (CDR LIML))
      (SCXS DMX (CDR LIML))
      (DCXSN (CADR LIML) (CDDR LIML))
      (SCXS DMX (CDR LIML))
      (ACXS DMX (CDR LIML))
      (SCXS DMX (CDR LIML))
%     (SCXS2 DMX (CADR LIML) (CADDR LIML) (CADDDR LIML))
%     (ACXS2 DMX (CADR LIML) (CADDR LIML) (CADDDR LIML))
%     (SCXS2 DMX (CADR LIML) (CADDR LIML) (CADDDR LIML))
      )) ) )

(DE DCXS (LIML) 
 (SLET ((VX 0))
  (FOREACH X IN LIML DO
   (SETQ VX
    (XIPLUS (XITIMES VX (XIADD1 (CAR X))) (GTSUCX (CAR X) (CDR X))) ) )
  VX) )

(DE DCXSN (DMX CNT)
% Get index for CNT non-symmetric indices with dimension DMX.
 (SLET ((VX 0) (DMY (XIADD1 DMX)))
  (REPEATS CNT (SETQ VX (XIPLUS (XITIMES VX DMY) (GTSUCX DMX NIL))))
  VX) )

(DE VGTSUCX (DMX X) (XIDIFFERENCE DMX (GTSUCX DMX X)))

(DE SCXS (DMX LIML) 
 (SLET (XL (DN (CAR LIML)) A)
  (FOREACH X IN (CDR LIML) DO
   (SETQ A SUBIX!$)
   (XPUSH XL (VGTSUCX DN X))
   (COND ((AND (NOT A) (XSSUB XL)) (SETQ SUBIX!$ T))) )
  (SYFINE DMX XL T) ) )

(DE ACXS (DMX LIML) 
 (SLET (XL (DN (CAR LIML)) (DNX (LENGTH (CDDR LIML))) A)
  (FOREACH X IN (CDR LIML) DO
   (SETQ A SUBIX!$)
   (XPUSH XL (COND
     ((XLESSP (SETQ X (VGTSUCX DN X)) DNX)
      (SETQ SUBIX!$ (SETQ A T))
      DNX )
     (T X) ))
   (COND ((AND (NOT A) (XASUB XL DNX)) (SETQ SUBIX!$ T)))
   (XIDECR DNX) )
  (SYFINE DMX XL NIL) ) )

(DE XSSUB (XL)
 (COND ((AND (CDR XL) (XGREATERP (CAR XL) (CADR XL)))
        (XSSUB (RPLACA (CDR XL) (XIADD1 (CADR XL))))
        (RPLACA XL 0) )) )

(DE XASUB (XL DN)
 (COND ((AND (CDR XL) (NOT (XLESSP (CAR XL) (CADR XL))))
        (XASUB (RPLACA (CDR XL) (XIADD1 (CADR XL))) (XIADD1 DN))
        (RPLACA XL DN) )) )

(COMMENT
(DE SCXS2 (DMX DN X Y)
 (SETQ X (VGTSUCX DN X))
 (SETQ Y (VGTSUCX DN Y))
 (COND ((XLESSP X Y) (SETQ SUBIX!$ (SETQ Y 0)) (XIINCR X)))
 (XIDIFFERENCE DMX (XIPLUS Y (XIQUOTIENT (XITIMES X (XIADD1 X)) 2))) )

(DE ACXS2 (DMX DN X Y)
 (COND
  ((AZEROP (SETQ X (VGTSUCX DN X))) (SETQ SUBIX!$ DMX))
  (T (SETQ Y (VGTSUCX DN Y))
     (COND ((NOT (XGREATERP X Y)) (SETQ SUBIX!$ (SETQ Y 0)) (XIINCR X)))
     (XIDIFFERENCE DMX
      (XIPLUS Y (XIQUOTIENT (XITIMES X (XISUB1 X)) 2)) ) ) ) )
)                                       % End of comment.

%### KARLIM.SL end
