%### SDET.SL

% Sheep determinant routines. Formerly called DET.

(FUNCTIONS MAPPERM DETMINORSG DETMINOR DET DETDM RIDSEQ COFACDM INVTEN
           INVTENDM)

% Name of a rank 2 tensor or a function of two args.
(SETFLUIDS TSRFN!$)

(LOCALFUNCTIONS MAPPERM)

(DE MAPPERM (INDL1 INDL2 PARITY TERM) 
% INDL1: list of column indices, INDL2: list of row indices,
% PARITY: sign (boolean),
% TERM: list of expressions to multiply minor with.
% Adds up the terms of a minor.
 (PROG (MINSW ELM)
% Combine each row index with first column index.
  (FOREACH INX2 IN INDL2 DO
   (SETQ MINSW PARITY) 
   (SETQ ELM (COND 
     ((TENP TSRFN!$) (XGETEN TSRFN!$ (LIST (CAR INDL1) INX2) NIL))
     (T (FUNCALL TSRFN!$ (CAR INDL1) INX2)) ))
   (COND 
    ((NOT (AZEROP ELM)) % If not zero then if
     (COND
      ((CDR INDL1)      % more column indices then apply MAPPERM on them
       (MAPPERM (CDR INDL1) (DELETE INX2 INDL2) MINSW (CONS ELM TERM)) )
                        % else multiply together and add up.
       (T (SMADDTERM (SMMULTC (CONS ELM TERM) 1 MINSW T))) )) )
   (SETQ PARITY (NOT PARITY)) ) ) )

(DE DETMINORSG (TSRFN!$ INDL1 INDL2 SGN TERM)
% TSRFN!$ (fluid): Name of a rank 2 tensor or a function of two
% arguments,
% INDL1: list of column indices, INDL2: list of row indices,
% SGN: sign (boolean), TERM: NIL or list of expressions to multiply
% minor with. Forms the minor with column indices INDL1 and row
% indices INDL2 from TSRFN!$.
% Each term is multiplied with the expressions in TERM and if SGN is
% true the sign is switched.
 (COND ((NOT (EQ (LENGTH INDL1) (LENGTH INDL2)))
        (ERRORFT 530 "Index list %r is not as long as %r in DETMINOR"
                INDL1 INDL2) ))
 (SHCOLLECT (MAPPERM INDL1 INDL2 SGN TERM)) )

(DE DETMINOR (TSR INDL1 INDL2)  % Form the determinant of a minor.
% TSR: Name of a tensor or a function of two arguments,
% INDL1: list of column indices, INDL2: list of row indices.
 (DETMINORSG TSR INDL1 INDL2 NIL NIL))

% Form the determinant of second rank tensor TSR.
% Dimension is taken form first index of TSR.
(DE DET (TSR) (DETDM TSR (GETDIM1 TSR)))

% Form the determinant.
% TSR: Name of a tensor or a function of two arguments, D: Dimension.
(DE DETDM (TSR D) (DETMINOR TSR (SEQ0L D) (SEQ0L D)))

(DE RIDSEQ (RIDIX D) (SLET (LS)          % Form a list of index values
  (UNTIL (IZEROP D)                      % from 0 to (but not including)
   (COND ((NOT (IEQN (IDECR D) RIDIX))   % D, with RIDIX excluded.
          (XPUSH LS D) )) )
  LS))

% The cofactor of element (TSR INDX1 INDX2).
(DE COFACDM (TSR INDX1 INDX2 D)
 (DETMINORSG TSR (RIDSEQ INDX1 D) (RIDSEQ INDX2 D)
  (NOT (XEVENP (XPLUS INDX1 INDX2))) NIL) )     % JEA.

(DE TINVTEN (TSR INDX1 INDX2 DETINV)
    (INVTENDM TSR INDX1 INDX2 (GETSCL DETINV) (GETDIM1 TSR)) )

(DE INVTEN (TSR INDX1 INDX2 DETINV)
 (INVTENDM TSR INDX1 INDX2 DETINV (GETDIM1 TSR)) )

(DE INVTENDM (TSR INDX1 INDX2 DETINV D)
 (DETMINORSG TSR (RIDSEQ INDX2 D) (RIDSEQ INDX1 D)
  (NOT (XEVENP (XPLUS INDX1 INDX2))) (LIST DETINV) ) )  % JEA.

% To construct inverse of a scalar but use symbolic if it is a sum.
% Written by JEA with modification by IF.
(DE MKINVDET (DETNAM)
 (SLET ((DETVAL (EGETEN DETNAM NIL NIL)))
  (SMEXPT
   (COND
    ((OR NODGNV (NOT (EQCAR DETVAL 'PLUS))) (COPYL DETVAL))
    (T
     (PUSHQ TNKEEP!* DETNAM)
     (PRTLF
      "%v is a sum. It is replaced by the symbol %v with prt name '%w'"
      DETNAM DETNAM DETNAM )
     DETNAM ) )
   -1 ) ) )

%### SDET.SL end
