%### EXFN2I.SL ======================================================
%
%       EXFN2I.SL
%       ---------
%
%       Copyright Inge Frick, Institute of Theoretical Physics,
%       Vanadisvagen 9, S-113 46 Stockholm, Sweden
%
%
%       This file contains definitions of some extensions to Standard
%       Lisp. It is used together with EXFN1I.SL, EXINTI.SL and
%       EXCOMI.SL. EXFN1I.SL + EXCOMI.SL + EXFN2I.SL (in that order)
%       is used to define the extensions for the compiler. EXFN1I.SL +
%       EXFN2I.SL + EXINTI.SL (in that order) is used to define the
%       extensions to the interpreter.
%
%       Some functions here are system dependent and have not been
%       defined. Look for the indicator MUSTDEFINE.
%       They are: PLIST, FOREACHID, TIME, UNREADCH, ORDERP, LEXP,
%                 BELL, RESET, QUIT, BYE
%
%====================================================================

(FUNCTIONS GCD GCD1 INCR DECR EVENP
           XPLUS XTIMES XADD1 XSUB1 XMINUS XDIFFERENCE
           XQUOTIENT XREMAINDER XEQN XLESSP XGREATERP XMINUSP
           XINCR XDECR XZEROP XONEP
           XIPLUS XITIMES XIADD1 XISUB1 XIMINUS XIDIFFERENCE
           XIQUOTIENT XIREMAINDER XIEQN XILESSP XIGREATERP XIMINUSP
           XIINCR XIDECR XIZEROP XIONEP
           XIMEMQ XIATSOC XIEQN
           AMEMQ AATSOC AEQN AZEROP AONEP
           IPLUS ITIMES IADD1 ISUB1 IMINUS IDIFFERENCE
           IQUOTIENT IREMAINDER IEQN ILESSP IGREATERP IMINUSP
           IINCR IDECR IZEROP IONEP
           RPLACW COPY COPYT PNTH NTH CNDR DELQIP RATSOC APPENDL NCONCL
           INTERSECTIONQ SETDIFFQ ADJOINQ PUSHQ POP REMPOP
           LASTPAIR XMEMQ XMEMQL MEM2Q LAST2PAIR SUBLA
           FLAG1 REMFLAG1 PLIST MAPOBL
           GETFN GETFNTYPE FNP EXPRP MACROP FEXPRP
           ERRSETT FEXPRCALL XOR PROGQ EXCHANGE
           FORL XIFORL IFORL FORX XIFORX IFORX HPFORX
           NOEVAGAINP % PSETQ DO DO!*
           INTERNP EXPLODEC EXPLODECHELP ULCASE OSCASE ILISPCASE
           MAKEID MAKESTR MAKEINT MAKEFLO
           LIST2VECTOR VECTOR VECTOR2LIST
           LISTP ORDERP LEXP BATCHP STDINP STDOUTP TTYINP FILEP
           PRIN2D PRIN1D
           UNREADCH SKIPWHITE SKIPTO PRIN2T TTAB SPACES QTRPRI BELL
           PGLINE CHANIN CHANOUT IOCHANS
           STDIN STDOUT STDINOUT FILEIN FILEOUT
           WARNING SETPROMPTER PROMPTER PROMPT SETFOLDCASE
           SETGCMESSAGE GCMESSAGE
           SYESP SNOP IYESP READIN DFEVAL DFPRINT DFEVALQ
           RESET QUIT BYE TIME)

(LOCALFUNCTIONS GCD1 EXPLODECHELP % HPFORX IOCHANS
                DFEVALQ DFPRINT)

(USEDPROPERTIES DIGITVALUE)

(USEDFLAGS DFEXPAND EVAL IGNORE)

(SETGLOBALS XCASE!*!$ARG SFSTKS!*)
(SETGLOBALS TAB!*)      % Must be defined in system dependent part.

% The following three constants are intended to be used in ULCASE to
% convert case to that used by:
(SETCONSTANTS
 (!*OSCASEUP T)      % the operating system for system commands,
                     % file names etc.
 (!*ILISPCASEUP T)   % the underlaying Lisp system.
 (!*SHPCASEUP T) )   % Sheep Lisp. This is the case used in FOLDCASE.
% Note that these are constants. They can be read but if they are
% changed errors might occur.

(SETFLUIDS 
 (FILELINELENGTH!* 76)
 (ASKREPEAT!*   5)
 (PROMPTSTRING!* "SL> ")
 (!*MSG         T)
 ((SUPPRESSANS!* (FOLDCASE NIL (COMPRESS (EXPLODEC "Finished")))))
 (!*SHPDEBUG T)
 !*INT !*SUPPRESSANS DFPRINT!* DFEVAL!* CLOC!* !*ECHO !*LOADMSG FIL!*
 YESPDEFAULT!* NOPDEFAULT!*)

(LINELENGTH 76)

% The following functions with prefix I and A operate on integers with
% absolute value < 1000. On many systems this means EQ can be used.
% Those with prefix A will also handle any argument but only guarantee
% to test equality when one argument is a small integer. Those with
% prefix I might give anything when an argument is not a small integer.
% Use I when both arguments are known to be small.
% Use A when one argument is known to be small.

(SAMEDEF AMEMQ MEMBER)

(SAMEDEF AATSOC ASSOC)

(SAMEDEF AEQN EQUAL)

(DS AZEROP (X) (AEQN X 0))                      % Any argument

(DS AONEP (X) (AEQN X 1))

(SAMEDEF IPLUS PLUS2)

(SAMEDEF ITIMES TIMES2)

(SAMEDEF IADD1 ADD1)

(SAMEDEF ISUB1 SUB1)

(SAMEDEF IMINUS MINUS)

(SAMEDEF IDIFFERENCE DIFFERENCE)

(SAMEDEF IQUOTIENT QUOTIENT)

(SAMEDEF IREMAINDER REMAINDER)

(SAMEDEF IEQN EQN)

(SAMEDEF ILESSP LESSP)

(SAMEDEF IGREATERP GREATERP)

(SAMEDEF IMINUSP MINUSP)

(DS IINCR (X) (SETQ X (IADD1 X)))

(DS IDECR (X) (SETQ X (ISUB1 X)))

(DS IEVENP (X) (AZEROP (IREMAINDER X 2)))

(SAMEDEF IZEROP AZEROP)

(SAMEDEF IONEP AONEP)

% The following arithmetic functions should be used when operating on
% integers that are sufficiently small (e.g. one word) that the
% operations can be open coded by a compiler. Numbers less than
% 2^30 are ok.

(SAMEDEF XPLUS PLUS2)

(SAMEDEF XTIMES TIMES2)

(SAMEDEF XADD1 ADD1)

(SAMEDEF XSUB1 SUB1)

(SAMEDEF XMINUS MINUS)

(SAMEDEF XDIFFERENCE DIFFERENCE)

(SAMEDEF XQUOTIENT QUOTIENT)

(SAMEDEF XREMAINDER REMAINDER)

(SAMEDEF XEQN EQN)

(SAMEDEF XLESSP LESSP)

(SAMEDEF XGREATERP GREATERP)

(SAMEDEF XMINUSP MINUSP)

(DS XINCR (X) (SETQ X (XADD1 X)))

(DS XDECR (X) (SETQ X (XSUB1 X)))

(DS XEVENP (X) (AZEROP (XREMAINDER X 2)))

(SAMEDEF XZEROP AZEROP)

(SAMEDEF XONEP AONEP)

% The following functions with prefix XI and XA operate on integers
% with absolute value < 64000. On some systems this means EQ can be
% used. Those with prefix XA will also handle any argument but only
% guarantee to test equality when one argument is a small integer.
% Those with prefix I might give anything when an argument is not a
% small integer.
% Use XI when both arguments are known to be small.
% Use XA when one argument is known to be small.

(SAMEDEF XAMEMQ MEMBER)

(SAMEDEF XAATSOC ASSOC)

(SAMEDEF XAEQN EQUAL)

(SAMEDEF XIPLUS PLUS2)

(SAMEDEF XITIMES TIMES2)

(SAMEDEF XIADD1 ADD1)

(SAMEDEF XISUB1 SUB1)

(SAMEDEF XIMINUS MINUS)

(SAMEDEF XIDIFFERENCE DIFFERENCE)

(SAMEDEF XIQUOTIENT QUOTIENT)

(SAMEDEF XIREMAINDER REMAINDER)

(SAMEDEF XIEQN EQN)

(SAMEDEF XILESSP LESSP)

(SAMEDEF XIGREATERP GREATERP)

(SAMEDEF XIMINUSP MINUSP)

(DS XIINCR (X) (SETQ X (XIADD1 X)))

(DS XIDECR (X) (SETQ X (XISUB1 X)))

(DS XIEVENP (X) (AZEROP (XIREMAINDER X 2)))

(SAMEDEF XIZEROP AZEROP)

(SAMEDEF XIONEP AONEP)

(DS INCR (X) (SETQ X (ADD1 X)))

(DS DECR (X) (SETQ X (SUB1 X)))

(DS EVENP (X) (AZEROP (REMAINDER X 2)))

% GCD(N:integer,M:integer):integer
% Returns the greatest common divisor of N and M.
(DE GCD (X Y) (COND
  ((NOT (AND (FIXP X) (FIXP Y)))
   (ERRORFT 415 "%r %r illegal args to GCD" X Y) )
  ((LESSP (SETQ X (ABS X)) (SETQ Y (ABS Y))) (GCD1 Y X))
  (T (GCD1 X Y)) ) )

(IFP (NOT (FLAGP 'GCD 'LOSE))
 (DE GCD1 (Y R) (COND
   ((AZEROP R) Y)
   (T (GCD1 R (REMAINDER Y R))) )) )


%(DE RPLACW (X Y) (RPLACD (RPLACA X (CAR Y)) (CDR Y)))
(DE RPLACW (X Y) (RPLAD X (CAR Y) (CDR Y)))

(DE COPY (X)
 (COND
  ((PAIRP X) (CONS (COPY (CAR X)) (COPY (CDR X))))
  (T X) ) )

(DS COPYT (X) (APPEND X NIL))

(DE PNTH (L N) (COND
  ((XIGREATERP N 1) (PNTH (CDR L) (XISUB1 N)))
  (T L) ))

(DS NTH (L N) (CAR (PNTH L N)))

% Similar to PNTH. Returns NIL if second argument is negative or to
% large.
(DE CNDR (U I)
 (COND
  ((MINUSP I) NIL)
  (T (WHILE (AND U (NOT (XIZEROP I))) (XIDECR I) (UNPUSH U)) U) ) )

(DE DELQIP (X L)
 (COND
  ((NULL L) NIL)
  ((EQ (CAR L) X) (DELQIP X (CDR L)))
  (T (RPLACD L (DELQIP X (CDR L)))) ) )

(DE RATSOC (X AL) (COND
 ((NULL AL) NIL)
 ((EQ X (CDAR AL)) (CAR AL))
 (T (RATSOC X (CDR AL))) ))

(DE APPENDL (L) (COND (L (APPEND (CAR L) (APPENDL (CDR L))))))

(DE NCONCL (L) (COND (L (NCONC (CAR L) (NCONCL (CDR L))))))

% UNIONQ moved to from exfn1i. JEA.

(DE INTERSECTIONQ (X Y)
 (COND
  ((NULL X) NIL)
  ((MEMQ (CAR X) Y) (CONS (CAR X) (INTERSECTIONQ (CDR X) Y)))
  (T (INTERSECTIONQ (CDR X) Y)) ) )

(DE SETDIFFQ (X Y)      % Returns list of elements in X not EQ any in Y.
 (COND
  ((NULL X) NIL)
  ((MEMQ (CAR X) Y) (SETDIFFQ (CDR X) Y))
  (T (CONS (CAR X) (SETDIFFQ (CDR X) Y))) ) )

(DE ADJOINQ (X L)               % Put X on list L if it isn't there.
 (COND ((MEMQ X L) L) (T (CONS X L))) )

(DS PUSHQ (L X) (SETQ L (ADJOINQ X L)))

(DS POP (L) (PROG1 (CAR L) (SETQ L (CDR L))))

(DS REMPOP (L X) (SETQ L (DELQIP X L)))

(DE LASTPAIR (X)                % Returns last pair in list X.
 (WHILE (CDR X) (SETQ X (CDR X)))
 X)

(DE XMEMQ (X Y)         % True if X is an element of Y at any level.
 (COND
  ((NOT (PAIRP Y)) (EQ X Y))
  (T (OR (XMEMQ X (CAR Y)) (XMEMQ X (CDR Y)))) ) )

(DE XMEMQL (L Y) % True if any element of list L is an element of Y at
 (COND          % any level. In that case the tail of L starting with
  ((NOT (PAIRP Y)) (MEMQ Y L))  % the found element is returned.
  (T (OR (XMEMQL L (CAR Y)) (XMEMQL L (CDR Y)))) ) )

(DE MEM2Q (X L)
% Same as MEMQ but returns pointer one before matching element.
 (COND (L (WHILE (AND (CDR L) (NOT (EQ (CADR L) X))) (SETQ L (CDR L)))
          (COND ((CDR L) L)) )) )

(DE LAST2PAIR (X)       % Returns next but last pair in list X.
 (COND
  ((NOT (AND X (CDR X) (CDDR X))) X)
  (T (SLET ((Y (CDR X)))
           (WHILE (CDR Y) (SETQ Y (CDR (SETQ X Y))))
           X )) ) )

% The same as SUBLIS except for using EQ instead of EQUAL on the masks
% in U.
(DE SUBLA (U V)
 (SLET (A B)
  (COND
   ((NOT (AND U V)) V)
   ((PAIRP V)
    (SETQ A (SUBLA U (CAR V)))
    (SETQ B (SUBLA U (CDR V)))
    (COND
     ((AND (EQ A (CAR V)) (EQ B (CDR V))) V)
     (T (CONS A B)) ) )
   ((SETQ U (ATSOC V U)) (CDR U))
   (T V) ) ) )

(DS FLAG1 (X Y) (FLAG (LIST X) Y))

(DS REMFLAG1 (X Y) (REMFLAG (LIST X) Y))

% PLIST(X:id):property list
% Returns the property list in some format. This is only for
% inspection so the exact format returned is not so important.
%(DE PLIST (X) (MUSTDEFINE PLIST "EXFN.SL"))

% FOREACHID(VAR:id,[FORM:ANY]):undefined
% VAR is in turn bound to all interned identifiers and all FORMs are
% evaluated with that binding of VAR.
% How this is done is system dependent. It could be something like:
%(DS FOREACHID (X !&REST L) (FOREACH X IN (OBLIST) DO . L))
%MUSTDEFINE FOREACHID

% IGET is a version of GET that always return NIL when the first
% argument not is an identifier.
(SAMEDEF IGET GET)

% GETFN(ARG:any):{NIL,codepointer,LAMBDA-expression}
% If ARG is a function then the definition is returned, else NIL.
(DE GETFN (X)
 (SLET ((A (GETD X))) (COND (A (CDR A)))) )

% GETFNTYPE(ARG:any):{NIL,EXPR,FEXPR,MACRO}
% If ARG is a function, return the function type, else NIL.
(DE GETFNTYPE (X)
 (SLET ((A (GETD X))) (COND (A (CAR A)))) )

% FNP(ARG:any):extraboolean
% FNP returns NIL if ARG is not a function, else true.
(SAMEDEF FNP GETD)

(DS EXPRP (X) (EQCAR (GETD X) 'EXPR))

(DS MACROP (X) (EQCAR (GETD X) 'MACRO))

(DS FEXPRP (X) (EQCAR (GETD X) 'FEXPR))

% ERRSETT([ARG:form]):any
% ERRSETT evaluates expressions ARG in an ERRSET. Error messages are 
% turned on if not a surrounding ERRSET has turned them off.
(DMACRO ERRSETT (!&REST U) (LIST 'ERRSET (MKPROGN U) '!*ERRMSG))

% CAPTURESFSTKS():alist
% CAPTURESFSTKS return an alist of form
% ((stack . value) (stack . value) ...) for those status stacks that
% have a value.
(DE CAPTURESFSTKS () (SLET (A)
  (FOREACH X IN SFSTKS!* CONC
   (COND ((SETQ A (EVAL X)) (LIST (CONS X A)))) ) ))

% UNBINDSFSTKS(OSTKS:alist,EV:any):EV
% If EV is not a pair then UNBINDSFSTKS restores the status stacks to
% the value earlier saved in OSTKS by CAPTURESFSTKS.
(DE UNBINDSFSTKS (OSTKS EV) (SLET (SV OSV A)
 (COND ((NOT (PAIRP EV)) (FOREACH X IN SFSTKS!* DO
         (COND ((AND (SETQ SV (EVAL X))
                 (NOT (AND (SETQ OSV (ATSOC X OSTKS))
                           (EQ SV (UNPUSH OSV)) )) )
                (SET X OSV)
                (SETQ A SV) (WHILE (NOT (EQ (CDR A) OSV)) (UNPUSH A))
                (SETQ OSV (GET X 'SFSTKFN))
                (COND
                 ((FLAGP X 'LASTONLY) (FUNCALL OSV (CAR A)))
                 (T (RPLACD A NIL)
                    (FOREACH Y IN SV DO (FUNCALL OSV Y)) ) ) )) )))
 EV))

% Apply a FEXPR.
(DS FEXPRCALL (FN ARGS) (APPLY (GETFN FN) (LIST ARGS)))

(DMACRO XOR (!&REST L)
  (FCHKARG L 2 NIL 'XOR)
  (CONS 'COND
    (FOREACH X ON L COLLECT
      (COND
        ((NULL (CDR X)) (LIST (CAR X) T)) % Bugfix for some Lisps. JEA.
        (T (LIST (CAR X)
                 (LIST 'NULL
%:LH               (COND ((CDDR X) (CONS 'OR (CDR X)))
%:LH                     (T (CADR X)) )
                   (CONSIFCDR 'OR (CDR X))  ))) ))) )

% PROGQ(PRED:form,[XPR:form]):extraboolean
% Type: NOEVAL, NOSPREAD
% Evaluate PRED. If the result is NIL return NIL, else evaluate the
% XPRs from left to right and return true.
(DMACRO PROGQ (!&REST L) (LIST 'COND (APPEND1 L T)))  %:LH

% EXCHANGE(VARX:id,VARY:id):any
% The values of VARX and VARY are exchanged.
(DS EXCHANGE (X Y) (SETQ Y (PROG1 X (SETQ X Y))))

% FORL(VAR:id,LIMIT:{id,list},[XPR:{form,label}]):NIL
% FORL does nested loops over numerical indices. It is used to
% manipulate arrays. LVARS is a list of LVARSPECS.
% <FORLform>    :== (<var> <limitform> . <progbody>)
%
% <limitform>   :== <endlim> | (<endlim>) | (<startlim> <endlim>) |
%                   (<startlim> <endlim> <step>)
%
% Default for <startlim> is 0 and for <step> is 1.
% <step> is not evaluated, it must be an integer, but it might be
% negative.
% <startlim> and <endlim> is only evaluated once.
% If <endlim> not is an atom then it must be given inside parenthesis.
%
% Evaluation of a FORL is done in the following way:
% 1 Set <var> to the value of <startlim>.
% 2 Go to 5.
% 3 Evaluate <progbody> as if it was the body of a PROG.
% 4 Increment <var> by adding <step> to it and go to 2
% 5 If <var> is less than (or if <step> is negative, greater than)
%   the value of <endlim> then goto 3.
% 6 Return NIL.

(DMACRO FORL (VAR LIM !&REST BODY)
 (HPFORX VAR LIM BODY NIL '(XGREATERP XLESSP XINCR XDECR XPLUS)) )

(DMACRO XIFORL (VAR LIM !&REST BODY)
 (HPFORX VAR LIM BODY NIL '(XIGREATERP XILESSP XIINCR XIDECR XIPLUS)) )

(DMACRO IFORL (VAR LIM !&REST BODY)
 (HPFORX VAR LIM BODY NIL '(IGREATERP ILESSP IINCR IDECR IPLUS)) )

% FORX is the same as FORL as described above except for that step 5
% is replaced by:
% 5 If <lvar> is not greater than (or if <step> is negative, not less
%   than) the value of <endlim> then go 3.
% This means that the limit is included for FORX but excluded for FORL.
(DMACRO FORX (VAR LIM !&REST BODY)
 (HPFORX VAR LIM BODY T '(XLESSP XGREATERP XINCR XDECR XPLUS)) )

(DMACRO XIFORX (VAR LIM !&REST BODY)
 (HPFORX VAR LIM BODY T '(XILESSP XIGREATERP XIINCR XIDECR XIPLUS)) )

(DMACRO IFORX (VAR LIM !&REST BODY)
 (HPFORX VAR LIM BODY T '(ILESSP IGREATERP IINCR IDECR IPLUS)) )

(DE HPFORX (VAR LIM BODY SW FL)
 (SLET ((STRT 0) (STEP 1) (LAB1 (GENSYM)) (LAB (GENSYM)) (FP FL)
        (VARL (LIST VAR)) LMN)
  (SETQ BODY (CONSA (LIST 'GO LAB1) LAB BODY))
  (COND ((PAIRP LIM) (SETQ LIM (COND
           ((CDR LIM)
            (SETQ STRT (CAR LIM))
            (COND ((CDDR LIM) (SETQ STEP (CADDR LIM))))
            (CADR LIM) )
           (T (CAR LIM)) ))))
  (COND ((AND (PAIRP LIM) (NOT (EQ (CAR LIM) 'QUOTE)))
         (XPUSH BODY (LIST 'SETQ (SETQ LMN (GENSYM)) LIM))
         (XPUSH VARL (SETQ LIM LMN)) ))
  (COND ((MINUSP STEP) (UNPUSH FP)))
  (SETQ LIM (LIST (CAR FP) VAR LIM))
  (SETQ FL (CDDR FL))
  (SETQ BODY (APPEND BODY (LIST
     (COND
      ((IONEP STEP) (LIST (CAR FL) VAR))
      ((IEQN STEP -1) (LIST (CADR FL) VAR))
      (T (LIST 'SETQ VAR (LIST (CADDR FL) VAR STEP))) )
     LAB1
     (LIST 'COND
      (LIST (COND (SW LIM) (T (LIST 'NOT LIM))) (LIST 'GO LAB)) ) )))
  (CONSA 'PROG VARL (LIST 'SETQ VAR STRT) BODY) ) )

% NOEVAGAINP is NIL if it doesn't matter that X is evaluated more than
% once.
(DE NOEVAGAINP (X) (NOT (COND
  ((PAIRP X) (EQ (CAR X) 'QUOTE))
  (T (OR (NULL X) (EQ X T) (CONSTANTP X))) )))

(COMMENT %PSETQ, DO and DO!* removed
% PSETQ(VAR:id,ARG:any,[VAR:id,ARG:any])
% First evaluates all ARGs, then sets each VAR to the value of its ARG.
% Note that the number of arguments must be even.
% This is the same as PSETQ in Lisp-machine Lisp.
(DMACRO PSETQ (!&REST L)
  (SLET ((A L) VR LVRS LARGS BODY)
   (WHILE (SETQ A (CDDR A))
    (COND
     ((NULL (CDR A))
      (ERRORFT 412 "Uneven number of arguments to PSETQ") ) )
    (XPUSH LVRS (SETQ VR (GENSYM)))
    (XPUSH LARGS (CADR A))
    (XPUSH BODY (LIST 'SETQ (CAR A) VR)) )
   (SETQ A (LIST 'SETQ (CAR L) (CADR L)))
   (COND
    ((CDDR L)
     (CONS (CONSA 'LAMBDA (REVERSIP LVRS) A (REVERSIP BODY))
           (REVERSIP LARGS) ) )
    (T A) ) ) )

% The MacLisp DO
(DMACRO DO (VARS TST !&REST BODY) (SLET (LAB PVRS LVRS LARGS REP GOS)
  (FOREACH X IN (REVERSE VARS) DO
   (COND
    ((ATOM X) (XPUSH PVRS X))
    ((NULL (CDR X)) (XPUSH PVRS (CAR X)))
    (T (COND ((CDDR X) (XPUSH REP (CADDR X)) (XPUSH REP (CAR X))))
       (COND
        ((CADR X) (XPUSH LVRS (CAR X)) (XPUSH LARGS (CADR X)))
        (T (XPUSH PVRS (CAR X))) ) ) ) )
  (COND
   (TST (SETQ GOS (LIST (LIST 'GO (SETQ LAB (GENSYM)))))
        (COND (REP (XPUSH GOS (CONS 'PSETQ REP))))
        (SETQ BODY (APPEND BODY GOS))
        (COND ((CAR TST)
               (XPUSH BODY
                (LIST 'COND
                 (LIST (CAR TST)
                  (LIST 'RETURN (MKPROGN (CDR TST))) ) ) ) ))
        (XPUSH BODY LAB) )
   (REP (ERRORFT 411 "DO with repeat but no test")) )
  (SETQ BODY (CONSA 'PROG PVRS BODY))
  (COND
   (LVRS (CONS (LIST 'LAMBDA LVRS BODY) LARGS))
   (T BODY) ) ))

% Similar to DO, but variables are updated sequentially and not in
% parallel.
(DMACRO DO!* (VARS TST !&REST BODY) (SLET (PVRS INI REP TMP)
  (FOREACH X IN (REVERSE VARS) DO (COND
    ((ATOM X) (XPUSH PVRS X))
    (T (XPUSH PVRS (CAR X))
       (COND ((CDR X)
              (COND ((CADR X)
                     (XPUSH INI (LIST 'SETQ (CAR X) (CADR X))) ))
              (COND ((AND (CDDR X)
                      (OR REP (NULL (CADR X))
                       (NOT (EQUAL (CADR X) (CADDR X))) ) )
                     (XPUSH REP (LIST 'SETQ (CAR X) (CADDR X)))
                     (COND ((NULL TMP) (SETQ TMP INI))) )) )) ) ))
  (COND
   (TST (COND
         (TMP (SETQ TMP (PROG1 (CDR TMP) (RPLACD TMP NIL))))
         (T (SETQ TMP INI) (SETQ INI NIL)) )
        (SETQ BODY (CONSA 'WHILE
           (MKPROGN (NCONC1 TMP (LIST 'NOT (CAR TST))))
           (APPEND BODY REP) ))
        (COND ((AND (CAR TST) (CDR TST))
               (SETQ BODY
                (LIST 'RETURN (CONSA 'PROGN BODY (CDR TST))) ) ))
        (SETQ BODY (LIST BODY)) )
   (REP (ERRORFT 411 "DO with repeat but no test")) )
  (CONS 'PROG (CONS PVRS (NCONC INI BODY))) ))
) % End of comment

% INTERNP(X:id):boolean
% True if X is on the oblist.
(DE INTERNP (X)
 (SLET ((Y (FOLDCASE NIL (COMPRESS (EXPLODE X)))) (Z (INTERN Y)))
  (COND ((EQ Y Z) (REMOB Z)))
  (EQ X Z) ) )

% EXPLODEC(X:atom):list of characters
% Corresponds to EXPLODE in the same way as PRIN2 corresponds to PRIN1.
(DE EXPLODEC (X)
 (SLET ((A (EXPLODE X)))
  (COND
   ((STRINGP X) (RPLACD (LAST2PAIR A) NIL) (CDR A))
   ((NOT (IDP X)) A)
   (T (EXPLODECHELP A)) ) ) )

(IFP (NOT (FLAGP 'EXPLODEC 'LOSE))
 (DE EXPLODECHELP (X) (COND
   ((NULL X) NIL)
   ((EQ (CAR X) '!!) (RPLACD (CDR X) (EXPLODECHELP (CDDR X))))
   (T (RPLACD X (EXPLODECHELP (CDR X)))) )) )

% ULCASE(ID:id,CSW:boolean):id
% ULCASE converts ID to an identifier with characters converted to the
% case determined by CSW. If CSW is NIL the result is lower case,
% otherwise it is upper case.
(DE ULCASE (ID CSW) (SLET ((A (EXPLODE ID)) B SW)
  (SETQ CSW (COND (CSW 'UPCASE) (T 'LOCASE)))
  (FOREACH X ON A DO
   (COND ((SETQ B (GET (CAR X) CSW)) (RPLACA X B) (SETQ SW T))) )
  (COND
   (SW (FOLDCASE NIL (INTERN (COMPRESS A))))    % No change of case. JEA
   (T ID) ) ))

(IFP (NOT (FLAGP 'ULCASE 'LOSE)) (PROGN
  (XDEFLIST LOCASE
   (!A !a) (!B !b) (!C !c) (!D !d) (!E !e) (!F !f) (!G !g)
   (!H !h) (!I !i) (!J !j) (!K !k) (!L !l) (!M !m) (!N !n)
   (!O !o) (!P !p) (!Q !q) (!R !r) (!S !s) (!T !t)
   (!U !u) (!V !v) (!W !w) (!X !x) (!Y !y) (!Z !z) )
  (XDEFLIST UPCASE
   (!a !A) (!b !B) (!c !C) (!d !D) (!e !E) (!f !F) (!g !G)
   (!h !H) (!i !I) (!j !J) (!k !K) (!l !L) (!m !M) (!n !N)
   (!o !O) (!p !P) (!q !Q) (!r !R) (!s !S) (!t !T)
   (!u !U) (!v !V) (!w !W) (!x !X) (!y !Y) (!z !Z) ) ))

% CONCATIDL(L:idlist):id
% CONCATIDL returns an id that is the concatenation of the ids in L.
(DE CONCATIDL (L)
 (FOLDCASE NIL (INTERN (COMPRESS (FOREACH X IN L CONC (COND
      ((IDP X) (EXPLODE X))
      (T (I!$FIXUPIDL (EXPLODEC X))) ))))) )

% MAKEID(CHARS:idlist,SW:any):id
% CHARS must be a list acceptable to COMPRESS i.e. a list of single
% character ids. The value of MAKEID is an id made out of all
% characters in CHARS (including special characters).
% The id is INTERNed if SW is true, otherwise not.
(DE MAKEID (X Y) (PROGN
  (SETQ X (FOLDCASE NIL (COMPRESS (I!$FIXUPIDL X))))
  (COND
   (Y (INTERN X))
   (T X) ) ))

(IFP (NOT (AND (FLAGP 'MAKEID 'LOSE) (FLAGP 'CONCATIDL 'LOSE)))
 (DE I!$FIXUPIDL (X)
  (SETQ X (CONS (CAR X) (FOREACH Z IN (CDR X) CONC
     (COND
      ((OR (LITER Z) (DIGIT Z)) (LIST Z))
      (T (LIST '!! Z)) ) )))
  (COND
   ((LITER (CAR X)) X)
   (T (CONS '!! X)) ) ) )

% MAKESTR(CHARS:charlist):string
% MAKESTR returns a string made from the character ids in CHARS.
% As a side effect CHARS is changed.
(DE MAKESTR (X)
 (COND
  ((NULL X) (COMPRESS '(!" !")))
  (T (SLET ((Y (LASTPAIR X)))
      (RPLACD Y '(!"))
      (SETQ X (FOLDCASE NIL (COMPRESS (CONS '!" X))))
      (RPLACD Y NIL)
      X)) ) )

% MAKEINT(DIGITS:digitlist,RADIX:{NIL,id,integer}):integer
% MAKEINT returns an integer made from the digits in DIGITS,
% with radix RADIX.
% Radix NIL is the same as decimal radix.
% If radix is an id other than NIL then it must have a MAKEINTFN
% property which must be a function.
(DE MAKEINT (X Y)
 (SLET (A)
  (COND
   ((NULL X) 0)
   ((NULL Y) (COMPRESS X))
   ((NUMBERP Y)
    (SETQ A (GET (CAR X) 'DIGITVALUE))
    (FOREACH B IN (CDR X) DO
     (SETQ A (PLUS (TIMES A Y) (GET B 'DIGITVALUE))) )
    A)
   ((SETQ A (IGET Y 'MAKEINTFN)) (FUNCALL A X))
   (T (ERRORFT 416 "%r is not a radix" Y)) ) ) )

(IFP (NOT (FLAGP 'MAKEINT 'LOSE)) (XDEFLIST DIGITVALUE
  (!0 0) (!1 1) (!2 2) (!3 3) (!4 4) (!5 5) (!6 6) (!7 7) (!8 8)
  (!9 9) (A 10) (B 11) (C 12) (D 13) (E 14) (F 15) ))

% MAKEFLO(INTEGERPART:digitlist,FRACTIONPART:digitlist,
%         EXPONENT:digitlist):float
% MAKEFLO returns a floating point number made from the digits as
% <INTEGERPART>.<FRACTIONPART>E<EXPONENT>
(DE MAKEFLO (X Y Z)
 (COND (Z (SETQ Y (APPEND Y (CONS 'E Z)))))
 (COMPRESS (APPEND X (CONS '!.
%:LH         (COND (Y) (T '(!0)))
             (OR Y '(!0)) ))) )

(DE LIST2VECTOR (L)
 (SLET ((A (MKVECT (XISUB1 (LENGTH L)))) (I 0))
  (WHILE L (PUTV A I (CAR L)) (XIINCR I) (UNPUSH L))
  A) )

(DS VECTOR (!&REST L) (LIST2VECTOR (LIST . L)))

(DE VECTOR2LIST (V)
 (SLET (A)
  (FORL I ((UPBV V) 0 -1) (XPUSH A (GETV V I)))
  A) )

(DE LISTP (X) (OR (NULL X) (PAIRP X)))

% ORDERP(X:any,Y:any):boolean
% ORDERP returns T if X and Y are EQ or in order and NIL if they are
% not in order. What ordering is used is system dependent.
%(DE ORDERP (X Y) (MUSTDEFINE ORDERP "EXFN.SL"))

% LEXP(X:id,Y:id):boolean
% Returns T if X and Y are in lexical (ASCII) order.
%(DE LEXP (X Y) (MUSTDEFINE LEXP "EXFN.SL"))

% BATCHP returns true if not running interactively i.e. not under the
% control of a tty. This information should really be obtained from
% the operating system.
%!!! Now replaced with variable !*INT which must be set to NIL at
%    startup if in batch. See function SLSPINI in file SLINI.SL.

% STDINP():boolean
% True if the current input channel is the standard input.
% The following definition can often be improved.
(DE STDINP () (SLET ((A (RDS NIL))) (RDS A) (NULL A)))

% STDOUTP():boolean
% True if the current output channel is the standard input.
% The following definition can often be improved.
(DE STDOUTP () (SLET ((A (WRS NIL))) (WRS A) (NULL A)))

% TTYINP():boolean
% True if the current input channel is the tty.
(DE TTYINP () (AND (STDINP) !*INT))

% FILEP(FILE:filespec):boolean
% FILEP is true if FILE can be opened for input.
(DE FILEP (X)
 (COND
  ((PAIRP (SETQ X (ERRSET (IFILEP X) NIL))) (CLOSE (CAR X)) T) ))

% Some compilers do not like (ERRSET (OPEN X 'INPUT) NIL) in FILEP. JEA.
(IFP (NOT (FLAGP 'FILEP 'LOSE))
 (DE IFILEP (Y) (OPEN Y 'INPUT)) )

% PRIN2D Is the same as PRIN2, except for that it sends out the output
% buffer immediately, without waiting for an end of line.
(SAMEDEF PRIN2D PRIN2)

% PRIN1D Is the same as PRIN1, except for that it sends out the output
% buffer immediately, without waiting for an end of line.
(SAMEDEF PRIN1D PRIN1)

% UNREADCH(CH:character):CH
% Puts the character CH back into the input buffer.
% Only the last read character may be put back.
%(DE UNREADCH (CH) (MUSTDEFINE UNREADCH "EXFN.SL"))

% SKIPWHITE skips ignored characters in the input, thereby positioning
% the input at the first significant input character. It requires
% UNREADCH or some other function that puts back a character into the
% input buffer.
(IFP (FNP 'UNREADCH)
 (DE SKIPWHITE ()
  (SLET (A)
   (UNTIL
    (NOT (OR (EQ (SETQ A (READCH)) '! ) (EQ A TAB!*) (EQ A !$EOL!$)))
    NIL)
   (COND ((NOT (EQ A !$EOF!$)) (UNREADCH A)))
   A) )
 % This version of SKIPWHITE is a noop that can be used if there is
 % no UNREADCH.
 (DS SKIPWHITE () NIL) )

% SKIPTO(CH:character id):undefined
% SKIPTO positions the input just after the first occurrence of the
% character CH.
(DE SKIPTO (CH)
 (SLET (A)
  (UNTIL (OR (EQ (SETQ A (READCH)) CH) (EQ A !$EOF!$)))
  A) )

(DE PRIN2T (X) (PROG1 (PRIN2 X) (TERPRI)))

% TTAB(N:integer)
% Prints enough spaces to position output at column N. If already past
% N then this will be at a new line.
(DE TTAB (N)
 (SLET ((A (IDIFFERENCE N (LPOSN))))
  (COND
   ((IMINUSP A) (TERPRI) (SPACES N))
   ((NOT (IZEROP A)) (SPACES A)) ) ) )

% SPACES(N:integer)
% Prints N spaces.
(DE SPACES (N) (REPEATS N (PRINC '! )))

% QTRPRI does a TERPRI if not at beginning of line.
(DE QTRPRI () (COND ((NOT (IZEROP (POSN))) (TERPRI))))

% BELL():undefined
% Rings the bell. This normally means print ascii 7 (cntrl-G).
%(DE BELL () (MUSTDEFINE BELL "EXFN.SL"))

% PGLINE returns a dotted pair: (linenumber . pagenumber) or NIL if
% the facility is not implemented. The returned value refers to the
% currently selected input.
(DE PGLINE () NIL)

% CHANIN(CHAN:channel,[ARG:form]):any
% Evaluates ARGs with input switched to channel CHAN.
(DMACRO CHANIN (CH !&REST BODY) (IOCHANS (LIST CH) NIL BODY))

% CHANOUT(CHAN:channel,[ARG:form]):any
% Evaluates ARGs with output switched to channel CHAN.
(DMACRO CHANOUT (CH !&REST BODY) (IOCHANS NIL (LIST CH) BODY))

(IFP (NOT (FLAGP 'IOCHANS 'LOSE)) (CPROGN
  (SETGLOBALS CHINSTK!* CHOUTSTK!*)
  (PUSHQ SFSTKS!* 'CHINSTK!*)
  (PUSHQ SFSTKS!* 'CHOUTSTK!*)
  (DEFLIST '((CHINSTK!* RDS) (CHOUTSTK!* WRS)) 'SFSTKFN)
  (FLAG '(CHINSTK!* CHOUTSTK!*) 'LASTONLY) ))

(DE IOCHANS (CHI CHO BODY)
 (SLET ((L (LIST (SETQ BODY (LIST 'PROG1 (MKPROGN BODY))))))
  (COND (CHI (NCONC1 BODY '(RDS (POP CHINSTK!*)))
             (XPUSH L (LIST 'XPUSH 'CHINSTK!* (CONS 'RDS CHI))) ))
  (COND (CHO (NCONC1 BODY '(WRS (POP CHOUTSTK!*)))
             (XPUSH L (LIST 'XPUSH 'CHOUTSTK!* (CONS 'WRS CHO))) ))
  (CONS 'PROGN L)  ) )

% STDIN([ARG:form]):any
% Evaluates ARGs with input switched to the STD.
(DMACRO STDIN (!&REST L) (IOCHANS (LIST NIL) NIL L))

% STDOUT([ARG:form]):any
% Evaluates ARGs with output switched to the STD.
(DMACRO STDOUT (!&REST L) (IOCHANS NIL (LIST NIL) L))

% STDINOUT([ARG:form]):any
% Evaluates ARGs with input and output switched to the STD.
(DMACRO STDINOUT (!&REST L) (IOCHANS (LIST NIL) (LIST NIL) L))

(IFP (NOT (AND (FLAGP 'FILEIN 'LOSE) (FLAGP 'FILEOUT 'LOSE))) (CPROGN
  (SETGLOBALS FILESTK!*)
  (COND ((NOT (MEMQ 'FILESTK!* SFSTKS!*))
         (SETQ SFSTKS!* (NCONC1 SFSTKS!* 'FILESTK!*)) ))  %:LH
  (DEFLIST '((FILESTK!* CLOSE)) 'SFSTKFN) ))

% FILEIN(FIL:file,[ARG:form]):any
% Evaluates ARGs with input switched to file FIL.
(DS FILEIN (X !&REST L)
 (PROG1 (CHANIN (CAR (XPUSH FILESTK!* (OPEN X 'INPUT))) . L)
  (CLOSE (POP FILESTK!*)) ) )

(IFP (NOT (FLAGP 'FILEOUT 'LOSE)) (CPROGN
  (SETGLOBALS LINELSTK!*)
  (COND ((NOT (MEMQ 'LINELSTK!* SFSTKS!*))
         (SETQ SFSTKS!* (NCONC1 SFSTKS!* 'LINELSTK!*)) ))  %:LH
  (DEFLIST '((LINELSTK!* LINELENGTH)) 'SFSTKFN)
  (FLAG '(LINELSTK!*) 'LASTONLY) ))

% FILEOUT(FIL:file,[ARG:form]):any
% Evaluates ARGs with output switched to file FIL.
(DS FILEOUT (X !&REST L)
 (PROGN (XPUSH LINELSTK!* (LINELENGTH FILELINELENGTH!*))
  (PROG1 (CHANOUT (CAR (XPUSH FILESTK!* (OPEN X 'OUTPUT))) . L)
   (CLOSE (POP FILESTK!*))
   (LINELENGTH (POP LINELSTK!*)) ) ) )

(DE EPRIN2 (X) (STDOUT (PRIN2D X)))

(DE EPRIN2T (X) (STDOUT (PRIN2 X) (TERPRI) X))

(DE EPRIN2Q (X) (STDOUT (QTRPRI) (PRIN2D X)))

%(DE EPRIN2QT (X) (STDOUT (QTRPRI) (PRIN2 X) (TERPRI) X))

(COMMENT
% WARNING(ARG:any):any
% If !*MSG is true, then WARNING prints a message preceded by *** on
% standard output. If ARG is not a pair then it is printed as is,
% otherwise it must be a list, the elements of which are printed with
% a space inbetween.
(DE WARNING (X)
 (COND (!*MSG (STDOUT
               (QTRPRI)
               (PRIN2 "***")
               (COND
                ((PAIRP X) (FOREACH Y IN X DO (PRINC '! ) (PRIN2 Y)))
                (T (PRINC '! ) (PRIN2 X)) ) ))) ) )

% SETPROMPTER(STR:string or id):string or id
% Sets the prompter and returns the previous prompter.
% If the system does prompt, the system prompter is set and PROMPT
% should be a noop.
% If the system doesn't prompt, the variable PROMPTSTRING!* is set.
(DE SETPROMPTER (STR) (PROG1 PROMPTSTRING!* (SETQ PROMPTSTRING!* STR)))

% PROMPTER(PRM:string,[SXPR:form]):any
% Evaluate all SXPR with the prompter temporarily set to PRM. This
% should be done in such a way that the old prompter is restored also
% when an error occurs. This can either be done by binding the
% prompter as a variable or by using SETPROMPTER together with
% SFSTKS!*.
(DS PROMPTER (PRM !&REST L) (PLET ((PROMPTSTRING!* PRM)) . L))

% PROMPT prompts for input. If the system does prompt automatically,
% PROMPT should be a noop.
(DE PROMPT () (COND ((TTYINP) (EPRIN2 PROMPTSTRING!*))))

% SETFOLDCASE(SW:boolean):boolean
% Sets character folding status and returns old value. If character
% folding is true, then upper and lower case characters are folded
% into one character set, upper or lower depending on system. Which it
% is can be seen in SHPCASEUP!*.
(DE SETFOLDCASE (SW) (PROG1 !*RAISE (SETQ !*RAISE SW)))

% FOLDCASE moved to exfn1i. JEA

% READCHCASE():character
% A READCH that is sensitive to the character folding status.
(SAMEDEF READCHCASE READCH)

% SETGCMESSAGE(SW:boolean):boolean
% Sets gcmessage status and returns old value. If gcmessage is true,
% then on most systems a message is printed when a garbage collection
% occurs.
(DE SETGCMESSAGE (SW) (PROG1 !*GC (SETQ !*GC SW)))

% GCMESSAGE(SW:boolean,[SXPR:form]):any
% Evaluate all SXPR while gcmessage status temporarily is set to SW.
(DS GCMESSAGE (SW !&REST L) (PLET ((!*GC SW)) . L))

% SYESP(MESSAGE:any):boolean
% If running in batch then SYESP always returns the value of
% YESPDEFAULT!* which initially is NIL.
% If running interactively then SYESP prints MESSAGE on the standard
% output, followed by "Answer Y or N: " and then reads a character
% from standard input.
%  If that character is N then NIL is returned, if it is Y then T is
% returned, else "Answer Y or N: " is printed again and a new
% character is read. This goes on until Y or N is read but not more
% than ASKREPEAT!* (normally 5) times. After the last time with no Y
% or N read, YESPDEFAULT!* is returned.
(DS SYESP (MSG) (IYESP MSG T YESPDEFAULT!*))

% SNOP(MESSAGE:any):boolean
% If running in batch then SNOP always returns the value of
% NOPDEFAULT!* which initially is NIL.
% Similar to SYESP, except that Y gives NIL and N gives T.
(DS SNOP (MSG) (IYESP MSG NIL NOPDEFAULT!*))

(DE IYESP (MSG YES DEFLT) (COND
  ((NOT !*INT) DEFLT)
  (T (SLET ((CT ASKREPEAT!*))
    (STDOUT (QTRPRI) (PRIN2L MSG) (TERPRI))
    (UNTIL (OR (EQ MSG !$EOF!$)
               (EQ (SETQ MSG (CAR (EXPLODEC MSG))) 'Y) (EQ MSG 'N)
               (IMINUSP (IDECR CT)) )
           (SETQ MSG (FOLDCASE T
             (STDIN (PROMPTER "Answer Y or N: " (PROMPT) (READ))) )) )
    (COND
     ((EQ MSG 'Y) YES)
     ((EQ MSG 'N) (NOT YES))
     (T (WARNINGFT 48 "Using default %r" (COND
          ((COND (YES (NOT DEFLT)) (T DEFLT)) 'N)
          (T 'Y) ))
        DEFLT) ) )) ))

% READIN(FILE:filespec,SW:extrabolean):"Finished loading"
% Reads in the file FILE. If SW is true the results are printed, else
% not.
(DE READIN (FIL READIN!$SW)
 (FILEIN FIL
  (PROG (CLOC!* READIN!$X)
LP (COND
    ((NOT (PAIRP (SETQ READIN!$X (ERRSETT (SKIPWHITE))))) (GO EQ))
    ((EQ (SETQ READIN!$X (CAR READIN!$X)) !$EOF!$) (GO EN)) )
   (SETQ CLOC!* (CONS FIL (PGLINE)))
   (COND
    ((NOT (PAIRP (SETQ READIN!$X (ERRSETT (READ))))) (GO EQ))
    ((EQ (SETQ READIN!$X (CAR READIN!$X)) !$EOF!$) (GO EN)) 
    ((PAIRP (ERRSETT (DFEVAL READIN!$X READIN!$SW))) (GO LP))
    ((NOT !*ECHO) (PRINT READIN!$X)) )
ER (COND ((SYESP "Continue loading? ") (GO LP)) )
   (ERRORFT 440 "Error return")
EQ (COND ((NOT (EQ READIN!$X !$EOF!$)) (GO ER)))
EN (RETURN "Finished loading") ) ) )

% DFEVAL(XPR:form,SW:extrabolean):any
% DFEVAL implements the !*DEFN, DFPRINT!* mechanism of RLISP.
% DFEVAL is designed to be a top level eval and print.
% If the global variable !*DEFN is false then XPR is evaluated and if
% SW is true then the result is printed. If the flag !*SUPPRESSANS is
% true then a result EQ to the value of SUPPRESSANS!* is not printed
% even if SW is true.
%  If !*DEFN is true then DFPRINT!* is applied (or if DFPRINT!* is
% NIL, PRETTYPRINT). This is further controlled by flagging the first
% element of XPR, EVAL, IGNORE or DFEXPAND.
(DE DFEVAL (DFEVAL!$X DFEVAL!$SW) (COND
  ((NOT !*DEFN) (COND
    (DFEVAL!$SW (I!$PRNEVL DFEVAL!$X
                 (AND !*INT (STDINP) (STDOUTP))
                 (AND !*SUPPRESSANS SUPPRESSANS!*) ))
    (T (EVAL DFEVAL!$X)) ))
  ((NOT (PAIRP DFEVAL!$X)) (DFPRINT DFEVAL!$X))
  ((EQ (CAR DFEVAL!$X) 'PROGN)
   (FOREACH Y IN (CDR DFEVAL!$X) DO (DFEVAL Y NIL)) )
  ((FLAGP (CAR DFEVAL!$X) 'IGNORE) (DFEVALQ DFEVAL!$X))
  ((FLAGP (CAR DFEVAL!$X) 'EVAL)
   (DFPRINT DFEVAL!$X)
   (DFEVALQ DFEVAL!$X) )
  ((AND (FLAGP (CAR DFEVAL!$X) 'DFEXPAND)
        (NOT (EQUAL DFEVAL!$X (SETQ DFEVAL!$SW (MACEXPD1 DFEVAL!$X)))) )
   (DFEVAL DFEVAL!$SW NIL) )
  (T (DFPRINT DFEVAL!$X)) ))

(DE I!$PRNEVL (PRNEVL!$X PRESW!$ POSTV!$)
 (COND (PRESW!$ (QTRPRI) (TERPRI)))
 (SETQ PRNEVL!$X (EVAL PRNEVL!$X))
 (COND ((NOT (AND POSTV!$ (EQ PRNEVL!$X POSTV!$)))
        (QTRPRI) (PRIN1 PRNEVL!$X) (TERPRI) ))
 PRNEVL!$X)

(IFP (NOT (FLAGP 'DFEVAL 'LOSE))
 (DE DFPRINT (X) (COND
   (DFPRINT!* (FUNCALL DFPRINT!* X))
   (T (QTRPRI) (PRETTYPRINT X) (TERPRI)) )) )

(IFP (NOT (FLAGP 'DFEVAL 'LOSE))
 (DE DFEVALQ (X) (COND
   (DFEVAL!* (FUNCALL DFEVAL!* X))
   (T (EVAL X)) )) )

% RESET(): doesn't return
% Supposed to reset to top-level of lisp. This implementation only
% resets to next ERRSET. If possible change in system dependent file.
%(DE RESET () (MUSTDEFINE RESET "EXFN.SL"))
(DE RESET () (ERROR 450 "Reset"))

% QUIT():undefined
% Quits lisp. May allow a return into Lisp for some implementations.
%(DE QUIT () (MUSTDEFINE QUIT "EXFN.SL"))

% BYE(): doesn't return
% Leaves lisp definitely.
%(DE BYE () (MUSTDEFINE BYE "EXFN.SL"))

% SHPDEBUG(SW:boolean):boolean
% If SW is true various debugging modes are turned on.
% What this function does is system dependent. On some systems SW can
% have more values than true and NIL.
% SHPDEBUG returns the old value of debugging.
(DE SHPDEBUG (X)
 (PROG1 !*SHPDEBUG
  (COND ((SETQ !*SHPDEBUG (SETQ !*BACKTRACE X))
         (SETQ !*MSG (SETQ !*GC (SETQ !*LOADMSG T))) )) ) )

% TIME():integer
% TIME returns the time in milliseconds since jobstart or something
% like that. It is not important what TIME = 0 corresponds to as the
% only use of TIME is the difference in value for to events.
%(DE TIME () (MUSTDEFINE TIME "EXFN.SL"))

% DATETIME():string
% DATETIME returns in unspecified form current date and time.
% Only to set the date e.g. an image was created, may return NIL for
% Lisp systems where this can not be done.
(DE DATETIME () NIL)

%### EXFN2I.SL end
