%### LOADI.SL =======================================================
%
%       LOADI.SL
%       --------
%
%       Copyright (c) 1982 by Inge Frick, Institute of Theoretical
%       Physics, Vanadisvagen 9, S-113 46 Stockholm, Sweden
%
%       The loading mechanism regards a full file description to
%       consist of three parts: Prefix, file name and extension.
%       File name and extensions might be separated by a
%       file-extension-separator.
%       There is no separator between prefix and file name, if one is
%       needed it has to be at the end of the prefix.
%       The prefix and/or the extension might be missing. If the prefix
%       is missing then a search through a prefix list is performed.
%       If the extension is missing then a search through an extension
%       list is performed for each prefix (or the specified prefix if
%       it wasn't missing).
%       The extension found determines what function is used to
%       actually load the file. If an extension has the property INFN,
%       then that property must be a function of two arguments, the
%       first argument is the name of the file to load and the second
%       argument is a switch that is on if the results of evaluations
%       are to be printed. If an extension has no INFN property then
%       the value of the fluid INFN!* is used as default loading
%       function.
%
%       Some functions here are system dependent and have not been
%       defined. Look for the indicator MUSTDEFINE.
%       They are: ADDEXTENSION, ADDPREFIX, PREFIXOF and FILNAM
%
%       Variables DEFAULTEXTENSION!*, MSEARCHLIS!* and MEXTLIS!* must
%       have an appropriate value.
%
%====================================================================

(FUNCTIONS LOAD RELOAD SLOAD DSKIN INCLUDE PRELOAD POSTLOAD MODULOAD
           TESTFILE ITESTFILE SETCASE OUTFILENAME ILOADEDP LOADEDP MLOAD
           LOADX LOADP FINDFILE IFINDFILE LOADFIL SEARCHFILN SEARCHFIL1)

(ENTRYPOINTS LOAD RELOAD SLOAD DSKIN INCLUDE PRELOAD POSTLOAD MODULOAD
             OUTFILENAME LOADEDP MLOAD LOADX LOADP FINDFILE)

(LOCALFUNCTIONS LOADFIL SEARCHFILN SEARCHFIL1)

(USEDPROPERTIES MACEXPDFN PATH EXPATH INFN)

(USEDFLAGS IGNORE)

(SETFLUIDS
 MLIST!*                        % List of loaded modules.
 MEXTLIS!*                      % Default extensions.
 FASLOUTEXT!*                   % Extension for created fasl files.
 MSEARCHLIS!*                   % Prefix search list.
%FIL!*                          % File or module name. In exfn2i.sl.
 (LOADARG!* NIL)                % Latest argument for LOAD.
 (!*LOADMSG T)                  % Switch, on for message from LOADX.
 (LOADMSGTAB!* 0)               % Position for Load message.
 DEFAULTEXTENSION!*             % Default extension.
 (HERE NIL)                     % Directory for calling file.
 (HERELIST NIL)     % Recursive list of directories for calling files.
 (MODULNAM!* UNKNOWN)           % Module name in LOADX.
 (!*IPRN T)                     % DSKIN normally prints out.
 (INFN!* READIN)                % Default loading function.
 (DSKSEARCHLIS!* (HERE))        % Search on calling files directory.
 ((DSKEXTLIS!* (LIST NIL))) )   % Don't add extension.

(DPUT DSK!: PATH (NIL))         % Make DSK mean present directory.

(DFEXPR LOAD (!&REST L)
% This is the standard loading function. It loads modules quietly
% using the standard prefix and extension lists. If a module is
% already loaded it will not be loaded. Loaded modules are marked
% as being loaded.
 (SETQ LOADARG!* L)
 (PLET ((LOADARG!* LOADARG!*))
  (FOREACH X IN L DO (MLOAD X)) 
  SUPPRESSANS!*) )

(DFEXPR RELOAD (!&REST L)
% Same as LOAD expt for that it will load a module also when it has
% been loaded before.
 (COND
  (L (SETQ LOADARG!* L))
  (T (SETQ L LOADARG!*)) )
 (PLET ((LOADARG!* LOADARG!*))
  (FOREACH X IN L DO (LOADX X MSEARCHLIS!* MEXTLIS!* T NIL))
  SUPPRESSANS!*) )

(DFEXPR SLOAD (!&REST L)
 (FOREACH X IN L DO (LOADX X DSKSEARCHLIS!* DSKEXTLIS!* T T))
 SUPPRESSANS!*)
% Loads a file (module) without adding any missing prefix and/or
% extension (this can be changed by changing DSKSEARCHLIS!* and/or
% DSKEXTLIS!*). A module will be loaded also when it has been
% previously loaded. A loaded module is marked as loaded. In case
% of a source-file being loaded, results of evaluations are printed.

(DFEXPR DSKIN (!&REST L)
 (FOREACH X IN L DO (LOADX X DSKSEARCHLIS!* DSKEXTLIS!* 'N !*IPRN))
 SUPPRESSANS!*)
% Same as SLOAD except for that files are not marked as loaded and
% that DSKIN is evaluated at compile-time and not at load-time.

% The following three functions are only defined for compatibility
% reasons.

% As loadi.sl here redefines INCLUDE to a definition not
% compatible with the temporary definition given in another file the
% current file (loadi.sl) must be the very last file to be loaded
% during the bootstrapping procedure to make a macro file.
(COND ((NOT (FLAGP 'INCLUDE 'LOSE)) (PROGN (REMD 'INCLUDE) NIL)))

(MOVD 'INCLUDE 'DSKIN)

(MOVD 'PRELOAD 'LOAD)

(MOVD 'POSTLOAD 'LOAD)

% MODULOAD should for some Lisps (e.g. Cambridge University Lisp)
% create a module from a source file, otherwise be the same as LOAD.
(MOVD 'MODULOAD 'LOAD)

(FLAG '(INCLUDE DSKIN MODULOAD) 'IGNORE)

(FOREACH X IN
  '(LOAD SLOAD INCLUDE DSKIN RELOAD PRELOAD POSTLOAD MODULOAD)
  DO (PUT X 'MACEXPDFN 'RETARG) )

%(DE ADDEXTENSION (X Y) (MUSTDEFINE ADDEXTENSION "LOADI.SL"))

%(DE ADDPREFIX (X Y) (MUSTDEFINE ADDPREFIX "LOADI.SL"))

%(DE ADDSUBDIR (X Y) (MUSTDEFINE ADDSUBDIR "LOADI.SL"))

(DS TESTFILE (X DUMMY1 DUMMY2) (ITESTFILE X))

(DE ITESTFILE (X) (COND ((FILEP X) X)))

(DS SETCASE (X) X)      % Noop if not changed.

(DE OUTFILENAME (PR FIL!* EX)
 (SLET ((A (FILNAM FIL!*)))
  (COND ((AND EX (NULL (CDDR A))) (SETQ FIL!* (ADDEXTENSION FIL!* EX))))
  (COND ((AND PR (NULL (CAR A))) (SETQ FIL!* (ADDPREFIX PR FIL!*))))
  (SETCASE FIL!*) ) )

(DE SETLOADED (X) (PUSHQ MLIST!* X))

(DE SETNOTLOADED (X) (REMPOP MLIST!* X))

(DS ILOADEDP (X) (MEMQ X MLIST!*))

(DE LOADEDP (X) (ILOADEDP X))

(DE MLOAD (X) (LOADX X MSEARCHLIS!* MEXTLIS!* NIL NIL))
% EXPR version of LOAD

(DE LOADX (FIL SRLIS EXLIS RESW IPRN)
% The same as LOADP, except that an error occurs if the file is
% not found.
 (OR
  (LOADP FIL SRLIS EXLIS RESW IPRN)
  (ERRORFT 441 "Can't find file %w" FIL) ) )

(DE LOADP (FIL!* SRLIS EXLIS RESW IPRN)
% Loads file or module FIL!* using prefix list SRLIS and extension
% list EXLIS if prefix and/or extension is missing from FIL!*.
% File name of FIL!* is used as name of module FIL!* when testing
% weather it has been loaded previously. What function is used to
% actually load the file, depends on the extension of the found file.
% If RESW is NIL then FIL!* will not be loaded if it has been loaded
% previously. If RESW is not eq 'N then module FIL!* will be marked as
% loaded if it is found. If IPRN is NIL then loading is done quietly,
% otherwise some loading functions will print the results of
% evaluations. If module is not found then NIL is returned otherwise
% module name as an id or (if already loaded) the string
% "Already loaded" is returned.
 (SLET ((NM (FILNAM FIL!*))     % Take apart file description.
        (EX (CDDR NM))          % Extension
        (MNM (CADR NM)) )       % Module-name = file name
  (COND ((AND EX (NOT (MEMQ EX EXLIS))) (SETQ RESW 'N))) % Not a Module
  (COND
   ((AND (NULL RESW) (ILOADEDP MNM)) "Already loaded")  % If not loaded
   ((SETQ NM (IFINDFILE NM SRLIS EXLIS))        % then if file found
    (LOADFIL (CAR NM) (OR (GET (CDR NM) 'INFN) INFN!*)  % load it.
             (COND ((NOT (EQ RESW 'N)) MNM)) IPRN) ) ) ) )

(DE FINDFILE (FIL!*)
% Find file FIL!* using search list MSEARCHLIS!* and extension list
% MEXTLIS!*. If found, return a dotted pair with car being the
% specification for the found file and cdr being the file extension.
% If not found, return NIL. FIL!* is a fluid that is used in IFINDFILE.
 (IFINDFILE (FILNAM FIL!*) MSEARCHLIS!* MEXTLIS!*) )

(DE IFINDFILE (NM SRLIS EXLIS)
% The fluid variable FIL!* is the name of a file. Find that file using
% search list SRLIS and extension list EXLIS. NM is the parsed file
% name (output from FILNAM).
% If found, return a dotted pair with car being the specification for
% the found file and cdr being the file extension. If not found,
% return NIL.
 (SLET ((FL FIL!*) (EX (CDDR NM)) EF X)
  (COND
   (EX (SETQ EF (MEMQ EX EXLIS)) % Got extension, set flag if standard
       (SETQ EXLIS NIL) )       % ext and do not search for extension.
   (T (COND
       ((NULL EXLIS) (SETQ EX DEFAULTEXTENSION!*))
       ((NULL (CDR EXLIS)) (SETQ EX (CAR EXLIS)) (SETQ EXLIS NIL)) )
      (COND (EX         % Put single extension into file description.
             (SETQ FL (ADDEXTENSION FL (SETCASE EX))) )) ) )
  (COND
   ((CAR NM)                            % Prefix is
    (SETQ SRLIS                         %  list of directories or
     (COND ((PAIRP (CAR NM)) (CAR NM))) )) %  single directory.
   ((AND SRLIS (PAIRP (CAR SRLIS))      % If search list has indicator
     (NULL (CDAR SRLIS)) )              %  for extension specific search
    (SETQ SRLIS (OR                     %  list, then if specific search
      (GET EX (CAAR SRLIS))             %  list use that
      (CDR SRLIS) )) ) )                %  else use rest of search list.
  (COND                                 % Search for file with:
   (EXLIS (SEARCHFILN FL SRLIS EXLIS))  %  Extension not known.
   ((NULL (SETQ X (GET EX 'LOADFN)))    %  Not a pseudo file.
    (SEARCHFIL1 FL SRLIS EX EF) )       %  Extension known.
   ((SETQ X (GET (CADR NM) X))
    (CONS (CONS FL X) EX) ) ) ) )       %  Pseudo file.

(DE LOADFIL (FL INFN MNM !*IPRN)
% Loads file FL using loading function INFN.
% If FL is a pair then it is a pseudo file, (CAR FL) is the file name
% and (CDR FL) is the function to execute by the INFN instead of
% loading. If MNM is not NIL then it is the name of the module being
% loaded, which means that MODULNAM!* is bound to MNM and the module
% will be  marked as loaded.
% If !*IPRN is nil then loading is done quietly, otherwise some loading
% functions will print the results of evaluations.
 (SLET ((SW (PAIRP FL)) (FIL!* (COND (SW (CAR FL)) (T FL))))
  (COND (!*LOADMSG (PRINTF "%tLoading file %w%n" LOADMSGTAB!* FIL!*)))
  (PLET ((MODULNAM!* (OR MNM MODULNAM!*))
         (HERE (COND ((NOT SW) (PREFIXOF FIL!*))))
         (LOADMSGTAB!* (IADD1 LOADMSGTAB!*)) )
   (PLET ((HERELIST (COND               % Set present prefix.
            ((OR SW (AND HERELIST (EQUAL HERE (CAR HERELIST))))
             HERELIST )
            (T (CONS HERE HERELIST)) )))
    (FUNCALL INFN (COND (SW (CDR FL)) (T FL)) !*IPRN) ) )
  (COND (!*LOADMSG (PRINTF "%tFile %w loaded%n" LOADMSGTAB!* FIL!*)))
  (COND (MNM (SETLOADED MNM)))          % Mark that it has been loaded.
  FIL!*) )

(DE SEARCHFILN (FIL SRLIS EXLIS)
% Searches for file with file name FIL, extension from EXLIS and prefix
% from SRLIS. The extension list is a list of ids. The prefix list has
% this format:
%  <prefixlist>    :== NIL|(<prefixelement> . <prefixlist>)
%  <prefixelement> :== <prefix>|(<prefix> . <extension list>)
%  <prefix>        :== <id>|<string>
% The case when a prefix element is an extension list with a prefix in
% front means that for that prefix the extension is taken not from
% EXLIS but rather from the given extension list. This way a complete
% control over the order of search can be achieved.
 (PROG (A B C W SR)
    (COND ((NULL EXLIS) (RETURN NIL)))
    (COND ((NULL SRLIS) (SETQ SRLIS '(NIL))))
 LP (COND
     ((PAIRP (SETQ A (CAR SRLIS)))  % There is a special extension list.
      (COND ((AND (PAIRP (SETQ C (CDR A)))
              (NULL (CAR C)) )      % Remove match all prefix if any.
             (UNPUSH C) ))
      (COND ((NULL (SETQ C (INTERSECTIONQ
                (COND ((LISTP C) C) (T (EVAL C)))
                EXLIS )))
             (GO LE) ))
      (SETQ A (CAR A)) )                % The prefix.
     (T (SETQ C EXLIS)) )%Use same extension list as for other prefixes.
 LT (COND ((MEMBER A W) (GO LE)))       % Seen this prefix before.
    (XPUSH W A)                         % Remember it.
    (COND
     ((EQ A 'HERE) (SETQ A HERE) (GO LT))
     ((EQ A 'HERELIST) (COND            % Same prefix as calling files.
       (HERELIST (XPUSH SR SRLIS) (SETQ SRLIS HERELIST) (GO LP))
       (T (GO LE)) )))
    (SETQ A (ADDPREFIX A FIL))          % Put prefix in front.
 LC (COND
     ((SETQ B (GET (CAR C) 'LOADFN))    % Extension for pseudo files.
      (COND ((SETQ B (GET FIL B))       % If it is a pseudo file return
             (RETURN (CONS              %  ((file . filefn) . ext)
               (CONS (ADDEXTENSION A (CAR C)) B)
               (CAR C) )) )) )
     ((PROGN (SETQ B (ADDEXTENSION A (SETCASE (CAR C)))) %Ordinary file.
       (SETQ B (TESTFILE B FIL T)) )    % If file found,
      (RETURN (CONS B (CAR C))) ) )     %  return (file . ext)
    (COND ((SETQ C (CDR C)) (GO LC)))   % Try next extension if any.
 LE (COND
     ((SETQ SRLIS (CDR SRLIS)) (GO LP)) % Try next prefix. If none
     (SR (SETQ SRLIS (CAR SR)) (UNPUSH SR) (GO LE)) )
    (RETURN NIL) ) )                    % return NIL.

(DE SEARCHFIL1 (FIL SRLIS EX EF)
% Similar to SEARCHFILN, but FIL has got an extension already.
 (PROG (A C SEENPRFX SR)
    (COND ((NULL SRLIS) (SETQ SRLIS '(NIL))))
 LP (COND ((PAIRP (SETQ A (CAR SRLIS))) % If special extension list then
           (COND
            ((NOT (PAIRP (SETQ C (CDR A))))
             (SETQ C (EVAL C)) )        % Name of ext. list
            ((CAR C) NIL)               % Extension must match
            ((NOT (AND EF (UNPUSH C) (CAR C)))
             (SETQ C NIL) ) )           % Any extension accepted
           (COND ((AND C (NOT (MEMQ EX C))) % If extension not there,
                  (GO LE) ))            % then skip this prefix.
           (SETQ A (CAR A))))           % Ignore special extension list.
 LT (COND ((NOT (MEMBER A SEENPRFX))    % Not seen this prefix before.
           (XPUSH SEENPRFX A)           % Remember it.
           (COND
            ((EQ A 'HERE) (SETQ A HERE) (GO LT))
            ((EQ A 'HERELIST)
             (COND (HERELIST            % Same prefix as calling files.
               (XPUSH SR SRLIS) (SETQ SRLIS HERELIST) (GO LP) )))
            ((SETQ A (TESTFILE (ADDPREFIX A FIL) FIL NIL)) % If file
             (RETURN (CONS A EX)) ) ) )) % found, return (file . ext).
 LE (COND
     ((SETQ SRLIS (CDR SRLIS)) (GO LP)) % Else try next prefix. If none
     (SR (SETQ SRLIS (CAR SR)) (UNPUSH SR) (GO LE)) )
    (RETURN NIL) ) )                    % return NIL.

%(DE FILNAM (FIL) (MUSTDEFINE FILNAM "LOADI.SL"))
% FIL is a file description. FILNAM takes FIL apart and returns
% (<prefix> <file name> . <extension>) where <file name> and
% <extension> are interned ids and <prefix> is either a string an id
% or a search list. When some part is missing, it is returned as NIL.
% Various special system dependent interpretations of FIL might be
% done, in which case the fluid variable FIL!* is updated. A relative
% prefix is returned as NIL. This function is necessarily
% system-dependent as it has to know how files are described.

%(DE PREFIXOF (FIL) (MUSTDEFINE PREFIXOF "LOADI.SL"))
% FIL is a real file description with no special interpretation.
% PREFIXOF returns the prefix part of FIL, or NIL if there is no
% prefix. Both absolute and relative prefixes are returned.
% This function is necessarily system-dependent as it has to know
% how files are described.

% ADDSUBDIRS. TP is NIL or a list of file prefixes for top directories.
% SUL is a list of subdirectories. ADDSUBDIRS adds to MSEARCHLIS!*
% a list constructed by concatenating the top directories with the
% subdirectories.
(DE ADDSUBDIRS (TP SUL)
 (SETQ MSEARCHLIS!* (NCONC MSEARCHLIS!* (COND
    ((NULL TP)
     (FOREACH Y IN (SUBDRS SUL) COLLECT (ADDSUBDIRS!-HELP NIL Y)) )
    (T (FOREACH Y IN (SUBDRS SUL) CONC
        (FOREACH X IN TP COLLECT (ADDSUBDIRS!-HELP X Y)) )) ))) )

(DE ADDSUBDIRS!-HELP (X Y)
 (SLET ((ESW (PAIRP Y)))
  (COND (ESW (SETQ ESW (CDR Y))
             (COND ((IDP ESW) (SETQ ESW (EVAL ESW))))
             (SETQ Y (CAR Y)) ))
  (COND
   ((PAIRP Y) (FOREACH DIR IN Y DO (SETQ X (ADDSUBDIR X DIR))))
   (T (SETQ X (ADDSUBDIR X Y))) )
  (COND
   (ESW (CONS X ESW))
   (T X) ) ))

% SUBDRS modifies original argument to ADDSUBDIRS in order to replace
% system specific things as some subdirectory names and extension
% lists. Added by JEA.
(DE SUBDRS (SX)
 (SLET (SB)
  (COND
   ((NULL SX) NIL)
   ((NOT (PAIRP SX))
    (OR (GET SX 'SUBDRS) SX) )
   ((PAIRP (CAR SX))
    (CONS (SUBDRS (CAR SX)) (SUBDRS (CDR SX))) )
   ((NULL (SETQ SB (GET (CAR SX) 'SUBDRS)))
    (CONS (CAR SX) (SUBDRS (CDR SX))) )
   ((EQ 'NONE SB) (SUBDRS (CDR SX)))
   ((PAIRP SB) (APPEND SB (SUBDRS (CDR SX))))
   (T (CONS SB (SUBDRS (CDR SX)))) )) )

(PUT 'LISPSYS 'SUBDRS (SETCASE (SYSTYPE 'LISPSYSTEM)))

%### LOADI.SL end
