;;;; -*- Mode: LISP; Syntax: Common-Lisp; Package: CL-USER; Base: 10 -*-

;;;; PACKAGE-SETUP, Module SLCP

;;; ***************************************************************************
;;;
;;;        (c) Copyright 1985 by Lucid Inc.
;;;        Unrestricted permission is granted to copy, modify
;;;        or redistribute this file.
;;;
;;; ***************************************************************************

;;; Set up Standard Lisp packages

;;; Programmer: Eric Benson

;;; Edit-History:
;;;
;;; Created: Saturday the twentieth of July, 1985; 7:41:29 pm
;;; 29-Apr-87 eb: Added REMPROP to shadowed symbols
;;;
;;; End-of-Edit-History

(eval-when (eval compile load)
  (unless (find-package 'sl)
    (make-package 'standard-lisp :nicknames '(sl))))

#+Excl
(import '(franz:memq) 'sl)

#+Symbolics
(import '(scl:memq scl:defsubst) 'sl)

#+Lucid
(import '(user::defsubst user::memq) 'sl)

(let ((sl (find-package 'sl)))
  (shadow '(minusp zerop get assoc member intern
		   error map mapc mapcan mapcar mapcon maplist
		   constantp length vectorp subst sublis equal
		   delete open close print prin1 princ function
		   remprop)
	  sl)
  #+(and Excl DG)
  (shadow 'nconc sl)
  (dolist (x '("CODEP" "EQN" "FIXP" "IDP" "ONEP" "PAIRP"
		  "COMPRESS" "EXPLODE2" "EXPLODEC"
		  "FLAG" "FLAGP" "PUT"
		  "REMFLAG" "DE" "DF" "DM" "GETD" "PUTD" "REMD"
		  "FLUID" "FLUIDP" "GLOBAL" "GLOBALP" "UNFLUID"
		  "EMSG*" "ERRORSET"
		  "GETV" "MKVECT" "PUTV" "UPBV" "DIVIDE"
		  "MAX2" "MIN2" "PLUS2"
		  "TIMES2" "DEFLIST" "DIGIT" "LITER"
		  "PAIR" "EVLIS" "EXPAND" "FUNCTION"
		  "EJECT" "EXPLODE" "READCH"
		  "LINELENGTH" "LPOSN" "PAGELENGTH" "POSN"
		  "PRIN2" "RDS" "WRS"
		  "*COMP" "*ECHO" "$EOF$" "$EOL$"
		  "*GC" "*RAISE" "EXPR" "FEXPR" "MACRO"
		  "INPUT" "OUTPUT"
		  "ERROR"
		  "MAP" "MAPC" "MAPCAN" "MAPCAR" "MAPCON" "MAPLIST"
		  "LENGTH"
		  "CONSTANTP"
		  "VECTORP"
		  "SUBST"
		  "SUBLIS"
		  "MEMQ"
		  "MINUSP" "ZEROP"
		  "GET" "REMPROP" "ASSOC" "MEMBER"
		  "LESSP" "ADD1" "SUB1" "MINUS"
		  "GREATERP"
		  "DIFFERENCE" "PLUS" "TIMES" "QUOTIENT" "REMAINDER"
		  "FIX"
		  "DELETE" "OPEN" "CLOSE" "PRINT" "PRIN1" "PRINC"
		  "EXPLODE" "READCH"
		  "DEFINE-SYNONYM" "DEFINE-INLINE"
		  "*STANDARD-LISP-READTABLE*" "SINGLE-CHARACTER-SYMBOLS"))
    (export (lisp:intern 
		#-(and Excl (not Allegro))
		x 
		#+(and Excl (not Allegro))
		(string-downcase x)
		sl) sl)))

;;;All single character symbols are exported.
(eval-when (eval compile load)

;;; STRING-CHAR-CODE-LIMIT is a machine-dependent value: the upper
;;; bound of values for CHAR-CODE of string characters.  In most
;;; implementations it will be the same as CHAR-CODE-LIMIT.  In
;;; Symbolics Common Lisp CHAR-CODE-LIMIT is 65536.
(defparameter sl::string-char-code-limit (min char-code-limit 256))
)

(defparameter sl::single-character-symbols
	      '#.(let ((array (make-array sl::string-char-code-limit)))
		   (dotimes (i sl::string-char-code-limit)
		     (setf (svref array i)
			   (make-string 1 :initial-element (code-char i))))
		   array))

(let ((sl (find-package 'sl))
      (syms sl::single-character-symbols))
  (dotimes (i sl::string-char-code-limit)
    (export (setf (svref syms i) (lisp:intern (svref syms i) sl)) sl)))
