%### OBLANL.SL ======================================================
%
%       OBLANL.SL
%       --------
%
%       Copyright (c) 1992 by Jan E. Aman and Inge Frick, Institute of
%       Theoretical Physics, Vanadisvagen 9, S-113 46 Stockholm, Sweden.
%
%       Analyses the OBLIST and print ids with their properties.
%
%       This version will work for PSL, CSL, Cambridge University Lisp
%       (Culisp) and hopefully for other Standard Lisp like Lisps at
%       least after minor modifications.
%
%====================================================================

% Not used to make Sheep Lisp, intended for debugging purposes only.

% Do in PSL: (WRS (OPEN "psl.obl" 'OUTPUT)) (DSKIN "oblanl.sl") (QUIT)
% Do in CSL: (WRS (OPEN "csl.obl" 'OUTPUT)) (RDF "oblanl.sl") (STOP 0)
% In Culisp is should work with
% (WRS (OPEN "culisp.obl" 'OUTPUT)) (RDF "oblanl.sl") (STOP 0)
% but only e.g. Unix redirection works: % lisp < oblanl.sl > culisp.obl

% This is for PSL only.
(COND
  ((NOT (GETD 'OBLIST))
    (LOAD COMPILER)                     % Load also compiler for PSL.
    (LOAD GSORT)                        % IDSORT is defined here.
    (COND
      ((GETD 'IDSORT) (DE MYSORT (L) (IDSORT L)))
      (T (DE MYSORT (L) L)) )
    (DE OBLIST NIL
      (PROG (OB)
        (MAPOBL (FUNCTION (LAMBDA (X) (SETQ OB (CONS X OB)))))
        (RETURN (MYSORT (REVERSE OB))) ) ) ))

% This is for Bare PSL.
(COND
  ((NOT (GETD 'BOUNDP))
    (DE BOUNDP (X) (PAIRP (ERRORSET X NIL NIL))) ))

(COND
  ((NOT (GETD 'SETDIFFQ))
   (DE SETDIFFQ (X Y)   % Returns list of elements in X not EQ any in Y.
     (COND
       ((NULL X) NIL)
       ((MEMQ (CAR X) Y) (SETDIFFQ (CDR X) Y))
       (T (CONS (CAR X) (SETDIFFQ (CDR X) Y))) ))))

(PROGN
  (COND ((BOUNDP '!$!$OBL) (SETQ O!$!$OBL !$!$OBL)))

  (SETQ !$!$OBL (OBLIST))

  (COND
    ((BOUNDP 'O!$!$OBL)
     (SETQ !$!$OBL (SETDIFFQ !$!$OBL O!$!$OBL)) ) )
 NIL )

(SETQ !$FUNS!$ NIL)

(SETQ !$EXPR!$ NIL)

(SETQ !$FEXPR!$ NIL)

(SETQ !$NEXPR!$ NIL)

(SETQ !$MACRO!$ NIL)

(SETQ !$MMACRO!$ NIL)

(SETQ !$OTHER!$ NIL)

(SETQ !$BOTH!$ NIL)

(SETQ !$VALUE!$ NIL)

(SETQ !$FLUID!$ NIL)

(SETQ !$GLOBAL!$ NIL)

(SETQ !$PROP!$ NIL)

(SETQ !$NIDP!$ NIL)

(SETQ !$FLAG!$ NIL)

(COND
  ((NOT (GETD 'XPUSH))
    (COND
      ((BOUNDP 'PROMPT!*)                       % Test for Culisp.
% Bootstrapping of macro definition in Culisp.
        (COND
          ((GETD 'DCM) NIL)                     % For ACN Culisp.
          ((GETD '!%DM) (SETQ DCM !%DM))        % For Reduce (JPFF).
          ((GETD 'DM) (SETQ DCM DM)) )          % For JPFF Culisp.
        (DCM XPUSH U
          (LIST 'SETQ (CAR U) (LIST 'CONS (CADR U) (CAR U))) ) )
      (T
        (DM XPUSH (U)
          (LIST 'SETQ (CADR U) (LIST 'CONS (CADDR U) (CADR U))) ) ) )))

(COND
  ((NOT (GETD 'MYSORT))
    (COND
      ((NOT (GETD 'SORT)) (DE MYSORT (L) L))
      (T
        (DE MYSORT (L) (SORT L 'MYORDERP))      % This is for CSL.
        (DE MYORDERP (X Y)
          (COND
            ((AND (NUMBERP X) (NUMBERP Y)) (ORDERP Y X))
            ((NUMBERP Y) T)
            ((NUMBERP X) NIL)
            (T (ORDERP X Y)) )) )) ))

% This is for Bare PSL and Reduce based on it:
(COND
  ((NOT (GETD 'PLIST))
    (DE PLIST (ID) (PROP ID)) ))

% This is for Culisp:
(COND ((AND (NOT (GETD 'PRIN1)) (GETD 'PRIN)) (SETQ PRIN1 PRIN)))
(COND ((AND (NOT (GETD 'PRIN2)) (GETD 'PRINC)) (SETQ PRIN2 PRINC)))

% Try to find out if FLAG is implemented as a PUT (with value T).
(SETQ FLAGASPUT!$ NIL)
(PROG (OFL RS)
  (SETQ OFL (FLAGP 'FLUID '!$!$FLG))
  (FLAG '(FLUID) '!$!$FLG)
  (SETQ RS (ERRORSET '(GET 'FLUID '!$!$FLG) NIL))
  (COND ((AND RS (PAIRP RS) (CAR RS)) (SETQ FLAGASPUT!$ T)))
  (COND ((NOT OFL) (REMFLAG '(FLUID) '!$!$FLG))) )

% Try to find out if functions are implemented as values.
(SETQ FNASVALU!$ NIL)
(PROG (OFL)
  (DE !$!$TST (X) X)
  (COND ((BOUNDP '!$!$TST) (SETQ FNASVALU!$ T)))
  (REMD '!$!$TST) )

(COND
  ((NOT (GETD 'FOREACH))
   (DEFLIST
    '((DO MAPC) (COLLECT MAPCAR))
    'IN)
   (DM FOREACH (ARGL)
     (PROG (ARG ONKEY L DOKEY BODY MFN)
       (SETQ ARGL (CDR ARGL))
       (COND
         ((NULL (SETQ MFN (GET (SETQ DOKEY (CADDDR ARGL))
                               (SETQ ONKEY (CADR ARGL)) )))
          (ERROR 410
           (LIST ONKEY " and " DOKEY " is unknown FOREACH combination"
            ) ) ))
       (SETQ ARG (LIST (CAR ARGL)))
       (RETURN (LIST MFN (SETQ L (CADDR ARGL))
         (COND
          ((NOT (AND (NULL (CDR (SETQ BODY (CDDDDR ARGL))))
                     (PAIRP (CAR BODY))
                     (EQUAL (CDAR BODY) ARG) ))
           (LIST 'FUNCTION (LIST 'LAMBDA ARG (CONS 'PROGN BODY) )) )
          ((IDP (SETQ BODY (CAAR BODY))) (LIST 'FUNCTION BODY)) % JEA
          (T BODY) ) )) ) ) ))

(FOREACH !$X IN !$!$OBL DO
  (COND
    ((NULL (SETQ !$Y (GETD !$X))) NIL)
    ((PROGN (XPUSH !$FUNS!$ !$X)
       (MEMQ (SETQ !$Y (CAR !$Y)) '(EXPR SUBR)) )
     (XPUSH !$EXPR!$ !$X) )
    ((MEMQ !$Y '(FEXPR FSUBR)) (XPUSH !$FEXPR!$ !$X))
    ((EQ !$Y 'NEXPR) (XPUSH !$NEXPR!$ !$X))
    ((EQ !$Y 'MMACRO) (XPUSH !$MMACRO!$ !$X))
    ((EQ !$Y 'MACRO) (XPUSH !$MACRO!$ !$X))
    (T (PRINT !$Y) (ERROR 0 "Unknown function type")) )
  (COND
    ((AND (BOUNDP !$X) (NOT (AND FNASVALU!$ !$Y)))
      (XPUSH !$VALUE!$ !$X)
      (COND (!$Y (XPUSH !$BOTH!$ !$X))) )
    ((NOT !$Y) (XPUSH !$OTHER!$ !$X)))
  (FOREACH !$Z IN (PLIST !$X) DO
    (COND
     ((NOT (PAIRP !$Z))
       (COND
         (FLAGASPUT!$ (ERROR 0 "Bad Property List"))
         ((EQ !$Z 'FLUID) (XPUSH !$FLUID!$ !$X))
         ((EQ !$Z 'GLOBAL) (XPUSH !$GLOBAL!$ !$X))
         ((MEMQ !$Z !$FLAG!$)
           (SETQ !$Y (GET !$Z 'HAS!$FLAG))
           (RPLACD (CAR !$Y) (SETQ !$Z (NCONS !$X)))
           (RPLACA !$Y !$Z) )
         (T (XPUSH !$FLAG!$ !$Z)
            (SETQ !$Y (NCONS !$X))
            (PUT !$Z 'HAS!$FLAG (CONS !$Y !$Y)) ) ))
     ((EQ (SETQ !$P (CAR !$Z)) 'FLUID) (XPUSH !$FLUID!$ !$X))
     ((EQ !$P 'GLOBAL) (XPUSH !$GLOBAL!$ !$X))
     ((EQ !$P 'VARTYPE)                 % This is for PSL.
      (COND
        ((EQ (CDR !$Z) 'FLUID) (XPUSH !$FLUID!$ !$X))
        ((EQ (CDR !$Z) 'GLOBAL) (XPUSH !$GLOBAL!$ !$X))
        (T (PRINT (CDR !$Z)) (ERROR 0 "Unknown VARTYPE")) ) )
     ((MEMQ !$P !$FLAG!$)
       (SETQ !$Y (GET !$P 'HAS!$FLAG))
       (RPLACD (CAR !$Y) (SETQ !$P (NCONS !$X)))
       (RPLACA !$Y !$P) )
% We have better check that all property values are T!
     ((EQ (CDR !$Z) 'T)
       (XPUSH !$FLAG!$ !$P)
       (SETQ !$Y (NCONS !$X))
       (PUT !$P 'HAS!$FLAG (CONS !$Y !$Y)) )
     ((MEMQ !$P '(HAS!$FLAG HAS!$PROP)) NIL)
     ((NOT (IDP !$P))
       (COND
         ((MEMQ !$P !$NIDP!$)
           (SETQ !$Y (GET 'HAS!$NIDP !$P))
           (RPLACD (CAR !$Y) (SETQ !$P (NCONS !$X)))
           (RPLACA !$Y !$P) )
         (T (XPUSH !$NIDP!$ !$P)
           (SETQ !$Y (NCONS !$X))
           (PUT 'HAS!$NIDP !$P (CONS !$Y !$Y)) ) ) )
     ((MEMQ !$P !$PROP!$)
       (SETQ !$Y (GET !$P 'HAS!$PROP))
       (RPLACD (CAR !$Y) (SETQ !$P (NCONS !$X)))
       (RPLACA !$Y !$P) )
     (T (XPUSH !$PROP!$ !$P)
        (SETQ !$Y (NCONS !$X))
        (PUT !$P 'HAS!$PROP (CONS !$Y !$Y)) ) )) )

(PROGN
 (FOREACH !$X IN
  '((!$FUNS!$ O!$FUNS!$ N!$FUNS!$)
    (!$EXPR!$ O!$EXPR!$ N!$EXPR!$)
    (!$FEXPR!$ O!$FEXPR!$ N!$FEXPR!$)
    (!$NEXPR!$ O!$NEXPR!$ N!$NEXPR!$)
    (!$MACRO!$ O!$MACRO!$ N!$MACRO!$)
    (!$MMACRO!$ O!$MMACRO!$ N!$MMACRO!$)
    (!$OTHER!$ O!$OTHER!$ N!$OTHER!$)
    (!$BOTH!$ O!$BOTH!$ N!$BOTH!$)
    (!$VALUE!$ O!$VALUE!$ N!$VALUE!$)
    (!$FLUID!$ O!$FLUID!$ N!$FLUID!$)
    (!$GLOBAL!$ O!$GLOBAL!$ N!$GLOBAL!$) )
  DO (SET (CAR !$X) (SETQ !$Y (REVERSE (EVAL (CAR !$X)))))
  (SET (CADDR !$X) (COND
    ((AND (BOUNDP (CADR !$X)) (SETQ !$Z (EVAL (CADR !$X))))
     (SETDIFFQ !$Y !$Z) )
    (T !$Y) )) )
 (FOREACH !$X IN
  '((!$PROP!$ O!$PROP!$ N!$PROP!$ HAS!$PROP !$PROPH!$ O!$PROPH!$
     N!$PROPH!$ )
    (!$FLAG!$ O!$FLAG!$ N!$FLAG!$ HAS!$FLAG !$FLAGH!$ O!$FLAGH!$
     N!$FLAGH!$ ) )
  DO (SET (CAR !$X) (SETQ !$Y (MYSORT (EVAL (CAR !$X)))))
  (SET (CADDR !$X) (COND
    ((AND (BOUNDP (CADR !$X)) (SETQ !$Z (EVAL (CADR !$X))))
     (SETDIFFQ !$Y !$Z) )
    (T !$Y) ))
  (SETQ !$X (CDDDR !$X))
  (SET (CADR !$X)
   (FOREACH !$W IN !$Y COLLECT (RPLACA (GET !$W (CAR !$X)) !$W)) ) )
 (FOREACH !$X IN
  '((!$NIDP!$ O!$NIDP!$ N!$NIDP!$ HAS!$NIDP !$NIDPH!$ O!$NIDPH!$
     N!$NIDPH!$ ) )
  DO (SET (CAR !$X) (SETQ !$Y (MYSORT (EVAL (CAR !$X)))))
  (SET (CADDR !$X) (COND
    ((AND (BOUNDP (CADR !$X)) (SETQ !$Z (EVAL (CADR !$X))))
     (SETDIFFQ !$Y !$Z) )
    (T !$Y) ))
  (SETQ !$X (CDDDR !$X))
  (SET (CADR !$X)
   (FOREACH !$W IN !$Y COLLECT (RPLACA (GET (CAR !$X) !$W) !$W) ) ) ) )

(DE L!$PRIN (!$X)
  (PRIN2 " (" ) (PRIN1 (LENGTH !$X)) (PRIN2 ") :" ) (TERPRI) (TERPRI)
  (FOREACH !$Y IN !$X DO (PRIN1 !$Y) (PRIN2 " "))
  (TERPRI) (TERPRI) )

(PROGN
  (TERPRI) (TERPRI) (PRIN2 "   IDs") (L!$PRIN !$!$OBL)
  (FOREACH !$X IN
   '((N!$FUNS!$ . "Functions") (N!$EXPR!$ . "EXPRs")
     (N!$FEXPR!$ . "FEXPRs") (N!$NEXPR!$ . "NEXPRs")
     (N!$MACRO!$ . "MACROs") (N!$MMACRO!$ . "MMACROs")
     (N!$OTHER!$ . "OTHERs (Neither Functions nor VALUEs)")
     (N!$BOTH!$ . "BOTH Functions and Values")
     (N!$VALUE!$ . "VALUEs")
     (N!$FLUID!$ . "FLUIDs") (N!$GLOBAL!$ . "GLOBALs") )
   DO (TERPRI) (PRIN2 "   ") (PRIN2 (CDR !$X))
      (L!$PRIN (EVAL (CAR !$X))) )
  (FOREACH !$X IN
    '(("Properties" N!$PROP!$ "Property " !$PROPH!$)
      ("Non id properties" N!$NIDP!$ "Property " !$NIDPH!$)
      ("Flags" N!$FLAG!$ "Flagged " !$FLAGH!$) )
   DO (TERPRI) (PRIN2 "   ") (PRIN2 (CAR !$X))
      (L!$PRIN (EVAL (CADR !$X)))
      (SETQ !$X (CDDR !$X))
      (FOREACH !$Y IN (EVAL (CADR !$X))
       DO (PRIN2 "   ") (PRIN2 (CAR !$X)) (PRIN1 (CAR !$Y))
          (L!$PRIN (CDR !$Y)) ) ) )

%### OBLANL.SL end
