%### LOAD.SL ========================================================
%
%       LOAD.SL
%       -------
%
%       Copyright (c) 1988 by Jan E. Aman, Inge Frick, Lars Hornfeldt,
%       Institute of Theoretical Physics, University of Stockholm,
%       Vanadisvagen 9, S-113 46  Stockholm, Sweden.
%
%       This file contains some changes to LOADI.SL that adapts it to
%       PSL (VMS and Unix) and to Cambridge University Lisp (Unix and
%       Atari TOS).
%
%====================================================================

(SETFLUIDS !*KEEPCASE CURRENTDIR!* FIL!* (!*MODULCOMP T))

(IFP (EQ 'VMS (SYSTYPE 'OS))
 (SETGLOBALS
  (HOMEDIR!* ("sys$login:")) )          % VMS Login directory.
 (SETGLOBALS
  (HOMEDIR!* (!~)) ) )                  % Unix Login directory.

(IFP (EQ 'VMS (SYSTYPE 'OS))
 (IFP (GETD 'MACHINENAME)
  (SETGLOBALS
   (LISPLIBDIR!* ("rfasl:" "rpsl:")) )  % PSL library directorys.
  (SETGLOBALS
   (LISPLIBDIR!* ("rfasl:" "pl:")) ) )  % PSL library directorys.
 (IFP (EQ 'PSL (SYSTYPE 'LISPSYSTEM))
  (SETGLOBALS
   (LISPLIBDIR!* ("$pl")) ) ) )         % PSL library directory.

(IFP (EQ 'VMS (SYSTYPE 'OS))
 (CPROGN
  (SETGLOBALS (BEGINDIR!* ![) (ENDDIR!* !]) (DIRSEPR!* !.))
  (DE ENDDIRP (CH) (MEMQ CH '(!] !>))) )        % [
 (CPROGN
  (IFP (EQ 'ATARI (SYSTYPE 'MACHINE))
   (SETGLOBALS (BEGINDIR!* !\) (ENDDIR!* !\) (DIRSEPR!* !\))
                        % This is for Unix:
   (SETGLOBALS (BEGINDIR!* !/) (ENDDIR!* !/) (DIRSEPR!* !/)) )
  (DE ENDDIRP (CH) (EQ CH ENDDIR!*)) ) )

% GETLOGICAL returns the value of the logical variable X or NIL if
% there is no value.
% GETLOGICAL can not be properly defined in the VMS version of PSL,
% nor for the Atari or old versions of Cambridge University Lisp.
% For newer versions of Culisp:
(IFP (FNP 'GET!-UNIX!-ENV) (SAMEDEF GETLOGICAL GET!-UNIX!-ENV))
% For Unix-PSL:
(IFP (FNP 'GETENV) (SAMEDEF GETLOGICAL GETENV))

(DE ADDEXTENSION (X Y)
  (MAKESTR
    (NCONC (EXPLODEC X)
      (COND
        ((EQ Y '! ) (NCONS '!.))
        (Y (CONS '!. (EXPLODEC Y))) ))))

% ABSDIRP determines if directory/file specification is absolute rather
% than relative to a given (e.g. current) directory.
(DE ABSDIRP (W)
 (COND
   ((OR
     (EQCAR W '!/)                      % Unix style absolute dir. spec.
     (MEMQ '!: W)                       % VMS or Atari device.
     (AND
      (SETQ W (OR (MEMQ '![ W) (MEMQ '!< W)))   % ]  VMS style
      (NOT (OR (EQCAR (CDR W) '!.) (EQCAR (CDR W) '!-))) ) )
    W ) ) )

(DE ADDSUBDIR (X Y)
  (SLET (Z)
    (COND
      ((NULL Y) X)
      ((ABSDIRP (SETQ Z (EXPLODEC Y))) (MAKESTR (ISTDIZEFNAM Z)))
      (T (MAKESTR (IADDSUBDIR X (ISTDIZEDIR Z)))) ) ) )

% ADDPREFIX Add prefix X to file specification Y.
(DE ADDPREFIX (X Y)
 (SLET ((W (STDIZEFNAM Y)) Z)
  (COND
    ((MEMQ ENDDIR!* W) (SETQ X (IADDSUBDIR X W)))
    ((EQ X '! ) (SETQ X (CONS BEGINDIR!* W)))   % Unix only.
    ((NULL X) (SETQ X W))
    (T
     (SETQ Z (LAST2PAIR (SETQ X (STDIZEDIR X))))
     (RPLACD
      (IFP (EQ 'VMS (SYSTYPE 'OS))
       (CDR Z)
       (COND ((AND (CDR Z) (NOT (ENDDIRP (CADR Z)))) (CDR Z)) (T Z)) )
      (IFP (EQ 'VMS (SYSTYPE 'OS)) W (CONS DIRSEPR!* W)) ) ) )
  (MAKESTR X) ) )

(DE IADDSUBDIR (X Y)
 (SLET ((N 0) (W Y) (V Y) Z U)
  (IFP (EQ 'VMS (SYSTYPE 'OS))
    (COND
      ((EQCAR W BEGINDIR!*)
       (SETQ V (SETQ W (CDR W)))
       (WHILE                           % Strip [-. -header(s):
        (EQCAR W '!-)
        (SETQ W
         (COND
           ((AND (PAIRP (CDR W))
                 (OR (EQ (CADR W) '!.) (ENDDIRP (CADR W))) )
            (CDDR W) )
           (T (CDR W)) ) )
        (IINCR N) ) ) )
    (WHILE
     (AND                               % Strip ../ -header(s):
      (EQCAR W '!.)
      (EQCAR (CDR W) '!.) )
     (SETQ W
      (COND
       ((EQCAR (CDDR W) DIRSEPR!*) (CDDDR W))
       (T (CDDR W)) ) )
     (IINCR N) ) )
  (COND
    ((AND (NOT (AZEROP N)) (SETQ Z (OR X CURRENTDIR!*)))
     (COND
       ((SETQ U (UPDIR Z N)) (SETQ X U))
       (T (SETQ X (STDIZEDIR X)) (SETQ W V)) ) )
    (X (SETQ X (STDIZEDIR X))) )
  (COND
    ((EQ X '! ) (CONS BEGINDIR!* Y))    % Unix only.
    ((NULL X) Y)
    (T
     (SETQ Z (LAST2PAIR X))
     (RPLACD
      (COND
        ((AND
          (CDR Z)
          (NOT
           (IFP (EQ 'VMS (SYSTYPE 'OS))
             (AND (ENDDIRP (CADR Z)) (MEMQ ENDDIR!* W))
             (ENDDIRP (CADR Z)) ) ) )
         (CDR Z) )
        (T Z) )
      (IFP (EQ 'VMS (SYSTYPE 'OS))
       (COND
        ((MEMQ ENDDIR!* W)
         (SUREHEAD
          (COND ((EQ (CADR Z) '!:) BEGINDIR!*) (T DIRSEPR!*))
          W ) )
        (T W) )
       (SUREHEAD DIRSEPR!* W) ) )
     (IFP (EQ 'VMS (SYSTYPE 'OS)) NIL
      (COND ((ENDDIRP (CADR (SETQ Z (LAST2PAIR X)))) (RPLACD Z NIL))) )
      X ) ) ) )

(DE SUREHEAD (X L)      %:b Ensure L start with X %LH
  (COND
    ((EQCAR L X) L)     % JEA
    (T (CONS X L)) ) )

(DE UPDIR (U N)
 (SLET (V W X K)
  (COND
   ((NULL U)
    (ERRORFT 443 "No parent directory in UPDIR, set CURRENTDIR!* ?") ))
  (SETQ W (CONS NIL (SETQ V (STDIZEDIR U))))
  (SETQ K (MINUS N))
  (WHILE 
   (IFP (EQ 'VMS (SYSTYPE 'OS))
    (AND
     (SETQ X (MEMQ DIRSEPR!* (CDR W)))
     (CDR X)
     (MEMQ ENDDIR!* (CDR W)) )
    (AND (SETQ X (MEMQ DIRSEPR!* (CDR W))) (CDR X)) )
   (SETQ W X) (IINCR K) )
  (COND
   ((IMINUSP K) NIL)
%   (ERRORFT 444 "Can't go up %l steps from %l" N U)
   (T
    (SETQ W (MEM2Q DIRSEPR!* (CONS NIL V)))
    (WHILE
     (NOT (IMINUSP (IDECR K)))
     (SETQ W (MEM2Q DIRSEPR!* (CDR W))) )
    (IFP (EQ 'VMS (SYSTYPE 'OS))
%    (PROGN (RPLACA (CDR W) ENDDIR!*) (RPLACD (CDR W) NIL))
     (RPLAD (CDR W) ENDDIR!* NIL)    %LH
     (RPLACD (COND ((EQ (CDR W) V) V) (T W)) NIL) )
    V ) ) ))

(DE STDIZEDIR (Y) (ISTDIZEDIR (EXPLODEC Y)))

(IFP (EQ 'VMS (SYSTYPE 'OS))
  (DE ISTDIZEDIR (Y)
    (ISTDIZEFNAM
      (COND
        ((NOT (MEMQ (CAR (LASTPAIR Y)) '(!/ !] !> !:))) % [ Any dir. end
         (NCONC1 Y '!/) )       % Ensure interpretation as directory.
        (T Y) ) ) )
  (DE ISTDIZEDIR (Y) (ISTDIZEFNAM Y)) )

(DE STDIZEFNAM (W) (ISTDIZEFNAM (EXPLODEC W)))

(IFP (EQ 'VMS (SYSTYPE 'OS))
% This version of STDIZEFNAM used to be known as VMSIFY.
 (DE ISTDIZEFNAM (W)
  (SLET (Z)
%  Replace <> with []:
   (COND ((SETQ Z (MEMQ '!< W)) (RPLACA Z '![ )))       % ]
   (COND ((SETQ Z (MEMQ '!> W)) (RPLACA Z '!] )))       % [
%  Remove ./ at the beginning.
   (COND ((AND (EQCAR W '!.) (EQCAR (CDR W) '!/)) (SETQ W (CDDR W))))
   (SETQ W
    (COND
     ((EQCAR W '!~)             % ~ stands for home directory.
      (COND
       ((NULL (CDR W)) (EXPLODEC (CAR HOMEDIR!*)))
       ((NOT (EQCAR (CDR W) '!/))       % We can't handle this case of
         (CONS '![ (CDR W)) )           % ~user/ but this may work.
       ((MEMQ '!/ (CDDR W))             % We can't handle ~/subdir/file
        (CONSA '![ '!. (CDDR W)) )      % without GETLOGICAL.
       (T
        (NCONC
         (EXPLODEC (CAR HOMEDIR!*))
         (COND ((CDR W) (CDDR W))) ) ) ) )
     ((EQCAR W '!/)
      (COND
        ((SETQ Z (MEMQ '!/ (CDR W)))
         (RPLACA Z '!:)
         (COND
           ((MEMQ '!/ (CDR Z))
            (RPLACD Z (CONS '![ (CDR Z))) ))))          % ]
      (CDR W) )
     ((AND (EQCAR W '!.) (EQCAR (CDR W) '!.))
      (CONSA '![ '!- (CDDR W)) )                        % ]
     ((MEMQ '!/ W)
      (CONSA '![ '!. W) )                               % ]
     (T W) ) )
   (SETQ Z W)
   (WHILE                % Replace all other ".." with "-" .
    (AND (SETQ Z (MEMQ '!. Z)) (EQCAR (CDR Z) '!.))
%   (RPLACA Z '!-) (RPLACD Z (CDDR Z)) (SETQ Z (CDR Z))
    (RPLAD Z '!- (CDDR Z)) (SETQ Z (CDR Z)) )  %LH
   (SETQ Z W) 
   (WHILE                % Replace all "/" but the last with "." .
    (SETQ Z (MEMQ '!/ Z))
    (RPLACA Z (COND ((MEMQ '!/ (CDR Z)) '!. ) (T '!] ))) )      % [
   W ) )   
 (IFP (EQ 'ATARI (SYSTYPE 'MACHINE))
  (DE ISTDIZEFNAM (W)
   (SLET ((U W))
    (WHILE U (COND ((EQ (CAR U) '!/) (RPLACA U '!\))) (UNPUSH U))
    (COND ((AND (EQCAR W '!\) (CDR W) (EQCAR (CDDR W) '!\))
           (RPLACA W (CADR W)) (RPLACA (CDR W) '!:) ))
    W) )
% For Unix: STDIZEFNAM used to be known as UNIXIFY.
% This is for Unix PSL and for Culisp versions where GETLOGICAL
% is useful:
  (IFP (FNP 'GETLOGICAL)
   (DE ISTDIZEFNAM (W)
    (SLET (Y)
     (COND
      ((AND
        (EQCAR W '!~)           % ~ stands for home directory.
        (OR (NULL (CDR W)) (EQ (CADR W) '!/ ))
        (SETQ Y (EXPLODEC (GETLOGICAL "HOME"))) )
       (COND ((CDR W) (RPLACD (LASTPAIR Y) (CDR W))))
       Y )
      (T W)) ) )
% This version for the case of no useful GETLOGICAL (old Culisp
% versions):
   (DE ISTDIZEFNAM (W) W) )
   ) )

% FIL is a real file description with no special interpretation.
% PREFIXOF returns the prefix part of FIL, or NIL if there is no
% prefix. Both absolute and relative prefixes are returned.
% This function is necessarily system-dependent as it has to know
% how files are described.
(DE PREFIXOF (FIL)
 (SLET ((X (CONS NIL (SETQ FIL (EXPLODEC FIL)))) Y SW)
  (IFP (EQ 'VMS (SYSTYPE 'OS))
   (PROGN
    (WHILE
     (SETQ Y (MEM2Q '!: (CDR X)))
     (SETQ X Y) (SETQ SW T) )           % VMS prefix.
    (COND ((SETQ Y (OR (MEM2Q '!> (CDR X)) (MEM2Q '!] (CDR X)))) % [
           (SETQ X Y) (SETQ SW T) ))
    (SETQ X (CDR X)) )
   (WHILE
    (SETQ Y (MEM2Q DIRSEPR!* (CDR X)))
    (SETQ X Y) (SETQ SW T) ) )          % Unix prefix.
  (COND
   (SW (RPLACD X NIL) (MAKESTR FIL))
   ((EQ (CAR FIL) BEGINDIR!*) '! ) ) ))         % Unix style only.

% FIL is a file description. FILNAM takes FIL apart and returns
% (<prefix> <file name> . <extension>) where <file name> and
% <extension> are interned ids and <prefix> is either a string an id
% or a search list. When some part is missing, it is returned as NIL.
% Various special system dependent interpretations of FIL might be
% done, in which case the fluid variable FIL!* is updated. A relative
% prefix is returned as NIL. This function is necessarily
% system-dependent as it has to know how files are described.
(DE FILNAM (FIL)
 (SLET (PX NM EX X Y)
  (SETQ FIL
   (STDIZEFNAM (IFP (FNP 'GETLOGICAL) (OR (GETLOGICAL FIL) FIL) FIL)) )
  (COND ((IFP (EQ 'VMS (SYSTYPE 'OS))
          (PROGN                        % VMS
           (SETQ Y (CONS NIL (CONS NIL FIL)))
           (WHILE
            (SETQ X (MEM2Q '!: (CDR Y)))% Find last ":".
            (SETQ Y X) )
           (AND Y (EQCAR (CDR Y) '!:) (NOT (EQCAR Y '!:))) )
          (AND                          % Unix
           (EQ (CAR FIL) DIRSEPR!* )            % Try to get path
           (SETQ Y (MEM2Q DIRSEPR!* FIL)) ) )   % from /name/
         (SETQ X (CDR Y)) (RPLACD Y NIL)
         (COND
          ((SETQ PX (GET (ULCASE (CONCATIDL FIL) !*SHPCASEUP) 'PATH))
           (SETQ FIL (CDR X)) )
          (T (RPLACD Y X)) ) ))
  (SETQ FIL!* (MAKESTR FIL))
  NIL
  (COND
    ((NOT !*KEEPCASE)
     (SETQ FIL
      (EXPLODEC (SETQ FIL!* (ULCASE FIL!* !*OSCASEUP))) ) ) )
  (SETQ X NIL)
  (SETQ NM FIL)
  (IFP (EQ 'VMS (SYSTYPE 'OS))
    (COND
      ((PAIRP (SETQ X (MEMQ ENDDIR!* NM)))      % VMS prefix
       (SETQ NM (CDR X)) ) )
    (WHILE
     (SETQ Y (MEMQ DIRSEPR!* NM))       % Unix prefix
     (SETQ X Y) (SETQ NM (CDR Y)) ) )
  (COND
    (X
     (RPLACD X NIL)
     (COND
       ((ABSDIRP FIL)
        (SETQ PX (MAKESTR FIL)) )       % Absolute prefix
       ((AND PX (NOT (PAIRP PX)))
        (SETQ PX (ADDPREFIX PX (MAKESTR FIL))) ) )
     (RPLACD X NM) ))
  (COND
    ((NULL NM) NIL)                     % No file name and no extension
    ((EQ (CAR NM) '!.)
     (SETQ EX NM) (SETQ NM NIL) )       % Extension but no file name
    ((SETQ Y (MEM2Q '!. NM))
     (SETQ EX (CDR Y))
     (RPLACD Y NIL) ) )
  (CONS PX
   (CONS
    (COND (NM (ULCASE (CONCATIDL NM) !*SHPCASEUP)))
    (COND
      (EX
       (COND
         ((CDR EX)
          (ULCASE (CONCATIDL (CDR EX)) !*SHPCASEUP) )
         (T '! ) ) )) )) ))

% Indicate loading function for files depending on extension.
(XDEFLIST INFN (SL READIN) (LAP LAPIN) (B FASLIN))

% Save original PSL versions. Not to be relied upon.
(IFP (EQ 'PSL (SYSTYPE 'LISPSYSTEM))
 (PROGN
  (COND ((NOT (FNP 'ILOAD)) (MOVD 'ILOAD 'LOAD)))
  (COND ((NOT (FNP 'IRELOAD)) (MOVD 'IRELOAD 'RELOAD)))
  (COND ((NOT (FNP 'IDSKIN)) (MOVD 'IDSKIN 'DSKIN))) ))

(DE SETCASE (X)
 (COND
   ((STRINGP X) X)
   (T (MAKESTR (EXPLODEC (ULCASE X !*OSCASEUP)))) ) )

(FLAG '(MODULOAD SETCASE) 'LOSE)        % Defined in the present file.

(INCLUDE "../loadi.sl")

(REMFLAG '(MODULOAD SETCASE) 'LOSE)     % Defined in the present file.

(IFP (EQ 'PSL (SYSTYPE 'LISPSYSTEM))
 (CPROGN
  (SETQ FASLOUTEXT!* 'B)        % Extension for created fasl files.
  (DPUT FLOADEXT SUBDRS (LAP B)) % Extensions for fast loading.
  (SETQ DEFAULTEXTENSION!* 'B)
% Pseudo files are not yet implemented for PSL.
% Probably not very important to do.
% (DPUT PSEUDO!-FILE LOADFN PSEUDO!-FILEFN)
% (DC PSEUDO!-FILE INFN (X DUMMY) (EVAL (LIST X)))
  )
 (IFP (EQ 'CULISP (SYSTYPE 'LISPSYSTEM))
  (CPROGN
   (SETQ FASLOUTEXT!* 'K)       % Extension for created fasl files.
   (DPUT FLOADEXT SUBDRS NONE)  % Extension for fast loading,
                                % currently not used in Culisp.
   (SETQ DEFAULTEXTENSION!* 'PSEUDO!-FILE)
   (DPUT PSEUDO!-FILE LOADFN PSEUDO!-FILEFN)
   (DC PSEUDO!-FILE INFN (X DUMMY) (EVAL (LIST X)))
   )))

% For Culisp define FASLOUT and FASLEND to be MODULE and ENDMODULE
% respectively.
(IFP (EQ 'CULISP (SYSTYPE 'LISPSYSTEM))
 (CPROGN
% (SAMEDEF FASLOUT MODULE)
% (SAMEDEF FASLEND ENDMODULE)
% SAMEDEF can not be used as the FUNCALL in UNBINDSFSTKS can not handle
% the macros defined by SAMEDEF in the same way.
  (MOVD 'FASLOUT 'MODULE)
  (MOVD 'FASLEND 'ENDMODULE) ))

(SETGLOBALS MODULESTK!*)
(COND ((NOT (MEMQ 'MODULESTK!* SFSTKS!*))
       (SETQ SFSTKS!* (NCONC1 SFSTKS!* 'MODULESTK!*)) ))
% (FLAG '(MODULESTK!*) 'LASTONLY)
(DEFLIST '((MODULESTK!* FASLEND)) 'SFSTKFN)

% MODULOAD is used as a machine independent command to create a module.
% It means for Culisp to create a Culisp module with contents from
% specified files, otherwise normally to load and compile files.
% Culisp version of MODULOAD originally by LH.
(IFP (EQ 'CULISP (SYSTYPE 'LISPSYSTEM))
 (DMACRO MODULOAD (FL1 !&REST FLL)
  (LIST 'IMODULOAD
   (MKQUOTE (CADR (FILNAM FL1)))
   (MKQUOTE (CONS FL1 FLL))) )
 (DS MODULOAD (!&REST FLL) (IMODULOAD NIL 'FLL)) )

% IMODULOAD is an expr version with module name as first argument.
(DE IMODULOAD (MN FLL)
 (SLET
  ((EX (DELETE FASLOUTEXT!* (DELETE 'PSEUDO!-FILE MEXTLIS!*))))
  (XPUSH COMPFLGSTK!* !*COMP)
  (SETQ !*COMP (OR !*COMP !*MODULCOMP))
  (IFP (EQ 'CULISP (SYSTYPE 'LISPSYSTEM))
    (COND (!*COMP (XPUSH MODULESTK!* NIL) (FASLOUT MN))) )
  (FOREACH FIL IN FLL DO (LOADX FIL MSEARCHLIS!* EX NIL NIL))
  (IFP (EQ 'CULISP (SYSTYPE 'LISPSYSTEM))
    (COND (!*COMP (FASLEND) (UNPUSH MODULESTK!*))) )
  (SETQ !*COMP (CAR COMPFLGSTK!*))
  (SETQ COMPFLGSTK!* (CDR COMPFLGSTK!*)) ) )

%### LOAD.SL end
