%### EXCOMI.SL ======================================================
%
%       EXCOMI.SL
%       ---------
%
%       Copyright (c) 1982 by Inge Frick, Institute of Theoretical
%       Physics, Vanadisvagen 9, S-113 46 Stockholm, Sweden.
%
%       This file contains definitions of some extensions to Standard
%       Lisp. This file is intended to be present when compiling
%       Sheep Lisp, it defines most of the extensions as
%       MACROs. FEXPRs defining the same functions can be found in
%       EXINTI.SL which is intended for the interpreter. The correct
%       way to handle this is of course to only have the interpreter
%       version, and replace the macros by COMPILERMACROs, COMPFNs,
%       PA1FNs or some other compiler construction. If this is
%       possible it is done by redefining CMMACRO in EXFN1I.SL.
%
%       The functions in this file have to be in a special order
%       because of bootstrapping.
%
%====================================================================

(FUNCTIONS REEVALFORM MKERRFORM NOFORMS !$ERRFORM !$ERRFL ERRFLAMBDA
           ERRFPROG ERRFSETQ ERRSET
           MAPRPLC MAPRPL MAPR!$FN REPEATS WHILE UNTIL PROG1 XCASE)

(ENTRYPOINTS REEVALFORM ERRSET
             MAPRPLC MAPRPL REPEATS WHILE UNTIL PROG1 XCASE)

(LOCALFUNCTIONS NOFORMS !$ERRFORM !$ERRFL ERRFLAMBDA ERRFPROG ERRFSETQ
                MAPR!$FN)

(USEDPROPERTIES ERRFORMFN)

(SETFLUIDS VBLS IVBLS)

% Create a form that can be evaluated again without reevaluating the
% arguments.
(CMMACRO REEVALFORM (!&REST L) (MKERRFORM L))

% CATCH and THROW is based on ideas by Rabbe Fogelholm and on ideas
% from PSL.

% This is a help routine to force arguments to a function to be
% evaluated before an ERRSET or CATCH. The purpose of this is that
% variables shouldn't have to be declared FLUID. This works for
% references to local variables, but it can't SETQ them.
(DE MKERRFORM (X)
 (SLET (VBLS IVBLS)
  (!$ERRFORM (SETQ X (MACEXPD X)))
  (COND
   ((OR (ATOM X) (NULL VBLS)) (MKQUOTE X))
   ((NOFORMS X)
    (CONSA 'LIST
     (MKQUOTE (CAR X))
     (FOREACH Y IN (CDR X) COLLECT (COND
       ((MEMQ Y VBLS) (LIST 'MKQUOTE Y))
       (T (MKQUOTE Y)) )) ) )
   (T (CONSA 'LIST
       (LIST 'FUNCTION (LIST 'LAMBDA VBLS X))
       (FOREACH Y IN VBLS COLLECT (LIST 'MKQUOTE Y)) )) ) ) )

(DE NOFORMS (L) (OR (NULL L) (AND (ATOM (CAR L)) (NOFORMS (CDR L)))))

(DE !$ERRFORM (X)
 (SLET (A)
  (COND
   ((ATOM X)
    (COND ((AND (IDP X)
            (NOT (OR (GLOBALP X)
                     (FLUIDP X)
                     (MEMQ X IVBLS)
                     (MEMQ X VBLS) )) )
           (XPUSH VBLS X) ))
    X)
   ((PAIRP (CAR X)) (!$ERRFL X))
   ((SETQ A (GET (CAR X) 'ERRFORMFN)) (FUNCALL A X))
   (T (!$ERRFL (CDR X))) ) ) )

(DE !$ERRFL (L) (FOREACH X IN L DO (!$ERRFORM X)))

(DE ERRFLAMBDA (X)
 (SLET ((IVBLS (APPEND (CADR X) IVBLS)))
  (!$ERRFL (CDDR X)) ) )

(DE ERRFPROG (X)
 (SLET ((IVBLS (APPEND (CADR X) IVBLS)))
  (FOREACH Y IN (CDDR X) DO (COND ((PAIRP Y) (!$ERRFORM Y)))) ) )

(DE ERRFSETQ (X)
 (!$ERRFORM (CADDR X))
 (SETQ X (CADR X))
 (COND ((NOT (OR (GLOBALP X) (FLUIDP X) (MEMQ X IVBLS)))
        (FLUID (LIST X))
        (WARNINGFT 30 "Variable %p MUST be declared FLUID " X)
        (REMPOP VBLS X) )) )

(XDEFLIST ERRFORMFN
 (LAMBDA ERRFLAMBDA) (PROG ERRFPROG) (QUOTE RETARG)  (GO RETARG)
 (SETQ ERRFSETQ) )

(CMMACRO ERRSET (BODY SW) (LIST
  (LIST 'LAMBDA '(!*ERRMSG SFER!&)
   (LIST 'UNBINDSFSTKS 'SFER!&
    (LIST 'ERRORSET (MKERRFORM BODY) '!*ERRMSG
     '(AND !*ERRMSG !*BACKTRACE) ) ) )
  SW '(CAPTURESFSTKS) ))

% MAPRPL and MAPRPLC replaces each element in a list and returns the
% list.
% The following are the interpreter versions:
%(DE MAPRPLC (L FN)
% (FOREACH X ON L DO (RPLACA X (FUNCALL FN (CAR X))))
% L)
%
%(DE MAPRPL (L FN)
% (FOREACH X ON L DO (RPLACA X (FUNCALL FN X)))
% L)
% It gets a little more complicated in the compiler version:
(CMMACRO MAPRPLC (L FN) (MAPR!$FN L FN T))

(CMMACRO MAPRPL (L FN) (MAPR!$FN L FN NIL))

% This is a help routine for the macro definition of MAPRPL and
% MAPRPLC.
(DE MAPR!$FN (U FN SW)
 (SLET (A (X (GENSYM)) (L (GENSYM)))
  (SETQ SW (COND (SW (LIST 'CAR X)) (T X)))
  (SETQ SW (COND
    ((AND (PAIRP FN) (MEMQ (CAR FN) '(FUNCTION QUOTE)))
     (LIST (CADR FN) SW) )
    (T (LIST 'FUNCALL (SETQ A (GENSYM)) SW)) ))
  (LIST 'SLET (CONS (LIST L U) (COND (A (LIST (LIST A FN)))))
   (LIST 'FOREACH X 'ON L 'DO (LIST 'RPLACA X SW))
   L) ) )

% REPEATS(COUNT:integer,[XPR:form]):NIL
% The XPRs are evaluated COUNT times.
(CMMACRO REPEATS (C !&REST BODY) (SLET ((A (GENSYM)))
  (LIST 'PROG (LIST A)
      (LIST 'SETQ A C)
 'LAB (LIST 'COND (LIST (LIST 'XMINUSP (LIST 'XDECR A)) '(RETURN NIL)))
      (CONS 'PROGN BODY)
      '(GO LAB) ) ))

(CMMACRO WHILE (TST !&REST BODY)
 (CONSA
  'PROG '()
   '(GO LAB2)
   'LAB1
   (APPEND BODY
    (LIST 'LAB2 (LIST 'COND (LIST TST '(GO LAB1)))) ) ) )

(COMMENT
(CMMACRO REPEAT (!&REST BODY)
 (LIST 'PROG '()
  'LAB (LIST 'COND (LIST (LIST 'NOT (MKPROGN BODY)) '(GO LAB))) ) ) )

(CMMACRO UNTIL (TST !&REST BODY)
 (CONSA 'PROG '()
  'LAB
  (APPEND1 BODY
   (COND
     (TST (LIST 'COND (LIST (LIST 'NOT TST) '(GO LAB))))
     (T '(GO LAB)) )) ) )

% PROG1([XPR:form]):any
% Evaluates the XPR from left to right and returns the value of the
% first XPR.
(CMMACRO PROG1 (X !&REST L) (SLET ((A (GENSYM)))
 (LIST 'SLET (NCONS (LIST A X)) (CONS 'PROGN L) A) ))

% XCASE is called as XCASE(X,Z,Y(0),Y(1),...,Y(n)).
% If the value of X is an integer in the range 0:n then Y(X) is
% evaluated, otherwise Z is evaluated.
% The interpreted version is as follows:
%(MACRO XCASE (CS D !&REST L)
% (PLET ((A (EVAL CS)))
%  (COND
%   ((AND (FIXP  A) (SETQ A (CNDR L A))) (CAR A))
%   (T D) ) ) )
% The following definition of XCASE will work, but it is much better
% if the compiler knows how to handle XCASE and produces a jump-table.
(CMMACRO XCASE (CS D !&REST L) (SLET ((N -1) SW TMP)
  (COND ((NOEVAGAINP CS) (SETQ SW CS) (SETQ CS 'XCASE!*!$ARG)))
  (SETQ TMP (CONS 'COND
    (NCONC1
     (FOREACH X IN L COLLECT
      (LIST (LIST 'AEQN CS (SETQ N (XADD1 N))) X) )
     (LIST 'T D) ) ))
  (COND
   (SW (LIST 'PROGN (LIST 'SETQ CS SW) TMP))
   (T TMP) ) ))

(UNFLUID '(VBLS IVBLS))

%### EXCOMI.SL end
