%### EXFN1I.SL ======================================================
%
%       EXFN1I.SL
%       ---------
%
%       Copyright Inge Frick, Institute of Theoretical Physics,
%       Vanadisvagen 9, S-113 46 Stockholm, Sweden
%
%
%       This file contains definitions of some extensions to Standard
%       Lisp. It is used together with EXFN2I.SL, EXINTI.SL and
%       EXCOMI.SL. EXFN1I.SL + EXCOMI.SL + EXFN2I.SL (in that order)
%       is used to define the extensions for the compiler. EXFN1I.SL +
%       EXFN2I.SL + EXINTI.SL (in that order) is used to define the
%       extensions to the interpreter.
%
%====================================================================

% Define the following FLAGS if they have not been defined already.
(COND ((NOT (FLAGP 'EXPR 'ISAFUNCTION))
       (PROGN                           % EXPR is always a function.
        (FLAG '(EXPR FEXPR MACRO) 'ISAFUNCTION) % Function types.
        % Functions of the following type get name as first arg.
        (FLAG '(MACRO) 'MACROARG)
        %The following must have one arg that is bound to list of args.
        (FLAG '(MACRO FEXPR) 'FNARY)
        )))

% COMMENT does absolutely nothing with its arguments.
% It is defined this early so that it can be used to comment out code.
(DM COMMENT (DUMMY) NIL)

(DM FUNCTIONS (DUMMY) NIL)

(FUNCTIONS
 COMMENT FUNCTIONS IFP WARNINGFT PUTD
 BOUNDP QSET MKQUOTE EQCAR QUOTEP ATSOC REVERSIP REVERSIPX REVERSEX
 NCONS XCONS RETARG XCOND FUNCALL
 MACEXPD MACEXPDL1 MACEXPDL MACEXPDCOND MACEXPDPRG
 XPUSH ERRORFT XERRORFT3
 STNIL APPEND1 NCONC1 RPLAD CONSIFCDR MKPROGN
 DMACRO !$MACFEX FCHKARG1 FCHKARG DSSB EVCONSTANTP
 DS CMMACRO DFEXPR DEXPR DFUN UNPUSH CONSA UNIONQ
 LOCALFUNCTIONS DE FOREACH SLET PLET ADDHLP
 MOVD SAMEDEF FOLDCASE
 CONCATID CONCATIDHELP PUTC DC XPUT IXPUT DPUT XDEFLIST XFLAGS
 GVALUE COMPILETIME NOTLOADTIME MACROTIME BOTHTIMES LOADTIME
 NOCOMP RESTORECOMP
 USEDPROPERTIES USEDFLAGS DECLAREFEXPR ENTRYPOINTS
 SETCONSTANTS SETGLOBALS SETLOCALGLOBALS SETFLUIDS
 QSETGLOBALS QSETLOCALGLOBALS QSETFLUIDS QSETQ SETGLBS
 SYSTYPE SETSYSTYPE)

%(LOCALFUNCTIONS CONCATIDHELP !$MACFEX MKPROGN)
% Can't be local functions as they are called by macros
% that are redefined to these interpreted functions.

(GLOBAL '(!$EOF!$ !$EOL!$))

(FLUID '(!*GC !*RAISE))

(GLOBAL '(!*COMP PUTCHR!* SFSTKS!* !*MACROTIME EMSG!* XCOND!* LISPFN
          SAVEDO!* !*SAVECOMP SYSTYPE!* !*RSHEEP))

(FLUID '(DSARGS!* !*ERRMSG HLPLIS!* LOCALFUNCTIONS!*
% !*ERRDEBUG THROWTAG!* THROWSIGNAL!*
         !*NOCHECK !*MACROEXPAND !*CREF !*DEFN !*BACKTRACE))

(SETQ HLPLIS!* 'LISPFN)

(SETQ PUTCHR!* '!.)

(SETQ !*ERRMSG T)

(SETQ !*BACKTRACE T)

% IFP(IFFORM:form,THENFORM:form,{ELSEFORM:form}):any
% IFP is used for conditional coding.
% IFFORM is evaluated at macro expansion time. If it is true then
% THENFORM is returned. If it is false then ELSEFORM (or if it is not
% present NIL) is returned.
(DM IFP (U!&) (COND
  ((EVAL (CADR U!&)) (CADDR U!&))
  ((CDDDR U!&) (CADDDR U!&)) ))

(IFP (NOT (GETD 'WARNINGFT))           % Needed for bootstrap
 (EVAL (PROGN
% WARNINGFT redefined to include warning message numbers not yet used
% but available for possible future use. Straight forward definition:
%(DS WARNINGFT (DUMMY ST !&REST AR) (MSGOUT ST "*** " (LIST . AR)))% JEA
% Safer definition accepting also calls without warning numbers:
   (DM WARNINGFT (U)
    (PROGN
     (SETQ U (CDR U))
     (COND ((AND U (NUMBERP (CAR U))) (SETQ U (CDR U)))) % JEA.
     (LIST 'MSGOUT "*** " (LIST 'QUOTE (CAR U)) (CONS 'LIST (CDR U))) ))
   (COND ((NOT (GETD 'MSGOUT))
          (DE MSGOUT (PR ST AR)
              (PROGN (PRIN2 PR) (PRINT (CONS ST AR)) NIL ) ) ))
   NIL )) )

% Redefine PUTD to look for the LOSE flag. JEA
(COND ((NOT (GETD 'I!$PUTD))
       (PUTD 'I!$PUTD (CAR (GETD 'PUTD)) (CDR (GETD 'PUTD))) ))

(DE PUTD (X Y Z)
 (COND
  ((FLAGP X 'LOSE) (WARNINGFT 36 "%p not defined - flagged LOSE" X) NIL)
  (T (I!$PUTD X Y
   (COND ((AND (NOT !*COMP) !*MACROEXPAND) (MACEXPD Z)) (T Z)) )) ) )

% A few general functions.

(IFP (NOT (FLAGP 'BOUNDP 'LOSE))
 (DE BOUNDP (X) (PAIRP (ERRORSET X NIL NIL))) )

(DE QSET (X Y) (COND ((NOT (BOUNDP X)) (SET X Y))))

(DE MKQUOTE (X) (LIST 'QUOTE X))

(DE EQCAR (X Y) (AND (PAIRP X) (EQ (CAR X) Y)))

(DE QUOTEP (X) (EQCAR X 'QUOTE))

(DE ATSOC (X Y) (COND
  ((NULL Y) NIL)
  ((EQ X (CAAR Y)) (CAR Y))
  (T (ATSOC X (CDR Y))) ))

(DE REVERSIP (L)
 (COND
  ((NOT (PAIRP L)) L)
  (T (REVERSIPX L NIL)) ))

(DE REVERSIPX (L Y)
 (PROG (X)
   (COND
    ((NULL L) (RETURN Y))
    ((NOT (PAIRP L)) (RETURN (CONS L Y))) )
 L (SETQ X (CDR L))
   (SETQ Y (RPLACD L Y))
   (COND ((SETQ L X) (GO L)))
   (RETURN Y) ) )

(DE REVERSEX (X Y) (COND
  (X (REVERSEX (CDR X) (CONS (CAR X) Y)))
  (T Y) ))

(DE NCONS (X) (CONS X NIL))

(DE XCONS (X Y) (CONS Y X))

(DE RETARG (X) X)       % A NOOP with one argument.

% XCOND the same as COND but it has an implied PROGN.
% XCOND should be used on a Lisp that doesn't have implied PROGN
% and/or no consequent. In the future COND is used, but it should be
% changed to XCOND for such a Lisp.
(DM XCOND (U)
 (PROG (X A)
  (SETQ U (CDR U))
O (COND
   (U (PROGN
       (SETQ A (CONS (COND
                      ((NULL (CDR (SETQ X (CAR U))))
                       (LIST (LIST 'SETQ 'XCOND!* (CAR X)) 'XCOND!*))
                      ((CDDR X) (LIST (CAR X) (CONS 'PROGN (CDR X))))
                      (T X) )
                     A))
       (SETQ U (CDR U))
       (GO O) ))
   ((NOT (AND A (EQ (CAAR A) T))) (SETQ A (CONS '(T NIL) A)))
   (T NIL) )
  (RETURN (CONS 'COND (REVERSIP A))) ) )

% (FUNCALL FN [ARGS]) is the same as (APPLY FN (LIST [ARGS]))
% FN must be an EXPR.
(IFP (NOT (FLAGP 'FUNCALL 'LOSE))
 (DM FUNCALL (U)
  (LIST 'APPLY (CADR U) (COND ((CDDR U) (CONS 'LIST (CDDR U))))))) % JEA

% MACEXPD(SXPR:any):any
% All macros in s-expression SXPR are expanded.
(DE MACEXPD (X)
 (PROG (A)
L (COND
   ((NOT (PAIRP X)) (RETURN X))
   ((PAIRP (CAR X)) (RETURN (MACEXPDL X)))
   ((SETQ A (GET (CAR X) 'MACEXPDFN)) (RETURN (FUNCALL A X)))
   ((EQUAL (SETQ A (MACEXPD1 X)) X) (RETURN (MACEXPDL1 X)))
   (T (SETQ X A)) )
  (GO L) ) )

% MACEXPD1(SXPR:any):any
% The top level of SXPR is macro-expanded if it is a macro form.
(DE MACEXPD1 (X)
 (PROG (A)
  (RETURN
   (COND
    ((AND (PAIRP X) (NOT (PAIRP (CAR X)))
          (EQCAR (SETQ A (GETD (CAR X))) 'MACRO) )
     (FUNCALL (CDR A) X) )
    (T X) ))))

(DE MACEXPDL1 (X) (CONS (CAR X) (MACEXPDL (CDR X))))

(DE MACEXPDL (L)
 (PROG (A)
G (COND ((NULL L) (RETURN (REVERSIP A))))
  (SETQ A (CONS (MACEXPD (CAR L)) A))
  (SETQ L (CDR L))
  (GO G) ) )

(DE MACEXPDCOND (L)
 (PROG (A)
G (COND ((NULL (SETQ L (CDR L))) (RETURN (CONS 'COND (REVERSIP A)))))
  (SETQ A (CONS (MACEXPDL (CAR L)) A))
  (GO G) ) )

(DE MACEXPDPRG (X) (CONS (CAR X) (MACEXPDL1 (CDR X))))

(IFP (NOT (FLAGP 'MACEXPD 'LOSE))
  (DEFLIST '((COND MACEXPDCOND) (LAMBDA MACEXPDPRG) (PROG MACEXPDPRG)
             (QUOTE RETARG) (GO RETARG) )
           'MACEXPDFN) )

% (XPUSH id any) pushes "any" onto the list that must be value of "id".
% "any" will be new beginning of the list. The value of XPUSH is not
% defined. XPUSH must be defined here as it is used in the general
% macro-defining function MACFEX.
(DM XPUSH (U) (LIST 'SETQ (CADR U) (LIST 'CONS (CADDR U) (CADR U))))

% Needed for bootstrap
(IFP (NOT (GETD 'ERRORFT))
 (EVAL (PROGN
   (DM ERRORFT (U)
    (LIST 'XERRORFT3 (CADR U) (CADDR U) (CONS 'LIST (CDDDR U))) )
   (COND ((NOT (GETD 'XERRORFT3))
          (DE XERRORFT3 (NM ST AR) (ERROR NM (CONS ST AR))) ))
   NIL )) )

(DE STNIL (X) (SET X NIL))      %:b Set identifier to NIL.
% Extending structures:
(DE APPEND1 (L X) (APPEND L (NCONS X))) %:b APPEND 1 element at end of L
(DE NCONC1  (L X) (NCONC  L (NCONS X))) %:b -ditto, but destructively
(DE RPLAD (L A D) (RPLACA (RPLACD L D) A))      %:b
% (RPLAD L A D) instead of (RPLACW L (CONS A D))

% Adds a prefix X to List, or take out the element of list L if alone:
(DE CONSIFCDR (X L)     %:b Cons if L has cdr, else car out L-element.
 (COND
  ((CDR L) (CONS X L))  % If L has cdr, cons in x.
  (T (CAR L)) ))        % Else car out the single element of L.

% This is a help routine to take care of implied progn in macro
% expansion.
(DE MKPROGN (L) (COND (L (CONSIFCDR 'PROGN L))))

% In the future DMACRO will be used instead of DM to define MACROs.
% DMACRO allows some of the MacLisp &keywords in the parameter list.
% Furthermore DMACRO (or rather !$MACFEX) can be changed to suit
% different systems way of defining MACROs.
(DM DMACRO (U) (PROGN
  (SETQ U (CDR U))
  (!$MACFEX (CAR U) (CADR U) (CDDR U) 'MACRO) ))

(DE !$MACFEX (FN ARGS OBODY TYP)
 (PROG
   (DSARGS!*    % FLUID. Set inside DSSB to variables know to DSSB.
    PARGS       % Variable list for PROG
    BODY        % Will contain the function body.
    HLP         % Help string.
    HLPLIS      % Help list name.
    LARG        % Last parameter or the functions LAMBDA args.
    ARGCD       % NIL or expression that steps down the argument list.
    REST        % Switch. True if processed &REST argument.
    SEEN        % Last seen keyword.
    MACR        % Indicates MACRO without !&MACRO keyword, so extra CDR
                % needed.
    DSB         % True if &DS keyword.
    AL          % Lower limit on number of arguments or NIL if
                % no arguments.
    AU          % Upper limit on number of arguments or NIL if
                % no upper limit.
    ARG         % Current variable or keyword.
    VAL         % NIL or default value form for &OPTIONAL argument.
    L           % Used to step down parameter list.
    X)          % Help variable.
   (SETQ MACR (FLAGP TYP 'MACROARG))  % Do we need to remove macro name?
   (SETQ L ARGS)
% !&DS or !&MACRO only at the beginning of the parameter list.
% Remove them.
I  (COND ((EQCAR L '!&DS)       % Check for &DS on parameter list.
          (SETQ DSB T) (SETQ L (CDR L)) ))
   (COND ((EQCAR L '!&MACRO)    % Check for &MACRO on parameter list.
          (COND
           (MACR (SETQ MACR NIL))       % We do want the macro name.
           (T (GO E)) )                 % Illegal if not a macro.
          (SETQ L (CDR L)) (GO I) )) % Go back and check for &DS again.
   (SETQ X L)   % L now points to parameter list with &DS and &MACRO
                % removed.
% First go through the parameter list to find the last argument (LARG)
%  and if variable nr of args.
R  (COND ((NULL X) (GO RE)))
   (SETQ ARG (CAR X))           % Look for keywords on parameter list.
   (COND
    ((MEMQ ARG '(!&OPTIONAL !&REST))    % Variable nr of args.
     (SETQ ARGCD T) )
    (T (SETQ LARG ARG)) )       % Not a keyword. Update last parameter.
   (SETQ X (CDR X))
   (GO R)
% Now see what we got.
% Are we doing a function with a fixed number of args, each bound to
% its parameter in a LAMBDA expression, or are are we doing an nary
% function with a single LAMBDA parameter that is bound to the list
% of arguments?
RE (COND                % Not variable nr of args or forced nary.
    ((NOT (OR ARGCD (FLAGP TYP 'FNARY)))
     (SETQ LARG NIL) )  % LARG will contain list of LAMBDA parameters.
    ((EQ TYP 'EXPR)     % An nary EXPR wouldn't be portable.
     (ERRORFT 400
      "%r is an EXPR, may not have variable nr of args" FN ) )
    (T (COND    % Nary, i.e. &OPTIONAL or &REST arg or forced nary.
        ((NULL LARG)
         (SETQ LARG (GENSYM)) ) % A FORCED NARY WITH NO ARGS.
        ((PAIRP LARG)
         (SETQ LARG (CAR LARG)) ) )     % OPTIONAL ARG WITH DEFAULT.
% LARG HAS NAME OF PARAMETER THAT WILL BE BOUND TO THE LIST OF ARGS.
       (SETQ ARGCD (LIST 'SETQ LARG (LIST 'CDR LARG))) ) )
% ARGCD STEPS DOWN.
% Loop through the parameter list, processing the parameters.
A  (COND ((NULL L) (GO AE)))
   (SETQ ARG (CAR L))
   (COND
    ((MEMQ ARG '(!&OPTIONAL !&REST !&MACRO !&DS))       % Keyword.
     (SETQ SEEN ARG) )
    % It is not a keyword, must be a parameter.
    ((NOT ARGCD) (XPUSH LARG ARG))      % Ordinary lambda arg.
    (REST (GO E))                       % &REST out of order.
    ((NULL SEEN)                        % Ordinary arg
     (SETQ AL (COND                     % Update lower limit.
       (AL (SETQ X ARGCD) (ADD1 AL))    % Earlier arg. CDR to next arg.
       (T (SETQ X LARG) 1) ))           % First arg.
     (XPUSH BODY (LIST 'SETQ ARG (LIST 'CAR X)))
     (XPUSH PARGS ARG) )
    ((EQ SEEN '!&OPTIONAL)              % Optional arg
     (SETQ AU (COND                     % Update upper limit.
       (AU                              % Earlier optional args.
        (SETQ X (LIST 'AND LARG ARGCD)) %  CDR to next arg if LARG
        (ADD1 AU) )                     % not NIL.
       (AL                              % Earlier ordinary args.
        (SETQ X ARGCD)                  %  CDR to next arg.
        (ADD1 AL) )
       (T (SETQ X LARG) 1) ))   % First arg. Don't need CDR first.
     (COND ((SETQ VAL (AND (PAIRP ARG) (CDR ARG))) % Default value?
            (SETQ ARG (CAR ARG))        % Yes. This is the parameter.
            (COND
             ((CDR VAL)                 % supplied-p parameter ?
              (XPUSH PARGS (CADR VAL))  % Yes. Put it in parameter list.
              (SETQ X (LIST 'SETQ (CADR VAL) X)) )) % Change predicate.
            (SETQ VAL (CAR VAL)) ))     % This is the default value.
     (XPUSH BODY (LIST 'SETQ ARG        % Set value of parameter
       (LIST 'COND
        (LIST X (LIST 'CAR LARG))       % if got arg, use it
        (LIST 'T VAL) ) ) )             % else use default.
     (XPUSH PARGS ARG) )
    ((EQ SEEN '!&REST) (COND            % REST parameter.
        (AU (XPUSH BODY                 % Optional arg before. Test if
             (LIST 'COND (LIST LARG ARGCD)) ))% possible to CDR to next.
        (AL (XPUSH BODY ARGCD)) )   % Ordinary arg before. CDR to next.
       (XPUSH PARGS (SETQ REST ARG))
       (SETQ AU NIL) )                  % No upper limit.
    (T (GO E)) )    % Must be !&DS or !&MACRO. They should not be here.
   (SETQ L (CDR L))
   (GO A)
% Get help information if there is any.
   (COND ((AND HLPLIS!* (CDR OBODY) (COND
            ((STRINGP (CAR OBODY)) (SETQ HLP (CAR OBODY)))
            ((EQCAR (CAR OBODY) 'HELP)
             (SETQ HLPLIS (CAR (SETQ HLP (CDAR OBODY))))
             (SETQ HLP (CADR HLP)) ) ))
          (SETQ OBODY (CDR OBODY)) ))
% Set DSARGS!* to point to list of all variables. DSARGS!* is used in
% DSSB below.
AE (COND
    ((NOT ARGCD)                        % Fixed nr of args.
     (COND (DSB (SETQ DSARGS!* (APPEND LARG PARGS)))) )
    ((EQCAR PARGS LARG)
     (COND (DSB (SETQ DSARGS!* PARGS)))
        % Remove the LAMBDA arg from the PROG args.
     (SETQ PARGS (CDR PARGS)) )
    (DSB (SETQ DSARGS!* (CONS LARG PARGS))) )
   (COND (DSB (SETQ OBODY (LIST (DSSB (MKPROGN OBODY))))))
B  (COND (OBODY
          (XPUSH BODY (CAR OBODY))
          (SETQ OBODY (CDR OBODY)) (GO B) ))
% If PROG, end with RETURN.
   (COND ((AND PARGS BODY)
          (RPLACA BODY (LIST 'RETURN (CAR BODY)))))
   (SETQ BODY (REVERSIP BODY))
   (SETQ FN (MKQUOTE FN))
% Check arguments.
   (COND ((AND ARGCD (NOT !*NOCHECK))   % If nary and !*NOCHECK
          (COND                         % off, check nr of args
           ((NOT (OR REST AU))          % Fixed nr of args.
            (XPUSH BODY (LIST 'FCHKARG1 LARG (OR AL 0) FN)) )
           ((OR AL AU)          % Upper limit or lower limit > 0.
            (XPUSH BODY (LIST 'FCHKARG LARG AL AU FN)) ) )))
% If necessary skip first argument.
   (COND ((AND MACR (OR (CDR BODY) (CAR BODY))) % Do we need a CDR
          (XPUSH BODY ARGCD) )) % first to skip the function name?
% Return a PUTD or PUTC expression.
   (SETQ BODY (LIST
     (COND
      ((FLAGP TYP 'ISAFUNCTION) 'PUTD)
      (T 'PUTC) )                       % Not function.
     FN
     (MKQUOTE TYP)
     (MKQUOTE (LIST 'LAMBDA
       (COND
        (ARGCD (NCONS LARG))            % Nary.
        (T (REVERSIP LARG)) )
       (COND
        (PARGS (CONS 'PROG (CONS (REVERSIP PARGS) BODY)))
        %((CDR BODY) (CONS 'PROGN BODY))
        %(T (CAR BODY))
        (T (MKPROGN BODY)) ) )) ))
   (RETURN (COND
     (HLP (LIST 'CPROGN
           (LIST 'ADDHLP (MKQUOTE (OR HLPLIS HLPLIS!*)) FN HLP)
           BODY ))
     (T BODY) ))
% Error return.
E  (ERRORFT 401 "%r illegal parameter list" ARGS) ) )

(DE FCHKARG1 (ARGL CN FN)
 (COND ((NOT (EQUAL (LENGTH ARGL) CN)) (FCHKARG ARGL CN CN FN))) )

(DE FCHKARG (ARGL MN MX FN) (PROGN
  (SETQ ARGL (LENGTH ARGL))
  (COND ((AND MN (LESSP ARGL MN))
         (ERRORFT 404 "Too few arguments for %r" FN) ))
  (COND ((AND MX (GREATERP ARGL MX))
         (ERRORFT 405 "Too many arguments for %r" FN) ))
  ARGL))

(DE DSSB (X) (COND
  ((AND (QUOTEP X) (PAIRP (CDR X)) (NULL (CDDR X))) (COND
    ((EVCONSTANTP (SETQ X (DSSB (CADR X)))) (MKQUOTE X))
    (T (LIST 'MKQUOTE X)) ))
  ((PAIRP X) (PROG (Y)
    (SETQ Y (DSSB (CDR X)))
    (SETQ X (DSSB (CAR X)))
    (RETURN (COND
      ((NULL Y) (COND
        ((QUOTEP X) (MKQUOTE (CDR X)))
        ((EVCONSTANTP X) (MKQUOTE (CONS X Y)))
        (T (LIST 'LIST X)) ))
      ((AND (EVCONSTANTP X) (EVCONSTANTP Y))
       (MKQUOTE (CONS (COND ((PAIRP X) (CADR X)) (T X))
                      (COND ((PAIRP Y) (CADR Y)) (T Y)) )) )
      ((EQCAR Y 'LIST) (CONS 'LIST (CONS X (CDR Y))))
      (T (LIST 'CONS X Y)) )) ))
  ((OR (NULL X) (CONSTANTP X) (MEMQ X DSARGS!*)) X)
  (T (MKQUOTE X)) ))

(DE EVCONSTANTP (X) (OR (NULL X) (EQ X T) (CONSTANTP X) (QUOTEP X)))

% DS defines substitution macros.
% !&DS has a special meaning only at beginning of argument list.
(DMACRO DS (FN ARGS !&REST BODY)
 (!$MACFEX FN (CONS '!&DS ARGS) BODY 'MACRO) )

% CMMACRO is similar to DMACRO. It is used when there is two
% definitions of a function: a FEXPR for the interpreter and a MACRO
% for the compiler. The MACRO is then defined by CMMACRO. In many
% systems CMMACRO can make a COMPILERMACRO or something else instead
% of a MACRO.
(DMACRO CMMACRO (FN ARGS !&REST BODY) (!$MACFEX FN ARGS BODY 'MACRO))

(DMACRO DFEXPR (FN ARGS !&REST BODY) (!$MACFEX FN ARGS BODY 'FEXPR))

(DMACRO DEXPR (FN ARGS !&REST BODY) (!$MACFEX FN ARGS BODY 'EXPR))

(DMACRO DFUN (FN TYP ARGS !&REST BODY) (!$MACFEX FN ARGS BODY TYP))

% (UNPUSH id) removes the first element from the list that must be
% value of "id". The new list is returned.
(DS UNPUSH (X) (SETQ X (CDR X)))

(DMACRO CONSA (X !&REST L) (COND
  (L (LIST 'CONS X (CONS 'CONSA L)))
  (T X) ))

% UNIONQ moved here from exfn2i. JEA
(DE UNIONQ (X Y)
 (COND
  ((NULL X) Y)
  ((MEMQ (CAR X) Y) (UNIONQ (CDR X) Y))
  (T (CONS (CAR X) (UNIONQ (CDR X) Y))) ) )

%:LH LOCALFUNCTIONS -Machine independent version (overwritten on Franz)
% Accumulates functions for later remobing.
% Sometimes (debugging) one might want to redefine (interpreted)
% a function which calls local functions;
% then we would need to undo their local declarations!
% To keep this option, don't remob when creating the image,
% but wait until ssheep.ini etc.  Sample sxxxx.ini file may contain:
% (DELQIP 'needed_fn  LOCALFUNCTIONS!*) % Temporary exclusion
% (FOREACH X IN LOCALFUNCTIONS!* DO (REMOB X)) % prior to remobing and
% (SETQ LOCALFUNCTIONS!* NIL)           % release of list storage
% Some lisps may remob also some functions used outside the defining
% file.
(DMACRO LOCALFUNCTIONS (!&REST L)  %:LH accumulate functions to remob
  (LENGTH (SETQ LOCALFUNCTIONS!* (UNIONQ LOCALFUNCTIONS!* L))) ) % JEA
% Calls to DFEXPR already here gives serious trouble. JEA.
%(DFEXPR LOCALFUNCTIONS (!&REST L)  %:LH accumulate functions to remob
%  (LENGTH (SETQ LOCALFUNCTIONS!* (UNIONQ LOCALFUNCTIONS!* L))) ) % JEA

% Redefinition of DE to allow for implied PROGN.
(DMACRO DE (FN ARGS !&REST BODY)
 (LIST 'PUTD (MKQUOTE FN) '(QUOTE EXPR)
  (MKQUOTE (LIST 'LAMBDA ARGS (MKPROGN BODY))) ) )

% A general MAPping macro.
% (FOREACH var keyword1 list keyword2 fn1 fn2 ...fnN) applies
% expressions "fn1" "fn2" ... "fnN" with variable "var" on arguments
% from list "list". This is done by one of the mapping functions
% determined from keyword1 and keyword2. Keyword1 is either IN or ON.
% "var" is in succes sion bound to each element of "list" in case of
% IN and to each tail of "list" in case of ON.
%  Keyword2 determines the value of FOREACH and is either DO, COLLECT,
% CONC or REPLACE. DO means return NIL, COLLECT means return a list of
% the results from "fnN", CONC means return a concatenation of the
% results by using NCONC and REPLACE means replace each element of
% "list" by the corresponding result and return "list".
%  FOREACH normally maps into the Lisp mapping functions MAP, MAPC,
% MAPCAR etc. It is expected that MAP etc. are open-coded by the
% compiler so that variables used in the mapped function do not need
% to be fluid. If this is not the case then FOREACH should instead map
% into PROG loops.
(DMACRO FOREACH (ARG ONKEY L DOKEY !&REST BODY)
 (PROG (MFN)
  (COND ((NULL (SETQ MFN (GET DOKEY ONKEY)))
         (ERRORFT 410
          "%r and %r is unknown FOREACH combination" ONKEY DOKEY ) ))
  (SETQ ARG (LIST ARG))
  (RETURN (LIST MFN L
    (LIST 'FUNCTION
     (COND
      ((AND (NULL (CDR BODY))
            (PAIRP (CAR BODY))
            (EQUAL (CDAR BODY) ARG)
            (NOT (AND (IDP (CAAR BODY))
                      (EQCAR (GETD (CAAR BODY)) 'MACRO) )) )
       (CAAR BODY) )
      (T (LIST 'LAMBDA ARG (MKPROGN BODY))) ) ) )) ) )

(IFP (NOT (FLAGP 'FOREACH 'LOSE))
 (DEFLIST
  '((DO MAP) (COLLECT MAPLIST) (CONC MAPCON) (REPLACE MAPRPL))
  'ON) )

(IFP (NOT (FLAGP 'FOREACH 'LOSE))
 (DEFLIST
  '((DO MAPC) (COLLECT MAPCAR) (CONC MAPCAN) (REPLACE MAPRPLC))
  'IN) )

% The following definition of FOREACH maps into PROG loops. It should
% be used if MAP, MAPC, MAPCAR etc. are not open-coded by the
% compiler. It is normally commented out.
(COMMENT DMACRO FOREACH (ARG ONKEY L DOKEY !&REST BODY)
 (PROG (ARGL RET INI)
  (SETQ BODY (MKPROGN BODY))
  (COND ((NULL (SETQ INI (GET DOKEY ONKEY)))
         (ERRORFT 410
          "%r and %r is unknown FOREACH combination" ONKEY DOKEY ) ))
  (COND ((AND (PAIRP BODY) (EQCAR (CDR BODY) ARG) (NULL (CDDR BODY)))
         (RETURN (LIST INI L (COND
            ((IDP (SETQ BODY (CAR BODY))) (MKQUOTE BODY))
            (T BODY) ))) ))
  (SETQ INI (LIST 'SETQ 'FO!$A1 L))
  (COND
   ((EQ DOKEY 'DO) (SETQ ARGL '(FO!$A1)))
   ((EQ DOKEY 'REPLACE)
    (SETQ ARGL '(FO!$A1 FO!$A2))
    (SETQ BODY (LIST 'RPLACA 'FO!$A1 BODY))
    (SETQ RET 'FO!$A2)
    (SETQ INI (LIST 'SETQ 'FO!$A2 INI)) )
   (T (SETQ BODY (COND
        ((EQ DOKEY 'CONC)
         (SETQ ARGL '(FO!$A1 FO!$A2 FO!$A3 FO!$A4))
         (LIST 'COND (CONS
           (LIST 'SETQ 'FO!$A4 BODY)
           '((NCONC FO!$A3 FO!$A4) (SETQ FO!$A3 FO!$A4)) )) )
        (T (SETQ ARGL '(FO!$A1 FO!$A2 FO!$A3))
           (LIST 'PROGN
            (LIST 'RPLACD 'FO!$A3 BODY)
            '(SETQ FO!$A3 (CDR FO!$A3)) ) ) ))
      (SETQ RET '(CDR FO!$A2))
      (SETQ INI
       (LIST 'PROGN INI '(SETQ FO!$A2 (SETQ FO!$A3 (NCONS NIL)))) ) ) )
  (RETURN (LIST 'PROG 'ARGL
                INI 
                '(GO FO!$L1)
        'FO!$L2 (LIST 'SETQ ARG (COND
                  ((EQ ONKEY 'ON) 'FO!$A1)
                  (T '(CAR FO!$A1)) ))
                BODY
                '(SETQ FO!$A1 (CDR FO!$A1))
        'FO!$L1 '(COND (FO!$A1 (GO FO!$L2)))
                (LIST 'RETURN RET) )) ))
% End of comment

% This is the same as the MacLisp LET. The name PLET is used to avoid
% collision with Reduce LET.
(DMACRO SLET (ARGS !&REST BODY) (PROG (PVRS)
% Similar to PLET but variables are bound sequentially as in a PROG.
% Note that the value of a VAR is available when its corresponding ARG
% has been evaluated, but it is undefined before that. SLET is not
% necessarily expanded into a PROG, i.e. GO and RETURN are not allowed.
  (SETQ BODY (LIST (LIST 'RETURN (MKPROGN BODY))))
  (FOREACH X IN (REVERSE ARGS) DO (XPUSH PVRS (COND
     ((NOT (PAIRP X)) X)
     (T (COND ((CDR X) (XPUSH BODY (CONS 'SETQ X)))) (CAR X)) )))
  (RETURN (CONSA 'PROG PVRS BODY)) ))

% PLET(VARS:list,[XPR:form]):any
% The elements of VARS are either a list (VAR:id ARG:form) or an id
% VAR which is equivalent to a list (VAR NIL).
% The VAR are bound to the ARG (this is done in parallel for all VAR),
% the XPR are evaluated, the old bindings of the VAR are restored and
% the value of the last XPR is returned. Note that when the ARG are
% evaluated then all the VAR have their old value.
%  The PLET is (but this is no guarantee) implemented as a macro
% expanding into a LAMBDA:  ((LAMBDA VRS (PROGN [XPR])) . ARGS) where
% VRS is a list of the VAR and ARGS is a list of the ARG.
(DMACRO PLET (ARG !&REST BODY) (SLET (VRS ARGS)
  (FOREACH X IN (REVERSE ARG) DO
   (COND
    ((ATOM X) (XPUSH VRS X) (XPUSH ARGS NIL))
    (T (XPUSH VRS (CAR X))
       (XPUSH ARGS (COND ((CDR X) (CADR X)))) ) ) )
  (CONS (LIST 'LAMBDA VRS (MKPROGN BODY)) ARGS) ))

% Add help information on subject FN. FN will be added to a list of
% help information HLPLIS. Revised by JEA.
(DE ADDHLP (HLPLIS FN HLP)
  (SLET ((EVHLPLIS (EVAL HLPLIS)))
    (COND
      ((AND HLPLIS (NOT (MEMQ FN EVHLPLIS)))
       (SET HLPLIS (NCONC1 EVHLPLIS FN)) ))
    (PUT FN 'HELP HLP) ))

% MOVD(X:id,Y:id):X
% Defines X to be the same function as Y. Error if Y is not a function.
(DE MOVD (X Y)
 (SLET ((A (GETD Y)))
  (COND ((NULL A) (ERRORFT 408 "Function %r is not defined" Y)))
  (PUTD X (CAR A) (CDR A)) ) )

% SAMEDEF cannot be an SMACRO as it confuses Rcref.
(COMMENT
(DMACRO SAMEDEF (X Y) 
 (LIST 'BOTHTIMES (LIST 'MOVD (MKQUOTE X) '(QUOTE DOSAMEDEF))
  (LIST 'DPUT X 'SAME!-AS Y) ) )

(DMACRO DOSAMEDEF (!&MACRO FN !&REST L) (CONS (GET FN 'SAME!-AS) L))

(DMACRO SAMEDEF (X Y) (COND
  ((FLAGP X 'LOSE) (WARNINGFT 36 "%p not defined - flagged LOSE" X) NIL)
  (T (LIST 'CPROGN      
      (LIST 'MOVD (MKQUOTE X) (MKQUOTE Y))
      (LIST 'CMMACRO X '(!&MACRO !&REST L)
       (LIST 'CONS (MKQUOTE Y) '(CDR L)) ) )) ))
)

% SAMEDEF must be defined with the aid of !&MACRO at least for Culisp,
% or there will be problems if the defined function is in a not yet
% loaded module. JEA.
(DMACRO SAMEDEF (X Y)
 (LIST 'DMACRO X '(!&MACRO !&REST L)
  (LIST 'CONS (MKQUOTE Y) '(CDR L)) ) )

% FOLDCASE(SW:boolean,[SXPR:form]):any
% Evaluate all SXPR while character folding status temporarily is set
% to SW.
(DS FOLDCASE (SW !&REST L) (PLET ((!*RAISE SW)) . L))

% CONCATID([IDN:id]):id
% CONCATID returns an id whos print-name is a concatenation of the
% names of the IDNs.
(DMACRO CONCATID (!&REST L) (COND
  ((NULL L) NIL)
  ((NULL (CDR L)) (CAR L))
  (T (LIST 'FOLDCASE NIL
      (LIST 'INTERN (LIST 'COMPRESS (CONCATIDHELP L))) )) ))

(DE CONCATIDHELP (U)
 (LIST 'NCONC (LIST 'EXPLODE (CAR U))
  (COND
   ((CDDR U) (CONCATIDHELP (CDR U)))
   (T (LIST 'EXPLODE (CADR U))) ) ) )

(DE ILOS!$PUT (X Y)
 (COND ((OR (FLAGP X 'LOSE) (FLAGP Y 'LOSE))
        (WARNINGFT 38 "%p property on %p not defined - LOSE flag" Y X)
        T )) )

(DE I!$EV (X) (AND (QUOTEP X) (CADR X)))

% CPROGN is the same as PROGN except maybe at the top level when
% compiling. When compiling, the arguments of CPROGN at top level are
% seen at top level too. This might be true for PROGN too, but it is
% not guaranteed.
(DS CPROGN (!&REST U) (PROGN . U))

(IFP (NOT (FLAGP 'PUTC 'LOSE)) (CPROGN
  (GLOBAL '(COMPFLGSTK!*))
  (COND ((NOT (MEMQ 'COMPFLGSTK!* SFSTKS!*))    % No PUSHQ defined, do
         (SETQ SFSTKS!* (CONS 'COMPFLGSTK!* SFSTKS!*)) )) % it this way.
  (DE SETCMPFL (X) (SETQ !*COMP X))
  (DEFLIST '((COMPFLGSTK!* SETCMPFL)) 'SFSTKFN)
  (FLAG '(COMPFLGSTK!*) 'LASTONLY) ))

% PUTC is similar to PUT. 
% The important difference from PUT is when compiling. The third
% argument is compiled if the second argument is flagged COMPILE.
% The following is for a resident compiler.
(DE PUTC (X Y Z) (COND
  ((ILOS!$PUT X Y) NIL)
  ((NOT (EQCAR Z 'LAMBDA)) (PUT X Y Z) X)
  (T (XPUSH COMPFLGSTK!* !*COMP)    % Has to be done this way as !*COMP
     (SETQ !*COMP (AND !*COMP (FLAGP Y 'COMPILE))) % is a global!
     (PUT X Y (PUTD (CONCATID X PUTCHR!* Y) 'EXPR Z))
     (SETQ !*COMP (CAR COMPFLGSTK!*))
     (SETQ COMPFLGSTK!* (CDR  COMPFLGSTK!*))
     X) ))

(COMMENT % For a separate compiler.
(DMACRO PUTC (FN TYP PROP) (COND
  ((ILOS!$PUT (I!$EV FN) (I!$EV TYP)) NIL)
  ((AND (QUOTEP TYP) (FLAGP (CADR TYP) 'COMPILE))
   (SLET
    ((NM (MKQUOTE (CONCATID
                   (COND ((QUOTEP FN) (CADR FN)) (T (GENSYM)))
                   PUTCHR!* (CADR TYP) ))))
    (LIST 'CPROGN
     (LIST 'PUTD NM '(QUOTE EXPR) PROP)
     (LIST 'IPUTC FN TYP NM) ) ) )
  (T (LIST 'IPUTC FN TYP PROP)) ))

(DE IPUTC (X Y Z)
 (COND ((NOT (ILOS!$PUT X Y))
        (PUT X Y (COND (!*MACROEXPAND (MACEXPD Z)) (T Z)))
        X)) )
) % End of comment.

% DC(NAME:id,PROP:id,VARS:idlist,[SXPR:form]):NAME
% The LAMBDA-expression (LAMBDA VARS (PROGN [SXPR])) is put on the
% property list of NAME, with the property indicator PROP.
% DC is similar to DE, except for putting the LAMBDA on the property
% list of NAME instead of in the function cell.
(DMACRO DC (FN PROP ARG !&REST BODY)
 (LIST 'PUTC (MKQUOTE FN) (MKQUOTE PROP)
  (MKQUOTE (LIST 'LAMBDA ARG (MKPROGN BODY))) ) )

(DMACRO XPUT (X TYP PROP)
 (COND
  ((ILOS!$PUT (I!$EV X) (I!$EV TYP)) NIL)
  (T (LIST 'IXPUT X TYP PROP)) ) )

(DE IXPUT (X Y Z) (COND ((NOT (ILOS!$PUT X Y)) (PUT X Y Z))))

(DS DPUT (X Y Z) (XPUT 'X 'Y 'Z))

(DMACRO XDEFLIST (TYP !&REST L) (COND
  ((FLAGP TYP 'LOSE)
   (WARNINGFT 40 "%p properties not defined - flagged LOSE" TYP) NIL )
  (T (SETQ TYP (MKQUOTE TYP))
     (CONS 'CPROGN (FOREACH X IN L COLLECT
       (LIST 'XPUT (MKQUOTE (CAR X)) TYP (MKQUOTE (CADR X))) )) ) ))

(DMACRO XFLAGS (FLG !&REST L) (COND
  ((FLAGP FLG 'LOSE)
   (WARNINGFT 42 "Ignored %p flagging - flagged LOSE" FLG) NIL )
  ((SETQ L (FOREACH X IN L CONC (COND
      ((FLAGP X 'LOSE)
       (WARNINGFT 44 "%p not flagged %p - LOSE flag" FLG X) NIL )
      (T (LIST X)) )))
   (LIST 'FLAG (MKQUOTE L) (MKQUOTE FLG)) ) ))

% GVALUE(ARG:form):form
% GVALUE is used to put computed constants in code.
% ARG is evaluated at macro expansion time and the quoted result is
% returned.
(DMACRO GVALUE (X) (MKQUOTE (EVAL X)))

% COMPILETIME([ARG:form]):NIL
% If compiling or cross-referencing, evaluate the arguments
(DMACRO COMPILETIME (!&REST L) (COND (!*DEFN (CONS 'CPROGN L))))

% NOTLOADTIME([ARG:form]):NIL
% Evaluate when interpreting or at compile time.
(DS NOTLOADTIME (!&REST L) (PROGN . L))

% MACROTIME([ARG:form]):NIL
% Evaluate at load time also, if compiling a macro-file.
(DMACRO MACROTIME (!&REST L)
 (CONS (COND (!*MACROTIME 'BOTHTIMES) (T 'NOTLOADTIME)) L) )

% BOTHTIMES([ARG:form]):NIL
(DS BOTHTIMES (!&REST L) (CPROGN . L))

% LOADTIME([ARG:form]):NIL
(DS LOADTIME (!&REST L) (PROGN . L))

% NOCOMP switches off compilation up to a matching RESTORECOMP.
(DE NOCOMP ()
  (PROGN
    (SETQ !*SAVECOMP !*COMP)
    (SETQ !*COMP NIL)))

% RESTORECOMP restores compilation status to original status.
(DE RESTORECOMP () (SETQ !*COMP !*SAVECOMP))

(DMACRO USEDPROPERTIES (!&REST DUMMY) NIL)

(DMACRO USEDFLAGS (!&REST DUMMY) NIL)

(DFEXPR DECLAREFEXPR (!&REST L)
 (FOREACH X IN L DO (PUT X 'CFNTYPE '(FEXPR))) )

(DMACRO ENTRYPOINTS (!&REST DUMMY) NIL)

(DMACRO SETCONSTANTS (!&REST L) (SETGLBS L 'GLOBAL 'SETQ))

(DMACRO SETGLOBALS (!&REST L) (SETGLBS L 'GLOBAL 'SETQ))

(DMACRO SETLOCALGLOBALS (!&REST L) (SETGLBS L 'GLOBAL 'SETQ))

(DMACRO SETFLUIDS (!&REST L) (SETGLBS L 'FLUID 'SETQ))

(DMACRO QSETGLOBALS (!&REST L) (SETGLBS L 'GLOBAL 'QSETQ))

(DMACRO QSETLOCALGLOBALS (!&REST L) (SETGLBS L 'GLOBAL 'QSETQ))

(DMACRO QSETFLUIDS (!&REST L) (SETGLBS L 'FLUID 'QSETQ))

(DS QSETQ (X Y) (QSET 'X Y))

(DE SETGLBS (U Y Z) (SLET (A B)
  (COND ((SETQ B (FOREACH X IN U CONC (COND
            ((FLAGP X 'LOSE)
             (WARNINGFT 46 "Variable %p not defined - flagged LOSE" X)
             NIL )
            (T (LIST (COND
                 ((NOT (PAIRP X)) X)
                 ((NOT (PAIRP (CAR X)))
                  (XPUSH A (LIST Z (CAR X) (MKQUOTE (CADR X))))
                  (CAR X) )
                 (T (XPUSH A (LIST Z (CAAR X) (CADAR X)))
                    (CAAR X) ) ))) )))
         (SETQ B (LIST Y (MKQUOTE B)))
         (COND
          (A (CONSA 'CPROGN B (REVERSIP A)))
          (T B) ) )) ))

(DE SYSTYPE (X)
 (COND
  ((NULL X) SYSTYPE!*)
  ((SETQ X (ATSOC X SYSTYPE!*)) (CDR X)) ) )

(DE SETSYSTYPE (X Y)
 (SLET ((A (ATSOC X SYSTYPE!*)) B)
  (COND
   (A (SETQ B (CDR A)) (RPLACD A Y) B)
   (T (XPUSH SYSTYPE!* (CONS X Y)) NIL) ) ) )

%(SETFLUIDS THROWVALUE!$)

(FLAG '(GVALUE COMMENT COMPILETIME NOTLOADTIME DECLAREFEXPR
        ENTRYPOINTS FUNCTIONS LOCALFUNCTIONS)
      'IGNORE)

(FLAG '(FLUID GLOBAL UNFLUID DS DMACRO CMMACRO BOTHTIMES) 'EVAL)

(FLAG '(MACROTIME SAMEDEF DC XPUT DPUT XDEFLIST IFP CPROGN
        SETGLOBALS SETLOCALGLOBALS SETFLUIDS SETCONSTANTS
        QSETGLOBALS QSETLOCALGLOBALS QSETFLUIDS)
      'DFEXPAND)

(NOTLOADTIME (COND (!*CREF      % To make RCREF work properly
   (REMFLAG '(DECLAREFEXPR ENTRYPOINTS
              FUNCTIONS LOCALFUNCTIONS )
            'IGNORE )
   (REMFLAG '(FLUID GLOBAL UNFLUID) 'EVAL)
   (REMFLAG '(DMACRO DS DM CMMACRO) 'EVAL)
   (REMFLAG '(SAMEDEF DC XPUT DPUT XDEFLIST IFP
              SETGLOBALS SETLOCALGLOBALS SETFLUIDS SETCONSTANTS
              QSETGLOBALS QSETLOCALGLOBALS QSETFLUIDS)
            'DFEXPAND) )))

%### EXFN1I.SL end
