%### LOAD.SL ========================================================
%
%       LOAD.SL
%       -------
%
%       Copyright (c) 1982 by Inge Frick, Institute of Theoretical
%       Physics, Vanadisvagen 9, S-113 46 Stockholm, Sweden.
%
%       This file consists of the system dependent parts of LOADI.SL.
%       specific to SLISP built on Franz Lisp on the VAX.
%
%====================================================================

(declare (special FIL!*))               % To avoid complaints

%(LOCALFUNCTIONS ADDEXTENSION ADDPREFIX)

(SETFLUIDS !*KEEPCASE !$ldprint)

(SETQ !$ldprint nil)

(DE ADDEXTENSION (X Y)
 (COND
  (Y (readlist (cons '!" (NCONC (EXPLODEC X)
         (CONS '!. (NCONC
                    (COND ((NOT (EQ Y '! )) (EXPLODEC Y)))
                    '(!") )) ))))
  (T X) ) )




(DE ADDPREFIX (X Y) (COND
  ((EQ X '! )
   (readlist (CONS '!" (CONS '!/ (NCONC (EXPLODEC Y) '(!"))))))
  (X (SLET ((Z (LASTPAIR (SETQ X (UNIXIFY X)))))
      (SETQ Y (NCONC (EXPLODEC Y) '(!")))
      (RPLACD Z (COND
        ((EQ (CAR Z) '!/ ) Y)
        (T (CONS '!/ Y)) ))
      (readlist (CONS '!" X)) ))
  (T Y) ))

(DE ADDSUBDIR (X Y) (COND
  ((NULL X) (MAKESTR (EXPLODEC Y)))
  (Y (ADDPREFIX X Y))
  (T X) ))

(DE UNIXIFY (W)
 (SLET ((X (SETQ W (EXPLODEC W))) Y Z SW)
  (WHILE
   (SETQ Y (MEMQ '!: X))
   (RPLACA Y '!/ )
   (SETQ X (CDR Y))
   (SETQ SW T) )
  (COND (SW (XPUSH W '!/ )))
  (COND
   ((AND X (MEMQ (CAR X) '(![ !<)))
    (COND
     ((EQCAR (CDR X) '!.) (RPLACW X (CDDR X)))
     (SW (RPLACW X (CDR X)))
     ((SETQ Y (GETLOGICAL 'SYS!$DISK))
      (SETQ W (NCONC (UNIXIFY Y) (CDR X))) )
     (T (RPLACA X '!/ )) )
    (COND ((NULL (SETQ Y (OR (MEMQ '!> X) (MEMQ '!] X))))
           (ERROR 0 (LIST "Wrong file specification" (MAKESTR W))) ))
    (RPLACA Y '!/ ) (SETQ Z (CDR Y)) (RPLACD Y NIL)
    (WHILE (SETQ Y (MEMQ '!. X))
           (SETQ X (COND
             ((EQ (CADR Y) '!/ ) (RPLACW X (CDR X)) NIL)
             (T (RPLACA Y '!/ )) )) )
    (NCONC W Z) )
   ((AND (EQ (CAR W) '!~)               % ~ stands for home directory
         (OR (NULL (CDR W)) (EQ (CADR W) '!/ ))
         (SETQ Y (OR (GETLOGICAL '!H!O!M!E)
                     (GETLOGICAL 'SYS!$LOGIN) )) )
    (SETQ Y (UNIXIFY Y))
    (COND ((CDR W)
           (SETQ X (LASTPAIR Y))
           (RPLACD X (COND ((EQ (CAR X) '!/ ) (CDDR W)) (T (CDR W)))) ))
    Y )    
   (T W) ) ) )

(DE TESTFILE (X FIL SW)
   (COND
    ((COND
      (!*KEEPCASE (FILEP X))          % JEA
      ((OR (AND (NOT (AND SW (IDP FIL))) (FILEP X))
           (FILEP (SETQ X (LKEY X))) )) )
     X) ) )

(DE SETCASE (X) (COND (!*KEEPCASE X) (T (LKEY X)))) % JEA

(DE PREFIXOF (FIL) (SLET ((X (UNIXIFY FIL)) (Z X) Y SW)
% FIL is a real file description with no special interpretation.
% PREFIXOF returns the prefix part of FIL, or NIL if there is no
% prefix. Both absolute and relative prefixes are returned.
% This function is necessarily system-dependent as it has to know how
% files are described.
  (SETQ X (CONS NIL X))
  (WHILE
   (SETQ Y (MEM2Q '!/ (CDR X)))         % Unix prefix
   (SETQ X Y)
   (SETQ SW T) )
  (COND
   (SW (RPLACD X NIL) (MAKESTR Z))
   ((EQ (CAR Z) '!/ ) '! ) ) ))

(DE FILNAM (FIL) (SLET (PX NM EX X Y) 
% FIL is a file description. FILNAM takes FIL apart and returns
% (<prefix> <file name> . <extension>). Where <file name> and
% <extension> are interned ids and <prefix> is either a string an id
% or a search list. When some part is missing, it is returned as NIL.
% Various special system dependent interpretations of FIL might be
% done, in which case the fluid variable FIL!* is updated. A relative
% prefix is returned as NIL.
% This function is necessarily system dependent as it has to know how
% files are described.
  (SETQ FIL (UNIXIFY (OR (GETLOGICAL FIL) FIL)))
  (COND ((AND (EQ (CAR FIL) '!/ ) (SETQ Y (MEM2Q '!/ FIL)))  % Try to
         (SETQ X (CDR Y)) (RPLACD Y NIL)        % get path from /name/
         (COND
          ((SETQ PX (GET (CONCATIDL (CDR FIL)) 'PATH))
           (SETQ FIL (CDR X)) )
          (T (RPLACD Y X)) ) ))
  (COND ((NOT !*KEEPCASE) (SETQ FIL (UKEYL FIL))))
  (SETQ FIL!* (CONCATIDL FIL))
  (SETQ X NIL) (SETQ NM FIL)
  (WHILE
   (SETQ Y (MEMQ '!/ NM))       % Unix prefix
   (SETQ X Y)
   (SETQ NM (CDR Y)) )
  (COND (X (RPLACD X NIL)
         (COND
          ((EQ (CAR FIL) '!/ )  % Absolute prefix
           (SETQ PX (CONCATIDL FIL)) )
          ((AND PX (NOT (PAIRP PX)))
           (SETQ PX (ADDPREFIX PX (CONCATIDL FIL))) ) )
         (RPLACD X NM) ))
  (COND
   ((NULL NM))                  % No file name and no extension.
   ((EQ (CAR NM) '!.)           % Extension but no file name.
    (SETQ EX NM)
    (SETQ NM NIL) )
   ((SETQ Y (MEM2Q '!. NM)) (SETQ EX (CDR Y)) (RPLACD Y NIL)) )
  (CONS PX (CONS
            (COND (NM (CONCATIDL NM)))
            (COND (EX (COND ((CDR EX) (CONCATIDL (CDR EX)))
                            (T '! ) ))) )) ))

(COMMENT
(DE FILNAM (FIL) (SLET ((X (GETLOGICAL FIL)) Y Z L U)
  (SETQ FIL (COND
    (X (SETQ X (UNIXIFY X)))
    (T (SETQ X (SETQ FIL (CONS NIL (UNIXIFY FIL))))
       (WHILE           % Do Unix $ expansion, i.e. expand $name/
        (AND X (SETQ Z (MEM2Q '!$ X)))
        (COND
         ((EQ (CAR Z) '!\)      % Do not expand \$, but replace it by $
          (RPLACA Z '!$) (SETQ X (RPLACD Z (CDDR Z))) )
         (T (COND ((SETQ Y (MEM2Q '!/ (SETQ X (CDR Z))))
                   (SETQ U (CDR Y)) (RPLACD Y NIL) )) % End at /
            (COND
             ((SETQ L (GETLOGICAL (MAKESTR (CDDR Z)))) % Expand
              (RPLACD Z (SETQ L (UNIXIFY L)))
              (SETQ X (COND (Y U)))
              (SETQ L (LASTPAIR L))
              (RPLACD L (COND
                ((AND Y (EQ (CAR L) '!/ )) (CDR U))
                (Y U) )) )
             (Y (RPLACD Y U)) ) ) ) ) % No expansion
       (CDR FIL) ) ))
  (COND
   ((AND (EQ (CAR FIL) '!/ ) (SETQ Y (MEM2Q '!/ FIL)))  % Try to get
    (SETQ Z (CDR Y)) (RPLACD Y NIL)     % path from /name/
    (COND
     ((SETQ L (GET (concatl X) 'PATH)) (SETQ FIL (CDR Z)))
     (T (RPLACD Y Z)) ) )
   (T (SETQ L NIL)) )
  (COND ((NOT !*KEEPCASE) (SETQ FIL (UKEYL FIL))))
  (SETQ FIL!* (MAKESTR FIL))
  (SETQ Z NIL) (SETQ X FIL)
  (WHILE
   (SETQ Y (MEMQ '!/ X))        % Unix prefix
   (SETQ Z Y)
   (SETQ X (CDR Y)) )
  (COND ((AND Z (EQ (CAR FIL) '!/ ))    % Seen an absolute prefix
         (RPLACD Z NIL)
         (COND
          ((NULL L) (SETQ L (MAKESTR FIL)))
          ((NOT (PAIRP L))              % prefix
           (SETQ L (ADDPREFIX L (MAKESTR FIL))) ) ) ))
  (SETQ Z NIL)
  (COND
   ((NULL X))                           % No file name and no extension.
   ((EQ (CAR X) '!.)                    % Extension but no file name.
    (SETQ Z X)
    (SETQ X NIL) )
   ((SETQ Y (MEM2Q '!. X)) (SETQ Z (CDR Y)) (RPLACD Y NIL)) )
  (COND ((AND Z                         % Extension
          (OR (SETQ Y (MEM2Q '!. Z)) (SETQ Y (MEM2Q '!; Z))) )
         (RPLACD Y NIL) ))              % Remove VMS version number.
  (CONS L
   (CONS
    (COND (X (concatl X)))
    (COND (Z (COND ((CDR Z) (concatl (CDR Z))) (T '! )))) )) ))
)

(DE GETLOGICAL (FIL) (SLET (A) % This is for SLISP on Franz
  (COND ((NOT (EQ
           (SETQ FIL (getenv (COND
              ((EQUAL (substring FIL -1) ":")
               (RPLACD (LAST2PAIR (SETQ A (EXPLODEC FIL))) NIL)
               (concatl A) )
              (T (concat FIL)) )))
           (GVALUE (concat "")) ))
         FIL)) ))

(XDEFLIST INFN (SL READIN) (!  READIN) (O XFASL) (L FRANZREADIN))
% Indicate loading function for files with extension SL and no
% extension.
% O is extension for fasl files in Franz Lisp.

(COND ((NOT (FNP 'IREADIN)) (PUTD 'IREADIN 'EXPR (GETFN 'READIN))))

(DE READIN (X Y) (SLET ((readtable SLISPREADTABLE)) (IREADIN X Y)))

(DE XFASL (X DUMMY) (SLET ((readtable SLISPREADTABLE)) (fasl X)))
% Fast-loader for SLISP on FRANZ-Lisp

(DE FRANZREADIN (X Y) (SLET ((readtable FRANZREADTABLE)) (IREADIN X Y)))

(DE load (X) (SLET ((INFN!* 'FRANZREADIN)) (MLOAD X)))

(eval!-when (eval compile)
  (FLAG '(TESTFILE ITESTFILE SETCASE) 'LOSE) )  % JEA

(INCLUDE "../loadi.sl")

(eval!-when (eval compile)
  (REMFLAG '(TESTFILE ITESTFILE SETCASE) 'LOSE) ) % JEA

(SETQ FASLOUTEXT!* 'O)          % Extension for created fasl files.

(DPUT FLOADEXT SUBDRS O)        % Extension for fast loading.

(SETQ DEFAULTEXTENSION!* 'O)    % FASL is default.

%### LOAD.SL end
