;### PTLSZT.L =======================================================
;
(include "chead.l")
;                       l i s z t   v 4
;
;
;
;                   A compiler for Franz lisp
;
; Copyright (c) 1980 ,  The Regents of the University of California.
; All rights reserved.
; author: j. foderaro
;
;--------------------------------------------------------------
; Modified to allow the compilation of Standard LISP too, mainly
; by means of a macro package defined in the file slmacs.o with
; sources slmac.l slfuns.l exfn1.sl ../exfn1i.sl excom.sl systyp.sl
; ../excomi.sl exfn2.sl ../exfn2i.sl load.sl ../loadi.sl ../prntfi.sl
; slmac1.l and slmacs.l.
; Modified code can be found by searching for markers '*IF*' or
; 'MAHM' (M.A.H. MacCallum) or 'JEA' (J.E. Aman).
;
; Inge Frick, Institute of Theoretical Physics,
; Vanadisvagen 9, S-113 46 Stockholm, Sweden.
;
;--------------------------------------------------------------

;Define the conditional coding macro IFP.  *IF*
(defun IFP macro (u) (cond
  ((eval (cadr u)) (caddr u))
  ((cdddr u) (cadddr u)) ))

;This is from cmacros.l
(IFP (not (getd 'Push)) (eval-when (compile load)
  (defmacro Push (x y) `(setq ,x (cons ,y ,x))) ))

;This is from macros.l
(IFP (not (getd 'push)) (eval-when (compile load)
  (defmacro push (x y) `(setq ,y (cons ,x ,y))) ))

;This is from macros.l
(IFP (not (getd 'unpush)) (eval-when (compile load)
  (defmacro unpush (x) `(setq ,x (cdr ,x))) ))

;later versions have some new names, use them.
;This is from macros.l
(IFP (not (getd 'if)) (eval-when (compile eval)

  (declare (special if-keyword-list))

  (setq if-keyword-list '(then thenret elseif else))

;This is from macros.l
;--- if
;
;  the keyword if expression is parsed using a simple four state
; automaton.  The expression is parsed in reverse.
; States:
;       init - have parsed a complete predicate,  then clause
;       col  - have collected at least one non keyword in col
;       then - have just seen a then, looking for a predicate
;       compl - have just seen a predicate after an then, looking
;               for elseif or if (i.e. end of forms).
;
  (defmacro if (&rest args)
   (let ((len (length args)))
      ;; first eliminate the non-keyword if macro cases
      (cond ((<& len 2)
             (error "if: not enough arguments " args))
            ((and (=& len 2)
                  (not (memq (cadr args) if-keyword-list)))
             `(cond (,(car args) ,(cadr args))))
            ; clause if there are not keywords (and len > 2)
            ((do ((xx args (cdr xx)))
                 ((null xx) t)
                 (cond ((memq (car xx) if-keyword-list)
                        (return nil))))
             `(cond (,(car args) ,(cadr args))
                    (t ,@(cddr args))))

            ;; must be an instance of a keyword if macro

            (t (do ((xx (reverse args) (cdr xx))
                    (state 'init)
                    (elseseen nil)
                    (totalcol nil)
                    (col nil))
                   ((null xx)
                    (cond ((eq state 'compl)
                           `(cond ,@totalcol))
                          (t (error "if: illegal form " args))))
                   (cond ((eq state 'init)
                          (cond ((memq (car xx) if-keyword-list)
                                 (cond ((eq (car xx) 'thenret)
                                        (setq col nil
                                              state 'then))
                                       (t (error "if: bad keyword "
                                                 (car xx) args))))
                                (t (setq state 'col
                                         col nil)
                                   (push (car xx) col))))
                         ((eq state 'col)
                          (cond ((memq (car xx) if-keyword-list)
                                 (cond ((eq (car xx) 'else)
                                        (cond (elseseen
                                                 (error
                                                    "if: multiples elses "
                                                    args)))
                                        (setq elseseen t)
                                        (setq state 'init)
                                        (push `(t ,@col) totalcol))
                                       ((eq (car xx) 'then)
                                        (setq state 'then))
                                       (t (error "if: bad keyword "
                                                 (car xx) args))))
                                (t (push (car xx) col))))
                         ((eq state 'then)
                          (cond ((memq (car xx) if-keyword-list)
                                 (error "if: keyword at the wrong place "
                                        (car xx) args))
                                (t (setq state 'compl)
                                   (push `(,(car xx) ,@col) totalcol))))
                         ((eq state 'compl)
                          (cond ((not (eq (car xx) 'elseif))
                                 (error "if: missing elseif clause " args)))
                          (setq state 'init)))))))) ) )

;Later versions have functions vith a : instead of macros with a _
(IFP (not (getd 'sys:time)) (eval-when (compile eval)
  (or (get 'syscall 'version) (load 'syscall))
  (putd 'sys:time (getd 'sys_time)) ))

(IFP (not (getd 'sys:getpid)) (eval-when (compile eval)
  (or (get 'syscall 'version) (load 'syscall))
  (putd 'sys:getpid (getd 'sys_getpid)) ))

(IFP (not (getd 'sys:unlink)) (eval-when (compile eval)
  (or (get 'syscall 'version) (load 'syscall))
  (putd 'sys:unlink (getd 'sys_unlink)) ))

(IFP (not (getd 'sys:chmod)) (eval-when (compile eval)
  (or (get 'syscall 'version) (load 'syscall))
  (putd 'sys:chmod (getd 'sys_chmod)) ))

;Patch to lversion.l
(setq compiler-name (concat compiler-name " patched to SLC"))
;End patch to lversion.l

; This cuts down the number of garbage collections. *IF*
(setq $gc_pct 0.80)
(setq $gc_midlim 40)
(setq $gc_lowlim 30)

;Patch to decl.l
; global franz special variables
(declare (special specvars*)) ;*IF* list of variables special by default

(declare (special v-extn liszt-read $EOF$ v-ofile v-root *IPRN ; *IF*
                  slisp-load-search-path))      ;*IF*, MAHM and JEA

(setq $EOF$ (ascii 127))        ;*IF*
(cond ((not (boundp 'lisp-library-directory))
       (setq lisp-library-directory "/usr/lib/lisp") ))
(cond ((not (boundp 'slisp-load-search-path))   ; *IF*
       (setq  slisp-load-search-path nil) ))    ; *IF*

;--- special handlers
(putprop 'progn 'cc-progn       'fl-exprcc)     ; *IF* this allows go
                                                ; inside progn.

(remprop 'progn 'fl-exprm)              ; *IF*


; list of known special variables in franz lisp
(setq specvars*                         ; *IF*
 '($gccount$ $gcprint $ldprint ER%all ER%brk ER%err ER%misc ER%tpl
   ER%undef ^w defmacro-for-compiling environment errlist errport
   evalhook float-format funcallhook gcdisable ibase linel nil
   piport poport prinlength prinlevel ptport readtable t top-level
   user-top-level lisp-library-directory Standard-Input
   Standard-Output lisp-object-directory))

; file-extensions special to standard lisp
(putprop 'sl 't 'SLISP)                 ; *IF*
(putprop 'shp 't 'SLISP)                ; *IF*
;End patch to decl.l

; cc-progn :: compile a "progn" expression              = cc-progn =
;*IF* this version allows go inside progn
;It replaces a macro in func.l
;
(defun cc-progn nil                     ; *IF*
  (makecomment '(beginning progn))
  (d-exps (cdr v-form)))


;Patch to tlev.l
;--- liszt - v-x : list containing file name to compile and optionally
;                and output file name for the assembler source.
;
(def liszt
  (nlambda (v-x)
           (prog (piport v-root v-ifile v-sfile v-ofile
                         v-Lfile v-mfile v-extn         ; *IF*
                         vp-ifile vp-sfile vps-crap
                         vps-include vns-include
                         asm-exit-status ntem temgc temcp
                         tmp
                         rootreal g-arrayspecs out-path
                         g-decls g-stdref pre-eval include-files
                         g-fname g-trueop g-falseop g-didvectorcode
                         tem temr starttime startptime startgccount
                         fl-asm fl-warn fl-verb fl-inter fl-xref fl-uci
                         fl-run fl-case fl-anno g-optionalp
                         fl-slisp       ;*IF*
                         liszt-process-forms in-line-lambda-number
                         g-skipcode g-dropnpcnt g-complrname)

                 ;in case "S" switch given, set asm-exit-status
                 ;  to 0 (so garbage won't be returned).
                 (setq asm-exit-status 0)

                 ; turn on monitoring if it exists
                 #+monitoring
                 (errset (progn (monitor t)     ; turn it on
                                (print 'monitor-on)
                                (terpr))
                         nil)
                 (setq starttime (sys:time)   ; real time in seconds
                       startptime (ptime)
                       startgccount $gccount$)
                 (setq in-line-lambda-number (sys:time))
                 (cond ((null (boundp 'internal-macros))
                        (setq internal-macros nil)))
                 (cond ((null (boundp 'macros))
                        (setq macros nil)))
                 (setq er-fatal 0)
                 (setq vps-include nil
                       vns-include nil)  ;stack of ports and names
                 (setq twa-list nil)
                 (setq liszt-eof-forms nil)

                 ; look for lisztrc file and return if error occured
                 ; in reading it
                 (cond ((eq (do-lisztrc-check) 'error)
                        (return 1)))

                 ; set up once only g variables
                 (setq g-comments nil
                       g-current nil            ; current function name
                       g-funcs nil
                       g-lits nil
                       g-trueloc nil
                       g-tran nil
                       g-allf nil               ; used in xrefs
                       g-reguse '((r4 0 . nil) (r3 0 . nil)
                                  (r2 0 . nil) (r7 0 . nil) (r1 0 . nil))
                       g-trancnt 0
                       g-ignorereg nil
                       g-trueop  'jneq          ; used in e-gotot
                       g-falseop 'jeql          ; used in e-gotonil
                       g-compfcn nil
                       g-litcnt 0)
                 (setq g-spec (gensym 'S))      ; flag for special atom
                 (setq g-fname "")              ; no function yet
                 (setq special nil)             ; t if all vrbs are special
                 (setq g-functype (gensym)
                       g-vartype  (gensym)
                       g-bindtype (gensym)
                       g-calltype (gensym)
                       g-bindloc  (gensym)
                       g-localf   (gensym)
                       g-arrayspecs (gensym)
                       g-tranloc  (gensym)
                       g-stdref   (gensym)
                       g-optionalp (gensym))

                 ; declare these special

                 (sstatus feature complr)
;Orig            (d-makespec 't)                ; always special
                 (mapc 'd-makespec specvars*)   ;*IF*

                 ; process input form
                 (setq fl-asm t         ; assembler file assembled
                       fl-warn t        ; print warnings
                       fl-verb t        ; be verbose
                       fl-macl nil      ; compile maclisp file
                       fl-anno nil      ; annotate
                       fl-inter nil     ; do interlisp compatibility
                       fl-tty nil       ; put .s on tty
                       fl-comments nil    ; put in comments
                       fl-profile nil   ; profiling
                       fl-tran    t     ; use transfer tables
                       fl-vms   nil     ; vms hacks
                       fl-case  nil     ; trans uc to lc
                       fl-xref  nil     ; xrefs
                       fl-run   nil     ; autorun capability
                       fl-uci   nil     ; uci lisp compatibility
                       fl-slisp nil     ; standard lisp compatibility ;*IF*
                       )

                 ; look in the environment for a LISZT variable
                 ; if it exists, make it the first argument
                 (if (not (eq '|| (setq tem (getenv 'LISZT))))
                     then (setq v-x (cons (concat "-" tem) v-x)))

                 (do ((i v-x (cdr i)))  ; for each argument
                     ((null i))
                     (setq tem (aexplodec (car i)))

                     (cond ((eq '- (car tem))   ; if switch
                            (do ((j (cdr tem) (cdr j)))
                                ((null j))
                                (cond ((eq 'S (car j)) (setq fl-asm nil))
                                      ((eq 'C (car j)) (setq fl-comments t))
                                      ((eq 'm (car j)) (setq fl-macl t))
                                      ((eq 'o (car j)) (setq v-ofile (cadr i)
                                                             i (cdr i)))
                                      ((eq 'e (car j)) (setq pre-eval (cadr i)
                                                             i (cdr i)))
                                      ((eq 'I (car j)) (setq include-files
                                                        (cons (cadr i)
                                                             include-files)
                                                        i (cdr i)))
                                      ((eq 'l (car j)) (setq v-mfile ; *IF*
                                                        (cons (cadr i) v-mfile)
                                                             i (cdr i)))
                                      ((eq 'L (car j)) (setq v-Lfile ; *IF*
                                                        (cons (cadr i) v-Lfile)
                                                             i (cdr i)))
                                      ((eq 'w (car j)) (setq fl-warn nil))
                                      ((eq 'q (car j)) (setq fl-verb nil
                                                             *IPRN nil)) ;*IF*
                                      ((eq 'Q (car j)) (setq fl-verb t
                                                             *IPRN t))   ;*IF*
                                      ((eq 'T (car j)) (setq fl-tty t))
                                      ((eq 'a (car j)) (setq fl-anno t))
                                      ((eq 'i (car j)) (setq fl-inter t))
                                      ((eq 'p (car j)) (setq fl-profile t))
                                      ((eq 'F (car j)) (setq fl-tran nil))
                                      ((eq 'v (car j)) (setq fl-vms t))
                                      ((eq 'r (car j)) (setq fl-run t))
                                      ((eq 'x (car j)) (setq fl-xref t))
                                      ((eq 'c (car j)) (setq fl-case t))
                                      ((eq 'u (car j)) (setq fl-uci  t))
                                      ((eq 's (car j)) (setq fl-slisp t)) ;*IF*
                                      ((eq '- (car j)))  ; ignore extra -'s
                                      (t (comp-gerr "Unknown switch: "
                                                    (car j))))))
                           ((null v-root)
;Ori                        (setq temr (reverse tem))
;Ori                        (cond ((and (eq 'l (car temr))
;Ori                                    (eq '\. (cadr temr)))
;Ori                               (setq rootreal nil)
;Ori                               (setq v-root (apply 'concat
;Ori                                     (reverse (cddr temr)))))
;Ori                              (t (setq v-root (car i)
;Ori                                       rootreal t))))
                            (setq v-ifile (car i))              ;*IF*
                            (setq tem (filnam v-ifile))         ;*IF*
                            (setq v-root (cadr tem)             ;*IF*
                                  v-extn (cddr tem) )           ;*IF*
                            (setq rootreal (null v-extn)) )     ;*IF*
                           (t (comp-gerr "Extra input file name: " (car i)))))

                 (cond ((get v-extn 'SLISP) (setq fl-slisp t))) ;*IF*

                 ;no transfer tables in vms
                 (cond (fl-vms (setq fl-tran nil)))

                 ; if verbose mode, print out the gc messages and
                 ; fasl messages, else turn them off.

                 (cond (fl-verb (setq $gcprint t
                                      $ldprint t))
                       (t  (setq $gcprint nil
                                 $ldprint nil)))

                 ; eval arg after -e
                 (if pre-eval
                    then (if (null (errset
                                      (eval (readlist (exploden pre-eval)))))
                            then (comp-gerr "-e form caused error: "
                                            pre-eval)))

                 ; load file after -i arg
                 (if include-files
                    then (catch
                            (mapc
                               '(lambda (file)
                                   (if (null (errset (load file)))
                                      then (comp-err
                                              "error when loading -i file: "
                                              file)))
                               include-files)
                            Comp-error))

                 ; -c says set reader to xlate uc to lc
                 (cond (fl-case (sstatus uctolc t)))

                 ; now see what the arguments have left us

                 ; default extension             
                 (cond ((null v-extn)                           ;*IF*
                        (setq v-extn (cond (fl-slisp 'sl)       ;*IF*
                                           (t 'l)))))           ;*IF*

                 (cond ((null v-root)
                        (comp-gerr "No file for input"))
                       ((or (portp
                             (setq vp-ifile
;Ori                               (car (errset (infile
;Ori                                               (setq v-ifile
;Ori                                                   (concat v-root '".l")))
;Ori                                            nil)))) ;((
                                (car (errset (infile v-ifile) nil))))   ;*IF*
                            (and rootreal
                             (portp
                              (setq vp-ifile
                               (car (errset
;Ori                                             (infile (setq v-ifile v-root))
                                     (infile                    ; *IF*
                                          (setq v-ifile         ; *IF*
                                           (concat v-root '\. v-extn))) ;*IF*
                                             nil)))))))
                       (t (comp-gerr "Couldn't open the source file :"
                                     (or v-ifile))))


                 ; determine the name of the .s file
                 ; strategy: if fl-asm is t (only assemble) use (v-root).s
                 ;           else use /tmp/(PID).s
                 ;
                 ; direct asm to tty temporarily
                 (setq v-sfile '"tty")
                 (setq vp-sfile nil)
                 (if (null fl-tty) then
                    (cond (fl-asm (setq v-sfile (concat '"/tmp/Lzt"
                                                          (boole 1 65535
                                                                 (sys:getpid))
                                                           '".s")))
                           (t (setq v-sfile
                                    (if v-ofile
                                        then v-ofile
                                        else (concat v-root '".s")))))
                   (cond ((not (portp (setq vp-sfile
                                          (car (errset (outfile v-sfile)
                                                       nil)))))
                        (comp-gerr "Couldn't open the .s file: "
                                   (or v-sfile)))))

                 ; determine the name of the .o file (object file)
                 ; strategy: if we aren't supposed to assemble the .s file
                 ;            don't worry about a name
                 ;           else if a name is given, use it
                 ;           else if use (v-root).o
                 ;  if profiling, use .o
                 (cond ((or v-ofile (null fl-asm)))             ;ignore
                       ((null fl-profile) (setq v-ofile (concat v-root '".o")))
                       (t (setq v-ofile (concat v-root ".o"))))

                 ; determine the name of the .x file (xref file)
                 ; strategy: if fl-xref is true, then use (v-root).x
                 ;
                 (cond (fl-xref
                        (cond ((not
                                (portp
                                 (setq vp-xfile
                                      (car (errset (outfile (setq v-xfile
                                          (concat
                                           (cond
                                            (v-ofile (cadr (filnam v-ofile)))
                                            (t v-root)) ".x"))))))))
                               (comp-gerr "Can't open the .x file"
                                          (or v-xfile))))))
                 (cond ((checkfatal) (return 1)))

                 ; g-complrname is a symbol which should be unique to
                 ; each fasl'ed file. It will contain the string which
                 ; describes the name of this file and the compiler
                 ; version.
                 (if fl-anno
                    then (setq g-complrname (concat "fcn-in-" v-ifile))
                         (Push g-funcs
                               `(eval (setq ,g-complrname
                                            ,(get_pname
                                                (concat v-ifile
                                                        " compiled by "
                                                        compiler-name
                                                        " on "
                                                        (status ctime)))))))


                 (setq readtable (makereadtable nil))   ; use new readtable


                 (cond (v-Lfile         ; *IF* load premacrofile if any
                        (mapc 'load (nreverse v-Lfile)))) ; *IF*

                 ; if the macsyma flag is set, change the syntax to the
                 ; maclisp standard syntax.  We must be careful that we
                 ; dont clobber any syntax changes made by files preloaded
                 ; into the compiler.

                 (cond (fl-macl
                        (IFP (getd 'getsyntax) ; *IF* Opus 38 or later?
                             (progn     ; *IF*  Yes!
                                (setsyntax '\/ 'vescape)      ;  143 = vesc
                                (cond ((eq 'vescape (getsyntax '\\))
                                       (setsyntax '\\ 'vcharacter)))
                                (cond ((eq 'vleft-bracket (getsyntax '\[))
                                       (setsyntax '\[ 'vcharacter)
                                       (setsyntax '\] 'vcharacter))) )
                             (progn     ; *IF* Opus 37 or older.
                                  (setsyntax '\/ 143)   ;  143 = vesc

                                  (cond ((equal 143 (status syntax \\))
                                         (setsyntax '\\ 2)))
                                  (setsyntax '\| 138)   ;  138 = vdq
                                  (cond ((equal 198 (status syntax \[))
                                         (setsyntax '\[ 2)
                                         (setsyntax '\] 2))) ) ) ; *IF*
                        (setq ibase  8.)
                        (sstatus uctolc t)

                        (d-makespec 'ibase)     ; to be special
                        (d-makespec 'base)
                        (d-makespec 'tty)

                        (errset (cond ((null (getd 'macsyma-env))
                                       (load 'machacks)))
                                nil))
                       (fl-uci (load "ucifnc")
                                (cvttoucilisp)))

                 (cond (fl-inter (putprop '* 'cc-ignore 'fl-exprcc) ;comment
                                 (remprop '* 'fl-expr)
                                 ))

                 (cond (fl-slisp (cvttoslisp))) ; *IF* slisp environment

                 (cond (v-mfile         ; *IF* load macrofile if any
                        (mapc 'load (nreverse v-mfile)))) ; *IF*

                 (cond ((checkfatal) (return 1)))  ; leave if fatal errors

                 (if fl-verb
                     then (comp-msg "Compilation begins with " compiler-name )
                          (comp-msg "source: "  v-ifile ", result: "
                                    (cond (fl-asm v-ofile) (t v-sfile))))

                 (setq piport vp-ifile)         ; set to standard input
                 (setq liszt-root-name v-root
                       liszt-file-name v-ifile)

                 (if fl-run then (d-printautorun))

                 (if fl-profile then (e-write1 '".globl mcount"))
        loop

                ; main loop of the compiler.  It reads a form and
                ; compiles it. It continues to compile forms from
                ; liszt-process-forms as long at that list is
                ; non-empty.  This allows one form to spawn off other
                ; forms to be compiled (an alternative to (progn 'compile))
                ;
                (cond ((atom (errset            ; list for debugging,
                                                ; errset for production.
;Ori                          (do ((i (read piport '<<end-of-file>>)
;Ori                                  (read piport '<<end-of-file>>)))
;Ori                              ((eq i '<<end-of-file>>) nil) ;)
                              (do ((i (apply liszt-read nil)    ;*IF*
                                      (apply liszt-read nil)))  ;*IF*
                                  ((eq i $EOF$) nil)            ;*IF*
                                  (setq liszt-process-forms
                                        (cons i liszt-process-forms))
                                  (do ((this (car liszt-process-forms)
                                             (car liszt-process-forms)))
                                      ((null liszt-process-forms))
                                      (unpush liszt-process-forms)
                                      (catch (liszt-form this) Comp-error)))))
                       (catch (comp-err "Lisp error during compilation")
                              Comp-error)
                       (setq piport nil)
                       (setq er-fatal (1+ er-fatal))
                       (return 1)))

                 (close piport)

                 ; if doing special character stuff (maclisp) reassert
                 ; the state

                 (cond (vps-include
                        (comp-note  " done include")
                        (setq piport (car vps-include)
                              vps-include (cdr vps-include)
                              v-ifile (car vns-include)
                              vns-include (cdr vns-include))
                        (go loop)))

                 (cond (liszt-eof-forms
                        (do ((ll liszt-eof-forms (cdr ll)))
                            ((null ll))
                            (cond ((atom (errset (liszt-form (car ll))))
                                   (catch
                                    (comp-note "Lisp error during eof forms")
                                    Comp-error)
                                   (setq piport nil)
                                   (return 1))))))

                 ; reset input base
                 (setq ibase 10.)
                 (setq readtable (makereadtable t))
                 (sstatus uctolc nil)   ; turn off case conversion
                                        ; so bindtab will not have |'s
                                        ; to quote lower case
                 (d-bindtab)

                 (d-printdocstuff)              ; describe this compiler

                 (cond ((portp vp-sfile)
                        (close vp-sfile)))  ; close assembler language file

                 ; check for fatal errors and don't leave if so
                 (cond ((checkfatal)
                        (if fl-asm                      ; unlink .s file
                            then (sys:unlink v-sfile))  ; if it is a tmp
                        (return 1)))            ; and ret with error status

                 (comp-note "Compilation complete")

                 (setq tem (Divide (difference (sys:time) starttime) 60))
                 (setq ntem (ptime))

                 (setq temcp (Divide (difference (car ntem) (car startptime))
                                    3600))

                 (setq temgc (Divide (difference (cadr ntem) (cadr startptime))
                                    3600))

                 (comp-note " Time: Real: " (car tem) ":" (cadr tem)
                        ", CPU: " (car temcp) ":" (quotient (cadr temcp) 60.0)
                         ", GC: " (car temgc) ":" (quotient (cadr temgc) 60.0)
                            " for "
                            (difference $gccount$ startgccount)
                            " gcs")

                 (cond (fl-xref
                        (comp-note "Cross reference being generated")
                        (print (list 'File v-ifile) vp-xfile)
                        (terpr vp-xfile)
                        (do ((ii g-allf (cdr ii)))
                            ((null ii))
                            (print (car ii) vp-xfile)
                            (terpr vp-xfile))
                        (close vp-xfile)))


                 ; the assembler we use must generate the new a.out format
                 ; with a string table.  We will assume that the assembler
                 ; is in /usr/lib/lisp/as so that other sites can run
                 ; the new assembler without installing the new assembler
                 ; as /bin/as
                 (cond (fl-asm                  ; assemble file
                         (comp-note "Assembly begins")
                         (cond ((not
                                 (zerop
                                    (setq tmp
                                          (apply 'process
                                                 (ncons
                                                    (concat
                                                       lisp-library-directory
                                                       "/as  -o "
                                                       v-ofile
                                                       " "
                                                       v-sfile))))))
                                (comp-gerr "Assembler detected error, code: "
                                           asm-exit-status)
                                (comp-note "Assembler temp file " v-sfile
                                           " is not unlinked"))
                               (t (comp-note "Assembly completed successfully")
                                  (sys:unlink v-sfile); unlink tmp file
                                  (if fl-run
                                      then (errset
                                            (sys:chmod v-ofile #O775)))))))

                 (setq readtable original-readtable)
                 #+monitoring
                 (errset (progn (monitor)       ; turn off monitoring
                                (print 'monitor-off))
                         nil)
                 (sstatus nofeature complr)
                 (return asm-exit-status))))

; *IF* cvttoslisp loads the slisp macro environment if it is not loaded
(defun cvttoslisp ()
 (prog (tmp)
  (cond ((not (getd 'GETD))
         (cond ((and (not (numberp (setq tmp (argv 0))))
                     (setq tmp (car (filnam tmp))) )
                (Push slisp-load-search-path tmp) ))
         (eval (list 'sstatus 'load-search-path
                     (cons '|.| (append  slisp-load-search-path
                                         (ncons lisp-library-directory) )) ))
         (return (load 'slmacs)) )) ) )

;take apart a filname in both Unix and Vms format
(defun filnam (w) (prog (x y z)           ;Start of *IF*
  (setq x (setq w (explodec w)))
loop
  (cond
   ((setq y (or (memq '\/ x) (memq '\: x)))
    (setq z y) (setq x (cdr y)) (go loop) )
   ((setq y (or (memq '\] x) (memq '\> x))) (setq z y) (setq x (cdr y))) )
  (setq w (cond (z (rplacd z nil) (implode w))))
  (cond
   ((not (setq y (memq '\. x)))
    (return (cons w (ncons (cond (x (implode x)) (t nil))))) ) )
  (setq z (cdr y))
  (rplacd y nil)
  (setq x (nreverse (cdr (nreverse x))))
  (cond
   ((or (setq y (memq '\. z)) (setq y (memq '\; z)))
    (rplacd y nil)
    (setq z (nreverse (cdr (nreverse z)))) ) )
  (return (cons w (cons (implode x) (implode z)))) ))

(def franz-read
  (lambda ()
     (read piport $EOF$)))

(setq liszt-read 'franz-read)           ;End of *IF*


;--- liszt-form - i : form to compile
;       This compiles one form.
;
(def liszt-form
  (lambda (i)
     (prog (tmp v-x)
          (cond ((and (dtpr i) (symbolp (car i))                ;*IF*
                      (setq tmp (get (car i) 'lisztfn)) )
                 (apply tmp (list i)) )                         ;*IF*
                ((not (dtpr (setq i (d-macroexpand i))))
                 (comp-note "ignored: " (or i)))         ;*IF*
                ((not (symbolp (car i))) (Push g-funcs `(eval ,i)))     ;*IF*
                ((setq tmp (get (car i) 'lisztfn)) (apply tmp (list i))) ;*IF*
                ((eq (car i) 'def) ; jkf mod
                 (cond (fl-verb (print (cadr i)) (terpr) (drain)))
                 (d-dodef i))
                ((memq (car i) '(liszt-declare declare))
                 (funcall 'liszt-declare  (cdr i)))
                ((eq (car i) 'eval-when) (doevalwhen i))
                ((and (eq (car i) 'progn) (equal (cadr i) '(quote compile)))
                 ((lambda (internal-macros)     ; compile macros too
                          (mapc 'liszt-form (cddr i)))
                       t))
                ((or (and (eq (car i) 'includef) (setq tmp (eval (cadr i))))
                     (and (eq (car i) 'include ) (setq tmp (cadr i))))
                 (cond ((or (portp (setq v-x
                                         (car (errset (infile tmp) nil))))
                            (portp (setq v-x
                                         (car
                                            (errset
                                               (infile
                                                  (concat
                                                     lisp-library-directory
                                                     "/"
                                                     tmp))
                                               nil))))
                            (portp (setq v-x
                                         (car (errset (infile (concat tmp
;((Ori                                                                '".l"))
                                                             '\. v-extn)) ;*IF*
                                                      nil)))))
                        (setq vps-include (cons piport vps-include))
                        (setq piport v-x)
                        (comp-note " INCLUDEing file: "  tmp)
                        (setq vns-include (cons v-ifile vns-include)
                              v-ifile tmp))
                       (t (comp-gerr "Cannot open include file: " tmp))))
                ((eq (car i) 'comment) nil)   ; just ignore comments
                (t ; we have to macro expand
                   ; certain forms we would normally
                   ; just dump in the eval list.  This is due to hacks in
                   ; the mit lisp compiler which are relied upon by certain
                   ; code from mit.
                   (setq i (d-fullmacroexpand i))

                   (Push g-funcs `(eval ,i)) ) ))))


(defun d-doeval (i) (Push g-funcs `(eval ,i)))          ;*IF*

;End of patch to tlev.l

;Patch of expr.l

;--- d-cmpnil :: compare an IADR to nil
;
(IFP (not (getd 'd-cmpnil)) (eval-when (compile load)
  (defmacro d-cmpnil (iadr) `(d-tst ,iadr)) ))



;--- d-exp :: compile a lisp expression
;       v-form : a lisp expression to compile
; returns an IADR which tells where the value was located.
;

(defun d-exp (v-form)
  (prog (first resloc tmp ftyp nomacrop)
    begin
        (if (atom v-form)
            then (setq tmp (d-loc v-form))              ;locate vrble
                 (if (null g-loc)
                     then (if g-cc then (d-cmpnil tmp))
                    else (d-move tmp g-loc) )
                 (d-handlecc)
                 (return tmp)

         elseif (atom (setq first (car v-form)))
           then ; the form (*no-macroexpand* <expr>)
                ; turns into <expr>, and prevents <expr> from
                ; being macroexpanded (at the top level)
                (if (eq '*no-macroexpand* first)
                   then (setq v-form (cadr v-form)
                              nomacrop t)
                        (go begin))
                (if (and fl-xref (not (get first g-refseen)))
                     then (Push g-reflst first)
                          (putprop first t g-refseen))
                 (setq ftyp (d-functyp first (if nomacrop then nil
                                                else 'macros-ok)))
                 ; if nomacrop is t, then under no circumstances
                 ; permit the form to be macroexpanded
                 (if (and nomacrop (eq ftyp 'macro))
                     then (setq ftyp 'lambda))
                 ; If the resulting form is type macro or cmacro,
                 ; then call the appropriate function to macro-expand
                 ; it.
                 (if (memq ftyp '(macro cmacro))
                    then (setq tmp v-form)      ; remember original form
                         (if (eq 'macro ftyp)
                             then (setq v-form (apply first v-form))
                           elseif (eq 'cmacro ftyp)
                             then (setq v-form (apply (get first 'cmacro)
                                                      v-form)))
                          ; If the resulting form is the same as
                          ; the original form, then we don't want to
                          ; macro expand again.  We call d-functyp and tell
                          ; it that we want a second opinion
                          (if (and (eq (car v-form) first)
                                   (equal tmp v-form))
                             then (setq ftyp (d-functyp first nil))
                             else (go begin))) ; retry with what we have

                 (if (and (setq tmp (get first 'if-fixnum-args))
                              (d-allfixnumargs (cdr v-form)))
                    then (setq v-form (cons tmp (cdr v-form)))
                         (go begin)
                  elseif (setq tmp (get first 'fl-exprcc))
                    then (d-argnumchk 'hard)
                         (return (funcall tmp))
                  elseif (setq tmp (get first 'fl-exprm))
                    then (d-argnumchk 'hard)
                         (setq v-form (funcall tmp))
                         (go begin)
                  elseif (setq tmp (get first 'fl-expr))
                    then (d-argnumchk 'hard)
                         (funcall tmp)
                  elseif (setq tmp (or (and (eq 'car first)
                                            '( a ))
                                       (and (eq 'cdr first)
                                            '( d ))
                                       (d-cxxr first)))
                    then (d-argcheckit '(1 . 1) (length (cdr v-form)) 'hard)
                         (return (cc-cxxr (cadr v-form) tmp))
                   elseif (eq 'nlambda ftyp)
                    then (d-argnumchk 'soft)
                         (d-callbig first `(',(cdr v-form)) nil)
                   elseif (or (eq 'lambda ftyp) (eq 'lexpr ftyp))
                     then (setq tmp (length v-form))
                          (d-argnumchk 'soft)
                          (d-callbig first (cdr v-form) nil)
                   elseif (eq 'array ftyp)
                    then (d-handlearrayref)
                  elseif (eq 'macro ftyp)
                    then (comp-err "infinite macro expansion " v-form)
                    else (comp-err "internal liszt err in d-exp" v-form))

         elseif (eq 'lambda (car first))
            then (c-lambexp)

         elseif (or (eq 'quote (car first)) (eq 'function (car first)))
            then (comp-warn "bizzare function name " (or first))
                 (setq v-form (cons (cadr first) (cdr v-form)))
                 (go begin)

         else (comp-err "bad expression" (or v-form)))

        (if (null g-loc)
            then (if g-cc then (d-cmpnil 'reg))
         elseif (memq g-loc '(reg r0))
            then (if g-cc then (d-cmpnil 'reg))
           else (d-move 'reg g-loc))
        (if g-cc then (d-handlecc))))

;End of patch to expr.l

;---- load ; *IF* this is a redefinition of load as defined in common1.l

(declare (localf load-a-file))
(declare (special gcdisable load-most-recent tilde-expansion))

(or (boundp 'load-most-recent) (setq load-most-recent nil))
(or (boundp 'tilde-expansion) (setq tilde-expansion t))

(defun load (filename &rest fasl-args)
  (cond ((not (or (symbolp filename) (stringp filename)))
         (error "load: illegal filename " filename)))
  (let (search-path name pred shortname tem extn)

       (cond (tilde-expansion (setq filename (tilde-expand filename))))

       (setq pred (filnam filename))
       
       ; a short name is less or equal 12 characters.  If a name is not
       ; short, then load will not try to append .l or .o
       (cond ((and (null (setq extn (cddr pred)))
                   (or (< (flatc (cadr pred) 13) 13)
                       (status feature long-filenames) ) )
              (setq shortname t) ))

       (setq search-path (cond
         ((car pred) '(||))
         (t (status load-search-path)) ))

       (setq extn
        (do ((xx search-path (cdr xx)))
            ((null xx) (error "load: file not found " filename))

            (setq pred (cond
              ((memq (car xx) '(nil ||)) (concat filename))
              ((memq (car (last (exploden (car xx)))) '(#/: #/] #/> #//) )
               (concat (car xx) filename) )
              (t (concat (car xx) "/" filename)) ))

            (cond
             ((probef (setq name pred)) (return extn))
             ((not shortname))
             ((probef (setq name (concat pred ".o"))) (return (cond
                ((and load-most-recent (probef (setq tem (concat pred ".l")))
                      (< (filestat:mtime (filestat name))
                         (filestat:mtime (filestat tem)) ) )
                 (setq name tem) 'l)
                (t 'o) )))
             ((probef (setq name (concat pred ".l"))) (return 'l)) ) ) )

       (cond
         ((or (eq extn 'l) (not (or (eq extn 'o) fasl-args)))
          (load-a-file name) )
         (t (fasl-a-file name (car fasl-args) (cadr fasl-args))) ) ) )

;The rest are definitions of some functions that might not be defined or
; might be compiled local in certain versions.

(IFP (not (and (getd 'filestat) (getd 'filestat:mtime)))
     (setq load-most-recent nil))

;--- fasl-a-file
; The arguments are just like those to fasl.  This fasl's a file
; and if the translink's are set, it does the minimum work necessary to rebind
; the links (so that the new functions just fasl'ed in will be used).
;
(IFP (not (getd 'fasl-a-file))          ; This is from common1.l (or auxfns1.l)
 (defun fasl-a-file (name map warnflag)
   (let ((translinkarg (status translink)))
      (prog1
         (fasl name map warnflag)
         (cond ((and translinkarg (setq translinkarg (status translink)))
                ; if translink was set before and is still set
                (cond ((eq translinkarg t)
                       (sstatus translink nil)  ; clear all links
                       (sstatus tranlink t))    ; set to make links
                      (t ; must be 'on'
                         (sstatus translink on) ; recompute all links
                         ))))))) )

(declare (special $ldprint))    ; print message before loading

(IFP (not (getd 'load-a-file))          ; This is from common1.l (or auxfns1.l)
 (defun load-a-file (fname)
   (cond ($ldprint (patom "[load ") (patom fname) (patom "]") (terpr)))
   (let ((Piport (infile fname))
         ; (gcdisable t)        ; too dangerous: removed for now
         ; don't gc when loading, it slows things down
         (eof (list nil)))
      (do ((form (read Piport eof) (read Piport eof)))
          ((eq eof form) (close Piport) t)
          (eval form)))) )

(IFP (not (getd 'tilde-expand)) (setq tilde-expansion nil))

(IFP (not (getd 'do-lisztrc-check))     ; This is from topl.l (or car.l)
 ;--- do-lisztrc-check
 ; look for a liszt init file named
 ;  .lisztrc  or  lisztrc or $HOME/.lisztrc or $HOME/lisztrc
 ; followed by .o or .l
 ; return the symbol 'error' if an error occured while reading.
 ;
 (defun do-lisztrc-check nil
   (do ((dirs `("." ,(getenv 'HOME)) (cdr dirs)))
       ((null dirs))
       (cond ((do ((name '(".lisztrc" "lisztrc") (cdr name)))
                  ((null name))
                  (cond ((do ((ext '(".o" ".l") (cdr ext))
                              (file))
                             ((null ext))
                             (cond ((probef
                                       (setq file (concat (car dirs)
                                                          "/"
                                                          (car name)
                                                          (car ext))))
                                    (cond ((atom (errset (load file)))
                                           (comp-msg
                                              "Error loading liszt init file "
                                              file N
                                              "Compilation aborted" N)
                                           (return 'error)))
                                    (return t))))
                         (return t))))
              (return t)))))
)

(IFP (not (getd 'd-bindtab))            ; This is from topl.l (or car.l)
 ;--- d-bindtab :: emit binder table when all functions compiled
 ;
 (defun d-bindtab nil
  (setq g-skipcode nil)   ; make sure this isnt ignored
  (e-writel "bind_org")
  (e-write2 ".set linker_size," (length g-lits))
  (e-write2 ".set trans_size," (length g-tran))
  (do ((ll (setq g-funcs (nreverse g-funcs)) (cdr ll)))
      ((null ll))
      (if (memq (caar ll) '(lambda nlambda macro eval))
          then (e-write2 '".long" (cdr (assoc (caar ll)
                                              '((lambda . 0)
                                                (nlambda . 1)
                                                (macro . 2)
                                                (eval . 99)))))
          else (comp-err " bad type in lit list " (car ll))))

  (e-write1 ".long -1")
  (e-write1 '"lit_org:")
  (d-asciiout (nreverse g-lits))
  (if g-tran then (d-asciiout (nreverse g-tran)))
  (d-asciiout (mapcar '(lambda (x) (if (eq (car x) 'eval)
                                       then (cadr x)
                                       else (caddr x)))
                      g-funcs))

  (e-write1 '"lit_end:"))
)

(IFP (not (or (getd 'd-bindtab) (getd 'd-asciiout)))
; This is from topl.l (or car.l)
 ;--- d-asciiout :: print a list of asciz strings
 ;
  (defun d-asciiout (args)
       (do ((lits args (cdr lits))
            (form))
           ((null lits))
           (setq form (explode (car lits))
                 formsiz (length form))
           (do ((remsiz formsiz)
                (curform form)
                (thissiz))
               ((zerop remsiz))
               (if (greaterp remsiz 60) then (sfilewrite '".ascii \"")
                   else (sfilewrite '".asciz \""))
               (setq thissiz (min 60 remsiz))
               (do ((count thissiz (1- count)))
                   ((zerop count)
                    (sfilewrite (concat '\" (ascii 10)))
                    (setq remsiz (difference remsiz thissiz)))
                   (if (eq ch-newline (car curform))
                       then (sfilewrite '\\012)
                    else (if (or (eq '\\ (car curform))
                                 (eq '\" (car curform)))
                             then (sfilewrite '\\))
                         (sfilewrite (car curform)))
                   (setq curform (cdr curform))))))
)

(IFP (not (getd 'd-printdocstuff))
 (defun d-printdocstuff nil nil) )      ; Noop

(IFP (not (getd 'd-allfixnumargs))
 (defun d-allfixnumargs (x) nil) )      ; Never true

(IFP (not (getd 'd-fullmacroexpand))    ; Noop
 (defun d-fullmacroexpand (i) i) )

(IFP (not (getd 'liszt-declare))
 (putd 'liszt-declare (cond
  ((getd 'complr-declare))
  (t (getd 'declare)) )) )

(IFP (not (getd 'doevalwhen))   ; This is from topl.l (or car.l)
 (def doevalwhen
   (lambda (v-f)
     (prog (docom dolod)
           (setq docom (memq 'compile (cadr v-f))
                 dolod (memq 'load (cadr v-f)))
           (mapc '(lambda (frm) (cond (docom (eval frm)))
                    (cond (dolod ((lambda (internal-macros)
                                    (liszt-form frm))
                                  t))))
                 (cddr v-f))))))

;--- d-macroexpand :: macro expand a form as much as possible
;  from util.l
; only macro expands the top level though.
(IFP (not (getd 'd-macroexpand))
 (defun d-macroexpand (i)
   (prog (first type)
      loop
      (if (and (dtpr i) (symbolp (setq first (car i))))
         then (if (eq 'macro (setq type (d-functyp first 'macro-ok)))
                 then (setq i (apply first i))
                      (go loop)
               elseif (eq 'cmacro type)
                 then (setq i (apply (get first 'cmacro) i))
                      (go loop)))
      (return i))) )

;--- d-functyp :: return the type of function   From datab.l
;       - name : function name
;
; If name had a macro function definition, we return `macro'.  Otherwise
; we see if name has a declared type, if so we return that.  Otherwise
; we see if name is defined and we return that if so, and finally if
; we have no idea what this function is, we return lambda.
;   This is not really satisfactory, but will handle most cases.
;
; If macrochk is nil then we don't check for the macro case.  This
; is important to prevent recursive macroexpansion.
;
(IFP (not (and (dtpr (setq aaa (get 'd-functyp 'fcn-info)))
             (or (= (setq aaa (car aaa)) 2)
                 (and (dtpr aaa) (= (cdr aaa) 2)) ) ))
 (progn 'compile
  (putprop 'd-functyp '((1 . 2)) 'fcn-info)
  (defun d-functyp (name &optional (macrochk t)) ; *IF* second arg optional
   (let (func ftyp)
      (if (atom name)
         then
              (setq func (getd name))
              (setq ftyp (if (and macrochk (get name 'cmacro)) ;compiler macro
                            then 'cmacro
                          elseif (bcdp func)
                            then (let ((type (getdisc func)))
                                    (if (memq type '(lambda nlambda macro))
                                       then type
                                     elseif (stringp type)
                                       then 'lambda     ; foreign function
                                       else (comp-warn
                                               "function "
                                               name
                                               " has a strange discipline "
                                               type)
                                            'lambda     ; assume lambda
                                    ))
                          elseif (dtpr func)
                            then (car func)
                          elseif (and macrochk (get name 'macro-autoload))
                            then 'macro))
              (if (memq ftyp '(macro cmacro)) then ftyp
               elseif (d-findfirstprop name 'functype) thenret
               elseif (get name g-functype) thenret  ; check if declared first
               elseif ftyp thenret
                 else 'lambda)
         else 'lambda))) ) )             ; default is lambda

;From expr.l
;--- d-argnumchk :: check that the correct number of arguments are given
; v-form (global) contains the expression to check
; class: hard or soft, hard means that failure is an error, soft means
;       warning
(IFP (not (getd 'd-argnumcheck))
 (defun d-argnumchk (class)
   (let ((info (car (get (car v-form) 'fcn-info)))
         (argsize (length (cdr v-form))))
      (if info then (d-argcheckit info argsize class)))) )

;--- d-argcheckit    From expr.l
; info - arg information form:  (min# . max#)  max# of nil means no max
; numargs - number of arguments given
; class - hard or soft
; v-form(global) - expression begin checked
;
(IFP (not (getd 'd-argcheckit))
 (defun d-argcheckit (info numargs class)
   (if (and (car info) (< numargs (car info)))
      then (if (eq class 'hard)
              then (comp-err
                      (difference (car info) numargs)
                      " too few argument(s) given in this expression:" N
                      v-form)
              else (comp-warn
                      (difference (car info) numargs)
                      " too few argument(s) given in this expression:" N
                      v-form))
    elseif (and (cdr info) (> numargs (cdr info)))
      then (if (eq class 'hard)
              then (comp-err
                      (difference numargs (cdr info))
                      " too many argument(s) given in this expression:" N
                      v-form)
              else (comp-warn
                      (difference numargs (cdr info))
                      " too many argument(s) given in this expression:" N
                      v-form)))) )

;### PTLSZT.L end
