;### SLMAC.L ========================================================
;
;       SLMAC.L
;       -------
;
;       Copyright (c) 1982 by Inge Frick, Institute of Theoretical
;       Physics, Vanadisvagen 9, S-113 46 Stockholm, Sweden.
;
;       This file defines macros for the compilation of 
;       Sheep Lisp files. It includes the file slfuns.l.
;       
;====================================================================

(declare (special specvars* fl-verb)
         (macros t))

(defun samedefX (newname oldname)
 (putprop newname oldname 'Sameas)
 (putprop newname 'Samemac 'cmacro)
 (putd newname (getd oldname))
 newname)

(defun Samemac macro (x) (rplaca x (get (car x) 'Sameas)))

(putd 'samedef (getd 'samedefX))

(include "slfuns.l")

; FUNCTION must be a macro, not just a cmacro, to opencode the MAP
; functions.
(putd 'FUNCTION (getd 'Samemac))
(remprop 'FUNCTION 'cmacro)


(special $EOF$ $EOL$ EMSG* *ERRMSG *BAKGAG *ECHO *MSG *RAISE T !*COMP
         FRANZREADTABLE SLISPREADTABLE)

(setq *COMP 'T)

(macros t)

(defun Switchargs (newname oldname)
 (putprop newname oldname 'Similar)
 (putprop newname 'Arswitch 'cmacro) )

(defun Arswitch macro (u)
 (cons (get (car u) 'Similar)
  (cons (caddr u)
   (cons (cadr u) (cdddr u)) ) ) )

(defmacro cm-ATOM (x) `(not (dtpr ,x)))

(putprop 'ATOM 'cm-ATOM 'cmacro)

(defmacro cm-GENSYM () '(gensym 'G))

(putprop 'GENSYM 'cm-GENSYM 'cmacro)

(putprop 'EXPR 't 'COMPILE)

(putprop 'FEXPR 't 'COMPILE)

(putprop 'MACRO 't 'COMPILE)

(defun PUTDlisztfn (x) (let ((u (cadr x)) (v (caddr x)) (w (cadddr x)))
  (cond
   ((not (and (dtpr u) (eq (car u) 'quote))) (d-doeval x))
   ((GET (setq u (cadr u)) 'LOSE)
    (cond (fl-verb (print u) (princ " lost (LOSE)") (terpri))) )
   ((and (dtpr v) (eq (car v) 'quote) (GET (setq v (cadr v)) 'COMPILE)
         (dtpr w) (eq (car w) 'quote) (dtpr (setq w (cadr w))) )
    (cond (fl-verb (print u) (terpr) (drain)))
    (d-dodef `(def ,u ,(cons (get v 'fnTYP) (cdr w))))
    (d-doeval `(remprop ',u 'FNtyP)) )
   (t (d-doeval x)) ) ))

(putprop 'PUTD 'PUTDlisztfn 'lisztfn)

(defmacro cmacro-ERRORSET (x y z) (cond
 ((and (dtpr x) (eq (car x) 'quote))
  `(let ((*ERRMSG ,y) (*BAKGAG ,z)) (errset ,(cadr x) nil)))
 (t `(ERRORSET ,x ,y ,z)) ))

(putprop 'ERRORSET 'cmacro-ERRORSET 'cmacro)

(defun cm-COND macro (u)
 (map '(lambda (y)
   (cond
    ((or (eq (car y) 'T) (equal (car y) '(quote T)))
     (rplaca y 't) )) )
  (cdr u) )
 (rplaca u 'cond) )

(putprop 'COND 'cm-COND 'cmacro)

(mapcar (function Switchargs)
 '(MAPLIST MAPCAR MAPCON MAPCAN)
 '(maplist mapcar mapcon mapcan) )

(defun cm-MAP macro (u)
 `(progn (map ,(caddr u) ,(cadr u) ,@(cdddr u)) nil) )

(defprop MAP cm-MAP cmacro)

(defun cm-MAPC macro (u)
 `(progn (mapc ,(caddr u) ,(cadr u) ,@(cdddr u)) nil) )

(defprop MAPC cm-MAPC cmacro)

(defmacro cm-DIGIT (x) `(GET ,x 'Digit))

(defprop DIGIT cm-DIGIT cmacro)

(defmacro cm-LITER (x) `(GET ,x 'Letter))

(defprop LITER cm-LITER cmacro)

(defmacro cm-EVLIS (x) `(mapcar 'eval ,x))

(defprop EVLIS cm-EVLIS cmacro)

(defmacro cm-RDS (x) `(prog1 piport (setq piport ,x)))

(defprop RDS cm-RDS cmacro)

(defmacro cm-WRS (x) `(prog1 poport (setq poport ,x)))

(defprop WRS cm-WRS cmacro)

(defun INCLUDElisztfn (x) (liszt-form (list 'include (LKEY (cadr x)))))

(putprop 'INCLUDE 'INCLUDElisztfn 'lisztfn)

;(defun PROGNlisztfn (x) (mapc 'liszt-form (cdr x)))

;(putprop 'PROGN 'PROGNlisztfn 'lisztfn)

(defun slisp-read () (slisp-liszt (READ)))

(defun slisp-liszt (i) (cond
  ((not (and (dtpr i) (symbolp (car i)))) i)
  ((eq (car i) 'PROGN) 
   `(progn 'compile ,@(mapcar 'slisp-liszt (cdr i))) )
  ((get (car i) 'EVAL) `(progn 'compile (eval-when (compile) ,i) ,i))
  ((get (car i) 'IGNORE) `(eval-when (eval compile) ,i))
  ((and (get (car i) 'DFEXPAND) (eq (d-functyp (car i) t) 'macro))
   (slisp-liszt (apply (car i) i)) )
  (t i) ))

(setq liszt-read 'slisp-read)

(FLAG '(FLUID GLOBAL UNFLUID) 'EVAL)

(include "reatab.l")

;### SLMAC.L end
